/*  rpc -- A console-based reverse polish notation calculator
    Copyright (C) 2000, 2001, 2002 Paul Pelzl

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



#include "Complex.h"
#include "RpcCalcFlags.h"


// Default constructor
Complex::Complex()
{
    cpxdata.re = 0;
    cpxdata.im = 0;
}


// Default destructor
Complex::~Complex()
{
}


// Copy constructor
Complex::Complex( const Complex& oldComplex )
{
    cpxdata = oldComplex.cpxdata;
}


// Initialize specific complex value
Complex::Complex( double a, double b )
{
    cpxdata.re = a;
    cpxdata.im = b;
}


// Initialize specific complex value, allowing for polar format
Complex::Complex( bool polar, int angleMode, double a, double b )
{
    if(polar)
    {
        if(angleMode == MODE_DEG)
        {
            cpxdata.re = a*cos(b*M_PI/180.0);
            cpxdata.im = a*sin(b*M_PI/180.0);
        }
        else
        {
            cpxdata.re = a*cos(b);
            cpxdata.im = a*sin(b);
        }
    }
    else
    {
        cpxdata.re = a;
        cpxdata.im = b;
    }
}


// Initialize specific complex value
Complex::Complex( Cpx input )
{
    cpxdata = input;
}


// Initialize specific real-valued complex
Complex::Complex( double a )
{
    cpxdata.re = a;
    cpxdata.im = 0.0;
}


// Assignment operator
Complex& Complex::operator =( const Complex& rhs )
{
    if( this != &rhs )
    {
        cpxdata = rhs.cpxdata;
    }

    return( *this );
}


// Real part
double Complex::Real( void ) const
{
    return( cpxdata.re );
}


// Imaginary part
double Complex::Imag( void ) const
{
    return( cpxdata.im );
}


// Magnitude
double Complex::Abs( void ) const
{
    return( cAbs(cpxdata) );
}


// Argument of complex number
double Complex::Arg( void ) const
{
    return( atan2(cpxdata.im, cpxdata.re) );
}


// Addition operator
Complex Complex::operator +( const Complex& rhs )
{
    Complex result;
    result.cpxdata = cAdd(cpxdata, rhs.cpxdata);
    return( result );
}


// Subtraction operator
Complex Complex::operator -( const Complex& rhs )
{
    Complex result;
    result.cpxdata = cSub(cpxdata, rhs.cpxdata);
    return( result );
}


// Multiplication operator
Complex Complex::operator *( const Complex& rhs )
{
    Complex result;
    result.cpxdata = cMul(cpxdata, rhs.cpxdata);
    return( result );
}


// Division operator
Complex Complex::operator /( const Complex& rhs )
{
    Complex result;
    result.cpxdata = cDiv(cpxdata, rhs.cpxdata);
    return( result );
}


// Change sign operator
Complex Complex::operator -( void )
{
    Complex result;
    result.cpxdata = cRmu(-1.0, cpxdata);
    return( result );
}


// Addition to double
Complex Complex::operator +( const double& rhs )
{
    Complex result;
    result.cpxdata.re = cpxdata.re + rhs;
    result.cpxdata.im = cpxdata.im;
    return( result );
}


// Subtraction by double
Complex Complex::operator -( const double& rhs )
{
    Complex result;
    result.cpxdata.re = cpxdata.re - rhs;
    result.cpxdata.im = cpxdata.im;
    return( result );
}


// Multiplication by double
Complex Complex::operator *( const double& rhs )
{
    Complex result;
    result.cpxdata = cRmu(rhs, cpxdata);
    return( result );
}


// Division by double
Complex Complex::operator /( const double& rhs )
{
    Complex result;
    Cpx divisor = cDef(rhs, 0.0);
    result.cpxdata = cDiv(cpxdata, divisor);
    return( result );
}


// Complex power function.
// Note: CCMATH provides no power function, hence the full
// implementation here.
// Uses formula z^x = exp( x*log(z) )
Complex Complex::Pow( const double& power )
{
    Complex result;
    
    result.cpxdata = cExp( cRmu( power, cLog(cpxdata) ) );    
    return( result );
}


// Complex power function (with complex argument).
// Note: CCMATH provides no power function, hence the full
// implementation here.
// Uses formula z^q = exp( q*log(z) )
Complex Complex::Pow( const Complex& power )
{
    Complex result;
    
    result.cpxdata = cExp( cMul( power.cpxdata, cLog(cpxdata) ) );
    return( result );
}


// Complex square root
Complex Complex::Sqrt( void )
{
    Complex result;
    result.cpxdata = cSqrt(cpxdata);
    return( result );
}


// Complex square
Complex Complex::Sq( void )
{
    Complex result;
    result.cpxdata = cMul(cpxdata, cpxdata);
    return( result );
}


// Complex exponential
Complex Complex::Exp( void )
{
    Complex result;
    result.cpxdata = cExp(cpxdata);
    return( result );
}


// Complex logarithm
Complex Complex::Log( void )
{
    Complex result;
    result.cpxdata = cLog(cpxdata);
    return( result );
}


// Ten raised to the complex power x
Complex Complex::Ten_x( void )
{
    Complex result;
    result = Complex(10.0).Pow( Complex(cpxdata) );
    return( result );
}


// Complex common logarithm
Complex Complex::Log10( void )
{
    Complex result;
    result.cpxdata = cRmu( 1.0/log(10.0), cLog(cpxdata) );
    return( result );
}


// Complex sine
Complex Complex::Sin( void )
{
    Complex result;
    result.cpxdata = cSin(cpxdata);
    return( result );
}


// Complex cosine
Complex Complex::Cos( void )
{
    Complex result;
    result.cpxdata = cCos(cpxdata);
    return( result );
}


// Complex tangent
Complex Complex::Tan( void )
{
    Complex result;
    result.cpxdata = cTan(cpxdata);
    return( result );
}



// Complex inverse sine
Complex Complex::aSin( void )
{
    Complex result;
    result.cpxdata = cAsin(cpxdata);
    return( result );
}


// Complex inverse cosine
Complex Complex::aCos( void )
{
    Complex result;
    result.cpxdata = cAcos(cpxdata);
    return( result );
}


// Complex inverse tangent
Complex Complex::aTan( void )
{
    Complex result;
    result.cpxdata = cAtan(cpxdata);
    return( result );
}


// Complex hyperbolic sine
Complex Complex::Sinh( void )
{
    Complex result;
    result.cpxdata = cSinh(cpxdata);
    return( result );
}


// Complex hyperbolic cosine
Complex Complex::Cosh( void )
{
    Complex result;
    result.cpxdata = cCosh(cpxdata);
    return( result );
}


// Complex hyperbolic tangent
Complex Complex::Tanh( void )
{
    Complex result;
    result.cpxdata = cTanh(cpxdata);
    return( result );
}


// Complex inverse hyperbolic sine
Complex Complex::aSinh( void )
{
    Complex result;
    result.cpxdata = cAsinh(cpxdata);
    return( result );
}


// Complex inverse hyperbolic cosine
Complex Complex::aCosh( void )
{
    Complex result;
    result.cpxdata = cAcosh(cpxdata);
    return( result );
}


// Complex inverse hyperbolic tangent
Complex Complex::aTanh( void )
{
    Complex result;
    result.cpxdata = cAtanh(cpxdata);
    return( result );
}


// Complex inverse 
Complex Complex::Inv( void )
{
    Complex result;
    result.cpxdata = cDiv( cDef(1.0, 0.0), cpxdata );
    return( result );
}





// arch-tag: wrapper for ccmath complex math
