/*  rpc -- A console-based reverse polish notation calculator
    Copyright (C) 2000, 2001, 2002 Paul Pelzl

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


/*  ComplexMatrix.cpp
 *  Class definition for complex matrices.  Makes use of the CCMATH linear
 *  algebra routines.
 */


#include "ComplexMatrix.h"
#include <iostream>

using namespace std;

// Default constructor; creates a 1x1 zero matrix.
ComplexMatrix::ComplexMatrix()
{
   mtrxdata = new Cpx[1];
   m = 1;
   n = 1;
   singular = false;
   mtrxdata[0] = cDef(0.0, 0.0);
}


// Default destructor
ComplexMatrix::~ComplexMatrix()
{
   delete[] mtrxdata;
   mtrxdata = NULL;
}


// Construct a specific mxn matrix.
ComplexMatrix::ComplexMatrix( Complex* elements, int rows, int cols )
{
   m = rows;
   n = cols;
   singular = false;
   
   mtrxdata = new Cpx[m*n];
   
   for( int i=0; i<m*n; i++ )
   {
      mtrxdata[i].re = (elements+i)->Real();
      mtrxdata[i].im = (elements+i)->Imag();
   }
}


// Construct a specific mxn matrix.
ComplexMatrix::ComplexMatrix( Cpx* elements, int rows, int cols )
{
   m = rows;
   n = cols;
   singular = false;
   
   mtrxdata = new Cpx[m*n];
   
   for( int i=0; i<m*n; i++ )
   {
      mtrxdata[i].re = (elements+i)->re;
      mtrxdata[i].im = (elements+i)->im;
   }
}


// Construct a specific mxn matrix from a real matrix.
ComplexMatrix::ComplexMatrix( double* elements, int rows, int cols )
{
   m = rows;
   n = cols;
   singular = false;
   
   mtrxdata = new Cpx[m*n];
   
   for( int i=0; i<m*n; i++ )
   {
      mtrxdata[i].re = elements[i];
      mtrxdata[i].im = 0.0;
   }
}


// Construct a specific mxn matrix from a real matrix.
ComplexMatrix::ComplexMatrix( RealMatrix old )
{
   m = old.m;
   n = old.n;
   singular = old.singular;
   
   mtrxdata = new Cpx[m*n];
   
   for( int i=0; i<m*n; i++ )
   {
      mtrxdata[i].re = old.mtrxdata[i];
      mtrxdata[i].im = 0.0;
   }
}


// Construct an mxn zero matrix.
ComplexMatrix::ComplexMatrix( int rows, int cols )
{
   m = rows;
   n = cols;
   singular = false;
   
   mtrxdata = new Cpx[m*n];
   
   for( int i=0; i<m*n; i++ )
   {
      mtrxdata[i] = cDef( 0.0, 0.0 );
   }
}



// Copy constructor.
ComplexMatrix::ComplexMatrix( const ComplexMatrix& oldMatrix )
{
   m = oldMatrix.m;
   n = oldMatrix.n;
   singular = oldMatrix.singular;
   
   mtrxdata = new Cpx[m*n];
   
   for( int i=0; i<m*n; i++ )
   {
      mtrxdata[i].re = oldMatrix.mtrxdata[i].re;
      mtrxdata[i].im = oldMatrix.mtrxdata[i].im;
   }
}


// Assignment operator
ComplexMatrix& ComplexMatrix::operator =( const ComplexMatrix& rhs )
{
   if( this != &rhs )  // check self-assignment
   {
      m = rhs.m;
      n = rhs.n;
      singular = rhs.singular;
      
      delete[] mtrxdata;
      
      mtrxdata = new Cpx[m*n];
      
      for( int i=0; i<m*n; i++ )
      {
         mtrxdata[i].re = rhs.mtrxdata[i].re;
         mtrxdata[i].im = rhs.mtrxdata[i].im;
      }
   }
   
   return( *this );
}



// Addition operator.  Adds an mxn matrix to an mxn matrix,
// element-by-element.  Bounds are checked, but errors are not
// returned.
ComplexMatrix ComplexMatrix::operator +( const ComplexMatrix& rhs )
{
   ComplexMatrix result;
   
   if( m == rhs.m && n == rhs.n )
   {
      result = ComplexMatrix(m, n);
      
      for(int i=0; i<m*n; i++)
      {
         result.mtrxdata[i] = cAdd( mtrxdata[i], rhs.mtrxdata[i] );
      }
   }
   
   return( result );
}


// Subtraction operator.  Subtracts an mxn matrix from an
// mxn matrix, element-by-element.  Bounds are checked, but errors
// are not returned.
ComplexMatrix ComplexMatrix::operator -( const ComplexMatrix& rhs )
{
   ComplexMatrix result;
   
   if( m == rhs.m && n == rhs.n )
   {
      result = ComplexMatrix(m, n);
      
      for(int i=0; i<m*n; i++)
      {
         result.mtrxdata[i] = cSub( mtrxdata[i], rhs.mtrxdata[i] );
      }
   }
   
   return( result );
}


// Multiplication operator.  Multiplies an mxk matrix by a kxn
// matrix, element-by-element.  Bounds are checked, but errors
// are not returned.
ComplexMatrix ComplexMatrix::operator *( const ComplexMatrix& rhs )
{
   ComplexMatrix result;
   if( n == rhs.m )     // check that the dimensions match
   {
      result = ComplexMatrix(m, rhs.n);  
      Cmmult( result.mtrxdata, mtrxdata, rhs.mtrxdata, m, n, rhs.n );
   }
   
   return( result );
}


// Scalar multiplication operator.  Multiplies an mxn matrix by
// a complex scalar.
ComplexMatrix ComplexMatrix::operator *( const Complex& rhs )
{
   ComplexMatrix result(m, n);
   
   for(int i=0; i<m*n; i++)
   {
      result.mtrxdata[i] = cMul( mtrxdata[i], rhs.cpxdata );
   }

   return( result );
}


// Scalar division operator.  Divides an mxn matrix by
// a complex scalar.
ComplexMatrix ComplexMatrix::operator /( const Complex& rhs )
{
   ComplexMatrix result(m, n);
   
   for(int i=0; i<m*n; i++)
   {
      result.mtrxdata[i] = cDiv( mtrxdata[i], rhs.cpxdata );
   }
   
   return( result );
}


// Addition by real matrix operation.  (Element-by-element addition)
ComplexMatrix ComplexMatrix::operator +( const RealMatrix& rhs )
{
   ComplexMatrix result;
   
   if( m == rhs.m && n == rhs.n )
   {
      result = ComplexMatrix(m, n);
      
      for(int i=0; i<m*n; i++)
      {
         result.mtrxdata[i] = cAdd( mtrxdata[i], cDef(rhs.mtrxdata[i],0.0) );
      }
   }
   
   return( result );
}


// Subtraction by real matrix operation.  (Element-by-element subtraction)
ComplexMatrix ComplexMatrix::operator -( const RealMatrix& rhs )
{
   ComplexMatrix result;
   
   if( m == rhs.m && n == rhs.n )
   {
      result = ComplexMatrix(m, n);
      
      for(int i=0; i<m*n; i++)
      {
         result.mtrxdata[i] = cSub( mtrxdata[i], cDef(rhs.mtrxdata[i],0.0) );
      }
   }
   
   return( result );
}


// Multiplication operator.  Multiplies an mxk complex matrix by a kxn real
// matrix.  Bounds are checked, but errors are not returned.
ComplexMatrix ComplexMatrix::operator *( const RealMatrix& rhs )
{
   ComplexMatrix result;
   Cpx* oldData;
   
   if( n == rhs.m )     // check that the dimensions match
   {
      result = ComplexMatrix(m, rhs.n);  
      // Convert the double storage to Cpx storage, so we can use Cmmult()
      oldData = new Cpx[rhs.m * rhs.n];
      for(int i=0; i<rhs.m * rhs.n; i++)
      {
         oldData[i] = cDef( rhs.mtrxdata[i], 0.0 );
      }
      Cmmult( result.mtrxdata, mtrxdata, oldData, m, n, rhs.n );
      delete[] oldData;
   }
   
   return( result );
}


// Scalar multiplication operator.  Multiplies an mxn matrix by
// a real scalar.
ComplexMatrix ComplexMatrix::operator *( const double& rhs )
{
   ComplexMatrix result(m, n);
   
   for(int i=0; i<m*n; i++)
   {
      result.mtrxdata[i] = cRmu( rhs, mtrxdata[i] );
   }

   return( result );
}


// Scalar division operator.  Divides an mxn matrix by
// a real scalar.
ComplexMatrix ComplexMatrix::operator /( const double& rhs )
{
   ComplexMatrix result(m, n);
   
   for(int i=0; i<m*n; i++)
   {
      result.mtrxdata[i] = cDiv( mtrxdata[i], cDef(rhs, 0.0) );
   }
   
   return( result );
}



// Negation operator.  Multiplies an mxn matrix by a scalar -1.0.
ComplexMatrix ComplexMatrix::operator -( void )
{
   ComplexMatrix result(m, n);

   for(int i=0; i<m*n; i++)
   {
      result.mtrxdata[i] = cMul( mtrxdata[i], cDef(-1.0, 0.0) );
   }
   
   return( result );
}


// Square matrix inversion.  Performs dimension checking, but
// does not return errors.
ComplexMatrix ComplexMatrix::Inv( void )
{
   ComplexMatrix result( *this );
   int error = 0;
   
   if( m == n )
   {         
      error = cMinv( result.mtrxdata, m );
   }
   
   // Trap for singular matrix
   if( error == -1 )
   {
      result.singular = true;
   }
   
   return( result );
}
   

// Real matrix transpose.
ComplexMatrix ComplexMatrix::Trans( void )
{
   Cpx* mtrxtemp = new Cpx[m*n];
   cMattr( mtrxtemp, mtrxdata, m, n );  
   ComplexMatrix temp( mtrxtemp, n, m );
   delete[] mtrxtemp;   
   
   return( temp );
}
         
         
         
         
         
         
         
 
// arch-tag: wrapper for ccmath complex matrices
