/*  rpc -- A console-based reverse polish notation calculator
    Copyright (C) 2000, 2001, 2002 Paul Pelzl

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


/*  RealMatrix.cpp
 *  Class definition for real matrices.  Makes use of the CCMATH linear
 *  algebra routines.
 */


#include "RealMatrix.h"
#include <iostream>
#include "ComplexMatrix.h"

using namespace std;


// Default constructor; creates a 1x1 zero matrix.
RealMatrix::RealMatrix()
{
   mtrxdata = new double[1];
   m = 1;
   n = 1;
   singular = false;
   mtrxdata[0] = 0.0;
}


// Default destructor
RealMatrix::~RealMatrix()
{
   delete[] mtrxdata;
   mtrxdata = NULL;
}


// Construct a specific mxn matrix.
RealMatrix::RealMatrix( double* elements, int rows, int cols )
{
   m = rows;
   n = cols;
   singular = false;
   
   mtrxdata = new double[m*n];
   
   for( int i=0; i<m*n; i++ )
   {
      mtrxdata[i] = *(elements+i);
   }
}



// Construct an mxn zero matrix.
RealMatrix::RealMatrix( int rows, int cols )
{
   m = rows;
   n = cols;
   singular = false;
   
   mtrxdata = new double[m*n];
   
   for( int i=0; i<m*n; i++ )
   {
      mtrxdata[i] = 0.0;
   }
}



// Copy constructor.
RealMatrix::RealMatrix( const RealMatrix& oldMatrix )
{
   m = oldMatrix.m;
   n = oldMatrix.n;
   singular = oldMatrix.singular;
   
   mtrxdata = new double[m*n];
   
   for( int i=0; i<m*n; i++ )
   {
      mtrxdata[i] = oldMatrix.mtrxdata[i];
   }
}


// Assignment operator
RealMatrix& RealMatrix::operator =( const RealMatrix& rhs )
{
   if( this != &rhs )  // check self-assignment
   {
      m = rhs.m;
      n = rhs.n;
      singular = rhs.singular;
      
      delete[] mtrxdata;
      
      mtrxdata = new double[m*n];
      
      for( int i=0; i<m*n; i++ )
      {
         mtrxdata[i] = rhs.mtrxdata[i];
      }
   }
   
   return( *this );
}



// Addition operator.  Adds an mxn matrix to an mxn matrix,
// element-by-element.  Bounds are checked, but errors are not
// returned.
RealMatrix RealMatrix::operator +( const RealMatrix& rhs )
{
   RealMatrix result;
   
   if( m == rhs.m && n == rhs.n )
   {
      result = RealMatrix(m, n);
      
      for(int i=0; i<m*n; i++)
      {
         result.mtrxdata[i] = mtrxdata[i] + rhs.mtrxdata[i];
      }
   }
   
   return( result );
}


// Subtraction operator.  Subtracts an mxn matrix from an
// mxn matrix, element-by-element.  Bounds are checked, but errors
// are not returned.
RealMatrix RealMatrix::operator -( const RealMatrix& rhs )
{
   RealMatrix result;
   
   if( m == rhs.m && n == rhs.n )
   {
      result = RealMatrix(m, n);
      
      for(int i=0; i<m*n; i++)
      {
         result.mtrxdata[i] = mtrxdata[i] - rhs.mtrxdata[i];
      }
   }
   
   return( result );
}


// Multiplication operator.  Multiplies an mxk matrix by a kxn
// matrix, element-by-element.  Bounds are checked, but errors
// are not returned.
RealMatrix RealMatrix::operator *( const RealMatrix& rhs )
{
   RealMatrix result;
   if( n == rhs.m )     // check that the dimensions match
   {
      result = RealMatrix(m, rhs.n);  
      Rmmult( result.mtrxdata, mtrxdata, rhs.mtrxdata, m, n, rhs.n );
   }
   
   return( result );
}


// Scalar multiplication operator.  Multiplies an mxn matrix by
// a real scalar.
RealMatrix RealMatrix::operator *( const double& rhs )
{
   RealMatrix result(m, n);
   
   for(int i=0; i<m*n; i++)
   {
      result.mtrxdata[i] = mtrxdata[i]*rhs;
   }

   return( result );
}


// Scalar division operator.  Divides an mxn matrix by
// a real scalar.
RealMatrix RealMatrix::operator /( const double& rhs )
{
   RealMatrix result(m, n);
   
   for(int i=0; i<m*n; i++)
   {
      result.mtrxdata[i] = mtrxdata[i]/rhs;
   }
   
   return( result );
}


// Addition operator.  Adds a complex matrix to a real matrix,
// element-by-element.  Bounds are checked, but errors are not
// returned.
ComplexMatrix* RealMatrix::operator +( const ComplexMatrix& rhs )
{
   ComplexMatrix* result = NULL;
   
   if( m == rhs.m && n == rhs.n )
   {
      result = new ComplexMatrix(m, n);
      
      for(int i=0; i<m*n; i++)
      {
         result->mtrxdata[i] = cAdd( cDef(mtrxdata[i],0.0), rhs.mtrxdata[i]);
      }
   }
   
   return( result );
}


// Subtraction operator.  Subtracts an mxn complex matrix from an
// mxn real matrix, element-by-element.  Bounds are checked, but errors
// are not returned.
ComplexMatrix* RealMatrix::operator -( const ComplexMatrix& rhs )
{
   ComplexMatrix* result = NULL;
   
   if( m == rhs.m && n == rhs.n )
   {
      result = new ComplexMatrix(m, n);
      
      for(int i=0; i<m*n; i++)
      {
         result->mtrxdata[i] = cSub( cDef(mtrxdata[i],0.0), rhs.mtrxdata[i]);
      }
   }
   
   return( result );
}


// Multiplication operator.  Multiplies an mxk complex matrix by a kxn
// real matrix, element-by-element.  Bounds are checked, but errors
// are not returned.
ComplexMatrix* RealMatrix::operator *( const ComplexMatrix& rhs )
{
   ComplexMatrix* result = NULL;
   Cpx* cpxData;
   
   if( n == rhs.m )     // check that the dimensions match
   {
      result = new ComplexMatrix(m, rhs.n);  
      // Convert the real matrix data to type Cpx.
      cpxData = new Cpx[m*n];
      for(int i=0; i<m*n; i++)
      {
         cpxData[i] = cDef(mtrxdata[i], 0.0);
      }
      Cmmult( result->mtrxdata, cpxData, rhs.mtrxdata, m, n, rhs.n );
   }
   
   return( result );
}
   

// Scalar multiplication operator.  Multiplies an mxn real matrix by
// a complex scalar.
ComplexMatrix* RealMatrix::operator *( const Complex& rhs )
{
   ComplexMatrix* result = new ComplexMatrix(m, n);
   
   for(int i=0; i<m*n; i++)
   {
      result->mtrxdata[i] = cRmu( mtrxdata[i], rhs.cpxdata );
   }

   return( result );
}


// Scalar division operator.  Divides an mxn real matrix by
// a complex scalar.
ComplexMatrix* RealMatrix::operator /( const Complex& rhs )
{
   ComplexMatrix* result = new ComplexMatrix(m, n);
   
   for(int i=0; i<m*n; i++)
   {
      result->mtrxdata[i] = cDiv( cDef(mtrxdata[i],0.0), rhs.cpxdata);
   }
   
   return( result );
}



// Negation operator.  Multiplies an mxn matrix by a scalar -1.0.
RealMatrix RealMatrix::operator -( void )
{
   RealMatrix result(m, n);

   for(int i=0; i<m*n; i++)
   {
      result.mtrxdata[i] = -mtrxdata[i];
   }
   
   return( result );
}


// Square matrix inversion.  Performs dimension checking, but
// does not return errors.
RealMatrix RealMatrix::Inv( void )
{
   RealMatrix result( *this );
   int error = 0;
   
   if( m == n )
   {         
      error = Minv( result.mtrxdata, m );
   }
   
   // Trap for singular matrix
   if( error == -1 )
   {
      result.singular = true;
   }
   
   return( result );
}
   

// Real matrix transpose.
RealMatrix RealMatrix::Trans( void )
{
   double* mtrxtemp = new double[m*n];
   Mattr( mtrxtemp, mtrxdata, m, n );  
   RealMatrix temp( mtrxtemp, n, m );
   delete[] mtrxtemp;   
   
   return( temp );
}
         
         
         
         
         
         
         
 
// arch-tag: wrapper for ccmath real matrices
