/*  rpc -- A console-based reverse polish notation calculator
    Copyright (C) 2000, 2001, 2002 Paul Pelzl

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



/*  RpcCalc.cpp
 *  Class definition for a generic RPN calculator object.  Owns an
 *  RpcStack.  Has all the calculator functionality, but lacks an
 *  interface.
 */


#include "RpcCalc.h"


// Default constructor.  Create a new calculator.
RpcCalc::RpcCalc()
{    
    // Load the stack from a file.
    char rpcStackFile[256];
    strcpy( rpcStackFile, getenv( "HOME" ) );
    strcat( rpcStackFile, "/.rpc/stack" );
    
    ifstream stackFile;
    stackFile.open( rpcStackFile );
    
    if( !stackFile.bad() )
    {
        stack.load( stackFile );
        backup();
    }
    else
    {
        // If we can't find the stack file, most likely the
        // directory doesn't exist.
        system( "mkdir ~/.rpc" );
    }
    
    // Load the deg/rad mode from a file.
    char rpcFlagFile[256];
    strcpy( rpcFlagFile, getenv( "HOME" ) );
    strcat( rpcFlagFile, "/.rpc/calcflags" );
    
    ifstream flagFile;
    flagFile.open( rpcFlagFile, ios::in | ios::binary );
    
    char version[10];
    
    if( !flagFile.bad() )
    {
        // Replace this with a load() function in future.
        flagFile.read( version, sizeof(version) );
        
        if( strcmp( version, VERSION ) == 0 )
        {
            flagFile.read( (char*)(&flags.angleMode), sizeof(flags.angleMode) );
            flagFile.read( (char*)(&flags.constMode), sizeof(flags.constMode) );
            flagFile.read( (char*)(&flags.complexMode), sizeof(flags.complexMode) );
        }
        else
        {
            flags.angleMode = MODE_RAD;
            flags.constMode = MODE_DEC;
            flags.complexMode = MODE_RECT;
        }
    }
    flagFile.close();
}


// Destructor.
RpcCalc::~RpcCalc()
{

    // Save the stack to a file.
    char rpcStackFile[256];
    strcpy( rpcStackFile, getenv( "HOME" ) );
    strcat( rpcStackFile, "/.rpc/stack" );
    
    ofstream stackFile;
    stackFile.open( rpcStackFile );
    
    if( !stackFile.bad() )
    {
        stack.save( stackFile );
    }
    else
    {
        // Try creating the .rpc directory, then reopen the file.
        system( "mkdir ~/.rpc" );

        stackFile.close();
        stackFile.open( rpcStackFile );
        
        if( !stackFile.bad() )
        {
            stack.save( stackFile );
        }
    }
    
    // Save the mode to a file.
    char rpcFlagFile[256];
    strcpy( rpcFlagFile, getenv( "HOME" ) );
    strcat( rpcFlagFile, "/.rpc/calcflags" );
    
    ofstream flagFile;
    flagFile.open( rpcFlagFile );
    
    char version[10];
    strcpy( version, VERSION );
    
    if( !flagFile.bad() )
    {
        // Replace with a save() function in the future.
        flagFile.write( version, sizeof(version) );
        flagFile.write( (char*)(&flags.angleMode), sizeof(flags.angleMode) );
        flagFile.write( (char*)(&flags.constMode), sizeof(flags.constMode) );
        flagFile.write( (char*)(&flags.complexMode), sizeof(flags.complexMode) );
    }
    flagFile.close();
}


// Copy constructor.
RpcCalc::RpcCalc( const RpcCalc& oldCalc )
{
    flags.angleMode = oldCalc.flags.angleMode;
    flags.constMode = oldCalc.flags.constMode;
    flags.complexMode = oldCalc.flags.complexMode;
    stack = oldCalc.stack;
    backupStack = oldCalc.backupStack;
}


// Assignment operator.
RpcCalc& RpcCalc::operator =( const RpcCalc& rhs )
{
    if( this != &rhs )
    {
        flags.angleMode = rhs.flags.angleMode;
        flags.constMode = rhs.flags.constMode;
        flags.complexMode = rhs.flags.complexMode;
        stack = rhs.stack;
        backupStack = rhs.backupStack;
    }
    
    return( *this );
}


// Return the number of stack elements.
int RpcCalc::numElements( void ) const
{
    return( stack.numElements() );
}


// View a specific stack element.  Note: This corresponds to the
// HP-style view of the stack: level 1 is actually the top stack
// element (stack.count).
RpcObject* const RpcCalc::viewElement( int level ) const
{
    RpcObject* temp = NULL;
    
    if( (level>0) && (level<=stack.numElements()) )
    {
        temp = stack.viewElement( stack.numElements()+1-level );
    }
    else
    {
        temp = new RpcError;
    }
    
    return( temp );
}


// Get the display string for the object at the specified
// stack level.  Note: This corresponds to the HP-style view
// of the stack: level 1 is actually the top stack element
// (stack.count).
void RpcCalc::getDisplayString( char* displayString, int level ) const
{    
    char* temp = NULL;

    if( (level>0) && (level<=stack.numElements()) )
    {
        temp = stack.getDisplayString( stack.numElements()+1-level, flags );
        
        // We use strncpy because displayString only has dimension DISP_SIZE,
        // but temp may be much larger.
        strncpy(displayString, temp, DISP_SIZE-1);
        displayString[DISP_SIZE-1] = '\0';
        
        delete[] temp;        // Watch those memory leaks...
    }
    else
    {
        strncpy(displayString, "<no object at this level>", DISP_SIZE-1 );
        displayString[DISP_SIZE-1] = '\0';
    }
}


// Get the display string for the object at the specified
// stack level.  Note: This corresponds to the HP-style
// view of the stack: level 1 is actually the top stack
// element (stack.count).
void RpcCalc::getFullscreenString( char* fullscreenString, int level ) const
{
    char* temp = NULL;
    
    if( (level>0) && (level<=stack.numElements()) )
    {
        temp = stack.getFullscreenString( stack.numElements()+1-level, flags );
        
        // We use strncpy because fullscreenString only has dimension
        // FULLSCREEN_SIZE, but temp may be much larger.
        strncpy(fullscreenString, temp, FULLSCREEN_SIZE-1);
        fullscreenString[FULLSCREEN_SIZE-1] = '\0';
        
        delete[] temp;        // Watch those memory leaks...
    }
    else
    {
        strncpy(fullscreenString, "<no object at this level>\n", FULLSCREEN_SIZE-1 );
        fullscreenString[FULLSCREEN_SIZE-1] = '\0';
    }
} 



// Enter an RpcObject on the stack.  
int RpcCalc::enter( const RpcObject* input )
{
    stack.push( input );
    return( ERRORCODE_NOERR );
}


// Duplicate the last stack entry.
int RpcCalc::dup( void )
{
    int errorState = ERRORCODE_NOERR;
    RpcObject* temp = NULL;
    
    if( numElements() > 0 )
    {   
        temp = stack.pop();
        stack.push( temp );
        stack.push( temp );
        delete temp;
    }
    else
    {
        errorState = ERRORCODE_TOOFEW;    // Too few elements.
    }
    
    return( errorState );
}


// Drop the last stack entry.
int RpcCalc::drop( void )
{
    int errorState = ERRORCODE_NOERR;
    RpcObject* junk = NULL;
    
    if( stack.numElements() > 0 )
    {
        junk = stack.pop();        // Pop it into oblivion.
        delete junk;
    }
    
    return( errorState );
}


// Drop a specific stack entry.
// Requires messy stack reconstruction, handled by RpcStack.
int RpcCalc::drop( int level )
{
    int errorState = ERRORCODE_NOERR;
    int actualLevel = stack.numElements()-level;
    
    if( level > 0 && level <= stack.numElements() )
    {
        stack.drop( actualLevel );
    }
    else
    {
        errorState = ERRORCODE_TOOFEW;
    }
    
    return( errorState );
}    


// Drop the first n stack elements.
int RpcCalc::dropn( int level )
{
    int errorState = ERRORCODE_NOERR;
    int i;
    RpcObject* objTemp = NULL;
    
    if( stack.numElements() >= level && level > 0 )
    {
        // Just pop off the first n elements.
        for(i=0; i<level; i++)
        {
            objTemp = stack.pop();
            delete objTemp;
        }
    }
    else
    {
        errorState = ERRORCODE_TOOFEW;
    }

    return( errorState );
}
    

// Keep only the first n stack elements.
int RpcCalc::keepn( int level )
{
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= level && level > 0 )
    {
        stack.keepn( level );
    }
    else
    {
        errorState = ERRORCODE_TOOFEW;
    }
    
    return( errorState );
}


// Cyclically roll stack elements up to specified level.
int RpcCalc::rollUp( int level )
{
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= level && level > 0 )
    {
        // Actual stack is inverted, so we are rolling _down_.
        stack.rollDown( level );
    }
    else
    {
        errorState = ERRORCODE_TOOFEW;
    }
    
    return( errorState );
}


// Cyclically roll stack elements down to specified level.
int RpcCalc::rollDown( int level )
{
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= level && level > 0 )
    {
        // Actual stack is inverted, so we are rolling _up_.
        stack.rollUp( level );
    }
    else
    {
        errorState = ERRORCODE_TOOFEW;
    }
    
    return( errorState );
}


// Swap the last two stack entries.
int RpcCalc::swap( void )
{
    int errorState = ERRORCODE_NOERR;
    RpcObject* temp1 = NULL;
    RpcObject* temp2 = NULL;
    
    if( stack.numElements() > 1 )
    {
        temp1 = stack.pop();
        temp2 = stack.pop();
        
        stack.push( temp1 );
        stack.push( temp2 );
        
        delete temp1;
        delete temp2;
    }
    else
    {
        errorState = ERRORCODE_TOOFEW;
    }

    return( errorState );
}    


void RpcCalc::clearStack( void )
{
    stack.clearAll();
}


// Restore the last stack state.
void RpcCalc::undo( void )
{
    // Back up the current stack
    RpcStack temp = stack;
    
    // Restore the old stack
    stack = backupStack;
    
    backupStack = temp;
}


// Back up the stack.
void RpcCalc::backup( void )
{
    backupStack = stack;
}


// Add the top two RpcObjects from the stack.
int RpcCalc::add( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    RpcObject* c = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 2 )
    {
        a = stack.pop();
        b = stack.pop();
        
        c = (*a)+b;
        
        if( c->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)c)->errorCode;
            stack.push(b);                // Put popped elements back on stack
            stack.push(a);       
        }
        else if( c->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(b);
            stack.push(a);
        }
        else
        {
            stack.push(c);        // Put result back on stack
        }
    }
    else
    {
        errorState = ERRORCODE_TOOFEW;        // Too few elements
    }
    
    delete a;                  // Clean up memory
    delete b;
    delete c;
    
    return( errorState );
}


// Subtract RpcObject at level 1 from RpcObject at level 2.
int RpcCalc::subtract( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    RpcObject* c = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 2 )
    {
        a = stack.pop();
        b = stack.pop();
        
        c = (*b)-a;
        
        if( c->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)c)->errorCode;   // Illegal types for subtract()
            stack.push(b);         // Put popped elements back on stack
            stack.push(a);
        }
        else if( c->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(b);
            stack.push(a);
        }
        else
        {
            stack.push(c);         // Put result back on stack
        }
    }
    else
    {
        errorState = ERRORCODE_TOOFEW;        // Too few elements
    }
    
    delete a;                  // Clean up memory
    delete b;
    delete c;
    
    return( errorState );
}


// Multiply RpcObject at level 1 by RpcObject at level 2.
int RpcCalc::multiply( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    RpcObject* c = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 2 )
    {
        a = stack.pop();
        b = stack.pop();
        
        c = (*b) * a;
        
        if( c->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)c)->errorCode;   // Illegal types for multiply()
            stack.push(b);                // Put popped elements back on stack
            stack.push(a);
        }
        else if( c->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(b);
            stack.push(a);
        }        
        else
        {
            stack.push(c);         // Put result back on stack
        }
    }
    else
    {
        errorState = ERRORCODE_TOOFEW;        // Too few elements
    }
    
    delete a;                  // Clean up memory
    delete b;
    delete c;
    
    return( errorState );
}


// Divide RpcObject at level 2 by RpcObject at level 1.
int RpcCalc::divide( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    RpcObject* c = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 2 )
    {
        a = stack.pop();
        b = stack.pop();
        
        c = (*b) / a;
        
        if( c->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)c)->errorCode;    // Illegal types for divide()
            stack.push(b);         // Put popped elements back on stack
            stack.push(a);
        }
        else if( c->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(b);
            stack.push(a);
        }       
        else
        {
            stack.push(c);         // Put result back on stack
        }
    }
    else
    {
        errorState = ERRORCODE_TOOFEW;        // Too few elements
    }
    
    delete a;                  // Clean up memory
    delete b;
    delete c;
    
    return( errorState );
}


// Change sign of stack level 1.
int RpcCalc::changeSign( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = -(*a);
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;   // Illegal type for changeSign()
            stack.push(a);                // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;
    
    return( errorState );
}


// Raise RpcObject at level 2 to power of level 1
int RpcCalc::pow( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    RpcObject* c = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 2 )
    {
        a = stack.pop();
        b = stack.pop();
        
        c = b->Pow(a);
        
        if( c->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)c)->errorCode;   // Illegal types for pow()
            stack.push(b);                // Put popped elements back on stack
            stack.push(a);
        }
        else if( c->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(b);
            stack.push(a);
        }
        else
        {
            stack.push(c);         // Put result back on stack
        }
    }
    else
    {
        errorState = ERRORCODE_TOOFEW;  // Too few elements
    }
    
    delete a;                        // Clean up memory
    delete b;
    delete c;
    
    return( errorState );
}


// Evaluate the object at level 2, mod the object at level 1.
int RpcCalc::mod( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    RpcObject* c = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 2 )
    {
        a = stack.pop();
        b = stack.pop();
        
        c = b->Mod(a);
        
        if( c->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)c)->errorCode;   // Illegal types for mod()
            stack.push(b);                // Put popped elements back on stack
            stack.push(a);
        }
        else if( c->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(b);
            stack.push(a);
        }
        else
        {
            stack.push(c);         // Put result back on stack
        }
    }
    else
    {
        errorState = ERRORCODE_TOOFEW;  // Too few elements
    }
    
    delete a;                        // Clean up memory
    delete b;
    delete c;
    
    return( errorState );
}


// Square root function.
int RpcCalc::sqrt( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Sqrt();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;  // Illegal type for sqrt()
            stack.push(a);               // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Place result on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Square function.
int RpcCalc::sq( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Sq();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;   // Illegal type for sq()
            stack.push(a);                // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }        
        else
        {
            stack.push(b);         // Place result on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Exponential function.
int RpcCalc::exp( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Exp();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;  // Illegal type for exp()
            stack.push(a);               // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Natural logarithm.
int RpcCalc::log( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Log();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode; // Illegal type for log()
            stack.push(a);              // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Ten raised to the power of x.
int RpcCalc::ten_x( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Ten_x();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode; // Illegal type for ten_x()
            stack.push(a);              // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Common logarithm.
int RpcCalc::log10( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Log10();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode; // Illegal type for log10()
            stack.push(a);              // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Sine function.
int RpcCalc::sin( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Sin(flags.angleMode);
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for sin()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Cosine function.
int RpcCalc::cos( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Cos(flags.angleMode);
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for cos()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Tangent function.
int RpcCalc::tan( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Tan(flags.angleMode);
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for tan()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Inverse sine function.
int RpcCalc::asin( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->aSin(flags.angleMode);
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for asin()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}



// Inverse cosine function.
int RpcCalc::acos( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->aCos(flags.angleMode);
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for acos()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Tangent function.
int RpcCalc::atan( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->aTan(flags.angleMode);
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for atan()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Hyperbolic sine function.
int RpcCalc::sinh( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Sinh();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for sinh()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Hyperbolic cosine function.
int RpcCalc::cosh( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Cosh();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for cosh()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Hyperbolic tangent function.
int RpcCalc::tanh( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Tanh();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for tanh()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Inverse hyperbolic sine function.
int RpcCalc::asinh( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->aSinh();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for asinh()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Inverse hyperbolic cosine function.
int RpcCalc::acosh( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->aCosh();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for acosh()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Inverse hyperbolic tangent function.
int RpcCalc::atanh( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->aTanh();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for atanh()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Euler gamma function.
int RpcCalc::gamma( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Gamma();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for gamma()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Log gamma function.
int RpcCalc::lngamma( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->LogGamma();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for lngamma()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Factorial function, calculated from gamma.
int RpcCalc::factorial( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Factorial();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for factorial()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}

// Error function.
int RpcCalc::erf( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Erf();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for erf()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Log gamma function.
int RpcCalc::erfc( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Erfc();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for erfc()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Log gamma function.
int RpcCalc::real( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Real();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for real()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Log gamma function.
int RpcCalc::imag( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Imag();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for lngamma()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Floor function.
int RpcCalc::floor( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Floor();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for floor()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}

// Ceiling function.
int RpcCalc::ceil( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Ceil();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for ceil()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Nearest integer function.
int RpcCalc::nearint( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->NearInt();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for nearint()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Inverse function.
int RpcCalc::inv( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Inv();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for inv()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Transpose function.
int RpcCalc::trans( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Trans();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for trans()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Absolute value.
int RpcCalc::abs( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Abs();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for abs()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Argument of complex number
int RpcCalc::arg( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->Arg( flags.angleMode );
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for arg()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}


// Convert to integer constant.
int RpcCalc::toConst( void )
{
    RpcObject* a = NULL;
    RpcObject* b = NULL;
    
    int errorState = ERRORCODE_NOERR;
    
    if( stack.numElements() >= 1 )
    {
        a = stack.pop();
        b = a->ToConst();
        
        if( b->getType() == ERROR_TYPE )
        {
            errorState = ((RpcError*)b)->errorCode;        // Illegal type for trans()
            stack.push(a);         // Put popped element back on stack
        }
        else if( b->getType() == SIMPLE_ERROR_TYPE )
        {
            errorState = ERRORCODE_BADARG;
            stack.push(a);
        }
        else
        {
            stack.push(b);         // Put result back on stack
        }
    }
    // else do nothing
    
    delete a;                      // Clean up memory
    delete b;

    return( errorState );
}



// arch-tag: calculator object
