/*  rpc -- A console-based reverse polish notation calculator
    Copyright (C) 2000, 2001, 2002 Paul Pelzl

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/




/*  RpcCalc.h
 *  Class declaration for a generic RPN calculator object.  It owns
 *  an RpcStack, and has all the calculator functionality.  Lacks
 *  an interface.
 */
 
#ifndef RPC_CALC_H
#define RPC_CALC_H 1
 
#include "RpcStack.h"
#include "RpcCalcFlags.h"
#include <cstdlib>
#include <cstring>

using namespace std;


// Note: if these get changed, you need to worry about
//       the array ordering in errorStrings[].
#define ERROR_NOERR 0
#define ERROR_BAD_ARG 1
#define ERROR_TOO_FEW 2

#define DISP_SIZE 256
#define FULLSCREEN_SIZE 65536

class RpcCalc
{
    protected:
        RpcStack stack;
        RpcStack backupStack;
        
    public:
        RpcCalc();                       // Create a new calculator
        ~RpcCalc();                      // Destroy calc
        
        RpcCalc( const RpcCalc& );                // Copy constructor
        RpcCalc& operator =( const RpcCalc& );    // Assignment

        RpcCalcFlags flags;              // Various calculator mode flags        
        
        int numElements( void ) const;   // Return the number of stack elements
        
        RpcObject* const viewElement( int ) const;  // View a specific stack element
        
        void getDisplayString( char*, int ) const;  // Get the display string for
                                              // the specified stack level.
        void getFullscreenString( char*, int ) const;   // Get the fullscreen display
                                              // string for the specified stack level.
        
        int enter( const RpcObject* );   // Place an RpcObject on stack.
        int dup( void );                 // Duplicate the last stack entry.
        int drop( void );                // Drop the last stack entry.
        int drop( int );                 // Drop a specific stack entry.
        int dropn( int );                // Drop first n stack levels.
        int keepn( int );                // Keep first n stack levels.
        int rollUp( int );               // Cyclically roll stack elements up.
        int rollDown( int );             // Cyclically roll stack elements down.
        int swap( void );                // Swap last two stack entries.
        void clearStack( void );         // Clear the stack contents.
        
        void undo( void );               // Restore the last stack state.
        void backup( void );             // Back up the stack.
        
        int add( void );                 // Add two stack elements
        int subtract( void );            // Subtract level 1 from level 2
        int multiply( void );            // Multiply two stack elements
        int divide( void );              // Divide level 2 by level 1
        
        int changeSign( void );          // Change sign of stack level 1
        
        int pow( void );                 // Raise to a power
        
        int mod( void );                 // Mod operator
        
        int sqrt( void );                // Square root
        int sq( void );                  // Square
        
        int exp( void );                 // Exponential
        int log( void );                 // Natural logarithm
        
        int ten_x( void );               // Ten raised to the power x
        int log10( void );               // Common logarithm
        
        int sin( void );                 // Sine
        int cos( void );                 // Cosine
        int tan( void );                 // Tangent
        
        int asin( void );                // Inverse sine
        int acos( void );                // Inverse cosine
        int atan( void );                // Inverse tangent
        
        int sinh( void );                // Hyperbolic sine
        int cosh( void );                // Hyperbolic cosine
        int tanh( void );                // Hyperbolic tangent
        
        int asinh( void );               // Inverse hyperbolic sine
        int acosh( void );               // Inverse hyperbolic cosine
        int atanh( void );               // Inverse hyperbolic tangent
        
        int gamma( void );               // Euler gamma function.
        int lngamma( void );             // Log gamma function.
        int factorial( void );           // Factorial (calculated from gamma).
        
        int erf( void );                 // Error function.
        int erfc( void );                // Complementary error function.
        
        int real( void );                // Real part of number.
        int imag( void );                // Imaginary part of number.
        
        int floor( void );               // Floor function.
        int ceil( void );                // Ceiling function.
        int nearint( void );             // Nearest integer function.
        
        int inv( void );                 // Inverse function.
        
        int trans( void );               // Transpose matrix.
        
        int abs( void );                 // Magnitude (modulus, absolute value)
        int arg( void );                 // Phase of complex
        
        int toConst( void );             // Convert to integer constant.
};

#endif



// arch-tag: calculator object (header)
