/*  rpc -- A console-based reverse polish notation calculator
    Copyright (C) 2000, 2001, 2002 Paul Pelzl

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



#include "RpcComplex.h"
#include "RpcComplexMatrix.h"

// Default constructor--a zero element.
RpcComplex::RpcComplex()
{
    type = COMPLEX_TYPE;
    data = Complex( 0.0, 0.0 );
}


// Default destructor.
RpcComplex::~RpcComplex()
{
}


// Copy constructor.
RpcComplex::RpcComplex( const RpcComplex& oldComplex )
{
    type = COMPLEX_TYPE;
    data = oldComplex.data;
}


// Assignment operator.
RpcComplex& RpcComplex::operator =( const RpcComplex& rhs )
{
    if( this != &rhs )
    {
        type = COMPLEX_TYPE;
        data = rhs.data;
    }
    
    return( *this );
}


// Initialize a specific complex value.
RpcComplex::RpcComplex( Complex value )
{
    type = COMPLEX_TYPE;
    data = value;
}


// Initialize a specific complex value.
RpcComplex::RpcComplex( Cpx value )
{
    type = COMPLEX_TYPE;
    data.cpxdata = value;
}


// Initialize a specific real value.
RpcComplex::RpcComplex( double value )
{
    type = COMPLEX_TYPE;
    data.cpxdata.re = value;
    data.cpxdata.im = 0.0;
}


// Return the display string for this object.
char* RpcComplex::getDisplayString( RpcCalcFlags flags ) const
{
    char* temp = new char[256];
    if( flags.complexMode == MODE_RECT )
    {   
        sprintf( temp, "( %.15g, %.15g )", data.Real(), data.Imag() );
    }
    else if(flags.angleMode == MODE_DEG)
    {
        sprintf( temp, "( %.15g <%.15g )", data.Abs(), data.Arg()*180.0/M_PI );
    }
    else
    {
        sprintf( temp, "( %.15g <%.15g )", data.Abs(), data.Arg() );
    }       
    return( temp );
}
    

// Return the fullscreen display string for this object.
char* RpcComplex::getFullscreenString( RpcCalcFlags flags ) const
{
    char* temp = new char[256];
    if( flags.complexMode == MODE_RECT )
    {
        sprintf(temp, "( %.15g, %.15g )\n", data.Real(), data.Imag() );
    }
    else if(flags.angleMode == MODE_DEG)
    {
        sprintf(temp, "( %.15g <%.15g )\n", data.Abs(), data.Arg()*180.0/M_PI );
    }
    else
    {
        sprintf(temp, "( %.15g <%.15g )\n", data.Abs(), data.Arg() );
    }
    return( temp );
}


    
// Addition operator.
RpcObject* RpcComplex::operator +( const RpcObject* rhs )
{
    RpcObject* temp;
    
    // Branch according to data type.
    switch( rhs -> getType() )
    {
        // Add a double to a complex.
        case DOUBLE_TYPE:
            temp = new RpcComplex( data + ((RpcDouble*)rhs)->getData() );
            break;
            
        // Add two complex types.
        case COMPLEX_TYPE:
            temp = new RpcComplex( data + ((RpcComplex*)rhs)->getData() );
            break;
    
        // Default is an error.
        default:
            temp = new RpcError(ERRORCODE_BADARG);
            break;
    }
    
    return( temp );
}


// Subtraction operator.
RpcObject* RpcComplex::operator -( const RpcObject* rhs )
{
    RpcObject* temp;
    
    // Branch according to data type.
    switch( rhs -> getType() )
    {
        // Subtract a double from a complex.
        case DOUBLE_TYPE:
            temp = new RpcComplex( data - 
                        Complex( ((RpcDouble*)rhs)->getData() ) );
            break;
            
        // Subtract a complex from a complex.
        case COMPLEX_TYPE:
            temp = new RpcComplex( data - ((RpcComplex*)rhs)->getData() );
            break;
            
        // Default is an error.
        default:        
            temp = new RpcError(ERRORCODE_BADARG);
            break;
    }
    
    return( temp );
}


// Multiplication operator.
RpcObject* RpcComplex::operator *( const RpcObject* rhs )
{
    RpcObject* temp;
    ComplexMatrix* dummy;
    
    // Branch according to data type.
    switch( rhs -> getType() )
    {
        // Multiply a complex by a double.
        case DOUBLE_TYPE:
            temp = new RpcComplex( data *
                        Complex( ((RpcDouble*)rhs)->getData() ) );
            break;
            
        // Multiply two complex objects.
        case COMPLEX_TYPE:
            temp = new RpcComplex( data * ((RpcComplex*)rhs)->getData() );
            break;
            
        // Multiply a complex by a real matrix.
        case REALMATRIX_TYPE:
            dummy = (((RpcRealMatrix*)rhs)->getData()) * data;
            temp = new RpcComplexMatrix( *dummy );
            delete dummy;
            break;
            
        // Multiply a complex by a complex matrix
        case COMPLEXMATRIX_TYPE:
            temp = new RpcComplexMatrix( (((RpcComplexMatrix*)rhs)->getData()) * data );
            break;
                
        // Default is an error.
        default:
            temp = new RpcError(ERRORCODE_BADARG);
            break;
    }
    
    return( temp );
}


// Division operator.
RpcObject* RpcComplex::operator /( const RpcObject* rhs )
{
    RpcObject* temp;
    
    // Branch according to data type.
    switch( rhs -> getType() )
    {
        // Divide a complex by a double.
        case DOUBLE_TYPE:
            temp = new RpcComplex( data /
                    Complex( ((RpcDouble*)rhs)->getData() ) );
            break;
            
        // Divide a complex by a complex.
        case COMPLEX_TYPE:
            temp = new RpcComplex( data / ((RpcComplex*)rhs)->getData() );
            break;
            
        // Default is an error.
        default:
            temp = new RpcError(ERRORCODE_BADARG);
            break;
    }
    
    return( temp );
}


// Change sign operator.
inline RpcObject* RpcComplex::operator -( void )
{
    RpcObject* temp = new RpcComplex( -data );
    return( temp );
}


// Power function.
RpcObject* RpcComplex::Pow( const RpcObject* power )
{
    RpcObject* temp;
    
    // Branch according to data type.
    switch( power -> getType() )
    {
        // Raise a complex to a double power.
        case DOUBLE_TYPE:
            temp = new RpcComplex( data.Pow(
                    ((RpcDouble*)power)->getData() ) );
            break;
                        
        // Raise a complex to a complex power.
        case COMPLEX_TYPE:
            temp = new RpcComplex( data.Pow(
                    ((RpcComplex*)power)->getData() ) );
            break;
        
        // Default is an error.
        default:
            temp = new RpcError(ERRORCODE_BADARG);
            break;
    }
    
    return( temp );
}


// Square root function.
inline RpcObject* RpcComplex::Sqrt( void )
{
    RpcObject* temp = new RpcComplex( data.Sqrt() );
    return( temp );
}


// Square function.
inline RpcObject* RpcComplex::Sq( void )
{
    RpcObject* temp = new RpcComplex( data*data );
    return( temp );
}


// Exponential function.
inline RpcObject* RpcComplex::Exp( void )
{
    RpcObject* temp = new RpcComplex( data.Exp() );
    return( temp );
}


// Natural logarithm.
inline RpcObject* RpcComplex::Log( void )
{
    RpcObject* temp = new RpcComplex( data.Log() );
    return( temp );
}


// Ten raised to the power x.
inline RpcObject* RpcComplex::Ten_x( void )
{
    RpcObject* temp = new RpcComplex( data.Ten_x() );
    return( temp );
}


// Common logarithm.
inline RpcObject* RpcComplex::Log10( void )
{
    RpcObject* temp = new RpcComplex( data.Log10() );
    return( temp );
}


// Sine function.
inline RpcObject* RpcComplex::Sin( int mode )
{
    RpcObject* temp = new RpcComplex( data.Sin() );
    return( temp );
}


// Cosine function.
inline RpcObject* RpcComplex::Cos( int mode )
{
    RpcObject* temp = new RpcComplex( data.Cos() );
    return( temp );
}


// Tangent function.
inline RpcObject* RpcComplex::Tan( int mode )
{
    RpcObject* temp = new RpcComplex( data.Tan() );
    return( temp );
}


// Inverse sine function.
inline RpcObject* RpcComplex::aSin( int mode )
{
    RpcObject* temp = new RpcComplex( data.aSin() );
    return( temp );
}


// Inverse cosine function.
inline RpcObject* RpcComplex::aCos( int mode )
{
    RpcObject* temp = new RpcComplex( data.aCos() );
    return( temp );
}


// Inverse tangent function.
inline RpcObject* RpcComplex::aTan( int mode )
{
    RpcObject* temp = new RpcComplex( data.aTan() );
    return( temp );
}


// Hyperbolic sine function.
inline RpcObject* RpcComplex::Sinh( void )
{
    RpcObject* temp = new RpcComplex( data.Sinh() );
    return( temp );
}


// Hyperbolic cosine function.
inline RpcObject* RpcComplex::Cosh( void )
{
    RpcObject* temp = new RpcComplex( data.Cosh() );
    return( temp );
}


// Hyperbolic tangent function.
inline RpcObject* RpcComplex::Tanh( void )
{
    RpcObject* temp = new RpcComplex( data.Tanh() );
    return( temp );
}


// Inverse hyperbolic sine function.
inline RpcObject* RpcComplex::aSinh( void )
{
    RpcObject* temp = new RpcComplex( data.aSinh() );
    return( temp );
}


// Inverse hyperbolic cosine function.
inline RpcObject* RpcComplex::aCosh( void )
{
    RpcObject* temp = new RpcComplex( data.aCosh() );
    return( temp );
}


// Inverse hyperbolic tangent function.
inline RpcObject* RpcComplex::aTanh( void )
{
    RpcObject* temp = new RpcComplex( data.aTanh() );
    return( temp );
}


// Real part of complex number.
inline RpcObject* RpcComplex::Real( void )
{
    RpcObject* temp = new RpcDouble( data.Real() );
    return( temp );
}


// Imaginary part of complex number.
inline RpcObject* RpcComplex::Imag( void )
{
    RpcObject* temp = new RpcDouble( data.Imag() );
    return( temp );
}


// Inverse of complex number.
inline RpcObject* RpcComplex::Inv( void )
{
    RpcObject* temp = new RpcComplex( data.Inv() );
    return( temp );
}


// Magnitude of complex number
inline RpcObject* RpcComplex::Abs( void )
{
    RpcObject* temp = new RpcDouble( data.Abs() );
    return( temp );
}


// Argument of complex number
inline RpcObject* RpcComplex::Arg( int angleMode )
{
    RpcObject* temp;
    if( angleMode == MODE_DEG )
    {
        temp = new RpcDouble( data.Arg()*180.0/M_PI );
    }
    else
    {
        temp = new RpcDouble( data.Arg() );
    }
    return( temp );
}


// Return the data.
inline Complex RpcComplex::getData( void )
{
    return( data );
}

    
    

// arch-tag: rpc complex data type
