/*  rpc -- A console-based reverse polish notation calculator
    Copyright (C) 2000, 2001, 2002 Paul Pelzl

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include "RpcComplexMatrix.h"

// Default constructor -- a 1x1 zero element.
RpcComplexMatrix::RpcComplexMatrix()
{
   type = COMPLEXMATRIX_TYPE;
   data = ComplexMatrix();
}


// Default destructor
RpcComplexMatrix::~RpcComplexMatrix()
{
}


// Copy constructor
RpcComplexMatrix::RpcComplexMatrix( const RpcComplexMatrix& oldMatrix )
{
   type = COMPLEXMATRIX_TYPE;
   data = oldMatrix.data;
}


// Assignment operator.
RpcComplexMatrix& RpcComplexMatrix::operator =( const RpcComplexMatrix& rhs )
{
   if( this != &rhs )   // check assignment
   {
      type = COMPLEXMATRIX_TYPE;
      data = rhs.data;
   }
   
   return( *this );
}


// Initialize a specific complex matrix.
RpcComplexMatrix::RpcComplexMatrix( ComplexMatrix value )
{
   type = COMPLEXMATRIX_TYPE;
   data = value;
}


// Initialize a specific complex matrix.
RpcComplexMatrix::RpcComplexMatrix( Complex* elements, int m, int n )
{
   type = COMPLEXMATRIX_TYPE;
   data = ComplexMatrix( elements, m, n );
}


// Return the display string for this object.
char* RpcComplexMatrix::getDisplayString( RpcCalcFlags flags ) const
{
   int i=0, j=0;
   int m = data.m, n = data.n;
   char* temp = new char[ m * n * 70 ];
   char numBuffer[70];
   Complex ctemp;
   double atemp;
   
   strcpy( temp, "[" );
   for( i=0; i<m; i++ )
   {
      strcat( temp, "[ " );
      for( j=0; j<n; j++ )
      {
         strcat( temp, "( " );
         if( flags.complexMode == MODE_RECT )
         {
             sprintf( numBuffer, "%.15g, %.15g", (data.mtrxdata[i*n+j]).re,
                                                 (data.mtrxdata[i*n+j]).im );
         }
         else
         {         
             ctemp = Complex( data.mtrxdata[i*n+j] );
             atemp = ctemp.Arg();
             if( flags.angleMode == MODE_DEG )
                 atemp *= 180.0/M_PI;
                 
             sprintf( numBuffer, "%.15g <%.15g", ctemp.Abs(), atemp );
         }
         strcat( temp, numBuffer );
         strcat( temp, " )" );
         if( j < (n-1) )
         {
            strcat( temp, ", " );
         }
      }
      strcat( temp, " ]" );
   }
   strcat( temp, "]" );
   return( temp );
}


// Return the fullscreen display string for this object.
char* RpcComplexMatrix::getFullscreenString( RpcCalcFlags flags ) const
{
   int i=0, j=0;
   int m = data.m, n = data.n;
   char* temp = new char[ m * n * 70 ];
   char numBuffer[70];
   int widthRe, widthIm, maxwidthRe[n], maxwidthIm[n];
   Complex ctemp;
   double atemp;
   
   // First get the maximum field width for each column (both real and
   // imaginary parts, separately, so everything will line up nicely)
   for( j=0; j<n; j++ )
   {
      maxwidthRe[j] = 0;
      maxwidthIm[j] = 0;
      for( i=0; i<m; i++ )
      {  
         if( flags.complexMode == MODE_RECT )
         { 
            sprintf( numBuffer, "%-.15g", data.mtrxdata[i*n+j].re );
            widthRe = strlen( numBuffer );
            sprintf( numBuffer, "%-.15g", data.mtrxdata[i*n+j].im );
            widthIm = strlen( numBuffer );
         }
         else
         {
            ctemp = Complex( data.mtrxdata[i*n+j] );
            atemp = ctemp.Arg();
            if( flags.angleMode == MODE_DEG )
               atemp *= 180.0/M_PI;
            
            sprintf( numBuffer, "%-.15g", ctemp.Abs() );
            widthRe = strlen( numBuffer );
            sprintf( numBuffer, "%-.15g", atemp );
            widthIm = strlen( numBuffer );
         }
            
         if( widthRe > maxwidthRe[j] )
            maxwidthRe[j] = widthRe;
         if( widthIm > maxwidthIm[j] )
            maxwidthIm[j] = widthIm;
      }
   }
      
   
   strcpy( temp, "[" );
   for( i=0; i<m; i++ )
   {
      if( i != 0 )
         strcat( temp, " " );
      strcat( temp, "[ " );
      for( j=0; j<n; j++ )
      {
         if( flags.complexMode == MODE_RECT )
         {
            sprintf( numBuffer, "( %*.15g, %*.15g )", maxwidthRe[j], 
                (data.mtrxdata[i*n+j]).re, maxwidthIm[j], (data.mtrxdata[i*n+j]).im );
         }
         else
         {
            ctemp = Complex( data.mtrxdata[i*n+j] );
            atemp = ctemp.Arg();
            if( flags.angleMode == MODE_DEG )
               atemp *= 180/M_PI;
            
            sprintf( numBuffer, "( %*.15g <%*.15g )", maxwidthRe[j], 
                ctemp.Abs(), maxwidthIm[j], atemp );
         }
           
         strcat( temp, numBuffer );
         if( j < (n-1) )
         {
            strcat( temp, ", " );
         }
      }
      
      if( i == m-1 )
      {
         strcat( temp, " ]" );
      }
      else
      {
         strcat( temp, " ]\n" );
      }
   }
   strcat( temp, "]\n" );
   
   return( temp );
}



// Addition operator.
RpcObject* RpcComplexMatrix::operator +(const RpcObject* rhs)
{
   RpcObject* temp;
   
   // Branch according to data type.
   switch( rhs -> getType() )
   {
      // Add a real matrix to a complex matrix.
      case REALMATRIX_TYPE:
         // check dimension.
         if( data.m == ((RpcRealMatrix*)rhs)->getData().m &&
               data.n == ((RpcRealMatrix*)rhs)->getData().n )
         {
            temp = new RpcComplexMatrix( data + ((RpcRealMatrix*)rhs)->getData() );
         }
         else
         {
            temp = new RpcError;
         }
         break;
         
      // Add a complex matrix to a complex matrix.
      case COMPLEXMATRIX_TYPE:
         // check dimension.
         if( data.m == ((RpcComplexMatrix*)rhs)->getData().m && 
               data.n == ((RpcComplexMatrix*)rhs)->getData().n )
         {
            temp = new RpcComplexMatrix( data + ((RpcComplexMatrix*)rhs)->getData() );
         }
         else
         {
            temp = new RpcError;
         }         
         break;
         
      // Default is an error.
      default:
         temp = new RpcError;
         break;
   }
   
   return( temp );
}

         

// Subtraction operator.
RpcObject* RpcComplexMatrix::operator -(const RpcObject* rhs)
{
   RpcObject* temp;
   
   // Branch according to data type.
   switch( rhs -> getType() )
   {
      // Subtract a real matrix from a complex matrix.
      case REALMATRIX_TYPE:
         // check dimension.
         if( data.m == ((RpcRealMatrix*)rhs)->getData().m &&
               data.n == ((RpcRealMatrix*)rhs)->getData().n )
         {
            temp = new RpcComplexMatrix( data - ((RpcRealMatrix*)rhs)->getData() );
         }
         else
         {
            temp = new RpcError;
         }
         break;
         
      // Subtract a complex matrix from a complex matrix.
      case COMPLEXMATRIX_TYPE:
         // check dimension.
         if( data.m == ((RpcComplexMatrix*)rhs)->getData().m && 
               data.n == ((RpcComplexMatrix*)rhs)->getData().n )
         {
            temp = new RpcComplexMatrix( data - ((RpcComplexMatrix*)rhs)->getData() );
         }
         else
         {
            temp = new RpcError;
         }
         break;
         
      // Default is an error.
      default:
         temp = new RpcError;
         break;
   }
   
   return( temp );
}



// Multiplication operator.
RpcObject* RpcComplexMatrix::operator *( const RpcObject* rhs )
{
   RpcObject* temp;
   
   // Branch according to data type.
   switch( rhs -> getType() )
   {
      // Multiply a complex matrix by a real scalar.
      case DOUBLE_TYPE:
         temp = new RpcComplexMatrix( data * ((RpcDouble*)rhs)->getData() );
         break;        
         
      // Multiply a complex matrix by a complex.
      case COMPLEX_TYPE:
         temp = new RpcComplexMatrix( data * ((RpcComplex*)rhs)->getData() );
         break;
      
      // Multiply a complex matrix by a real matrix.
      case REALMATRIX_TYPE:
         temp = new RpcComplexMatrix( data * ((RpcRealMatrix*)rhs)->getData() );
         break;
      
      // Multiply two complex matrices.
      case COMPLEXMATRIX_TYPE:
         // check dimension.
         if( data.n == ((RpcComplexMatrix*)rhs)->getData().m )
         {
            temp = new RpcComplexMatrix( data * ((RpcComplexMatrix*)rhs)->getData() );
         }
         else
         {
            temp = new RpcError;
         }
         break;
         
      // Default is an error.
      default:
         temp = new RpcError;
         break;
   }
   
   return( temp );
}



// Division operator.
RpcObject* RpcComplexMatrix::operator /( const RpcObject* rhs )
{
   RpcObject* temp;
   
   // Branch according to data type.
   switch( rhs -> getType() )
   {
      // Divide a complex matrix by a double.
      case DOUBLE_TYPE:
         temp = new RpcComplexMatrix( data / ((RpcDouble*)rhs)->getData() );
         break;
         
      // Divide a complex matrix by a complex.
      case COMPLEX_TYPE:
         temp = new RpcComplexMatrix( data / ((RpcComplex*)rhs)->getData() );
         break;
               
      // Default is an error.
      default:
         temp = new RpcError;
         break;
   }
   
   return( temp );
}



// Change sign operator.
inline RpcObject* RpcComplexMatrix::operator -( void )
{
   RpcObject* temp = new RpcComplexMatrix( -data );
   return( temp );
}



// Matrix inversion.
inline RpcObject* RpcComplexMatrix::Inv( void )
{
   RpcObject* temp;
   
   // check for square matrix
   if( data.m == data.n )
   {
      ComplexMatrix inverse = data.Inv();
      if( inverse.singular )
      {
         temp = new RpcError( ERRORCODE_SINGULAR );
      }
      else
      {
         temp = new RpcComplexMatrix( inverse );
      }
   }
   else
   {
      temp = new RpcError;
   }
   
   return( temp );
}



// Matrix inversion.
inline RpcObject* RpcComplexMatrix::Trans( void )
{
   RpcObject* temp = new RpcComplexMatrix( data.Trans() );
   return( temp );
}



// Return the data.
inline ComplexMatrix RpcComplexMatrix::getData( void )
{
   return( data );
}







// arch-tag: rpc complex matrix data type
