/*  rpc -- A console-based reverse polish notation calculator
    Copyright (C) 2000, 2001, 2002 Paul Pelzl

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



/*  RpcDouble.cpp
 *  Definition of a the rpc double precision data type.
 */
 
#include "RpcDouble.h"
#include "RpcComplex.h"
#include "RpcRealMatrix.h"
#include "RpcIntegerConst.h"
#include "RpcError.h"


const double deg2rad = M_PI / 180.0;


// Default constructor: creates a zero element.
RpcDouble::RpcDouble()
{
    type = DOUBLE_TYPE;
    data = 0.0;
}


// Default destructor.
RpcDouble::~RpcDouble()
{
}


// Constructor to initialize specific value.
RpcDouble::RpcDouble( double input )
{
    type = DOUBLE_TYPE;
    data = input;
}


// Copy constructor.
RpcDouble::RpcDouble( const RpcDouble& oldObject )
{
    type = DOUBLE_TYPE;
    data = oldObject.data;
}


// Assignment operator.
RpcDouble& RpcDouble::operator =( const RpcDouble& oldObject )
{ 
    if( this != &oldObject)
    {
        type = DOUBLE_TYPE;
        data = oldObject.data;
    }
    
    return( *this );
}


// Return the display string for this object.
char* RpcDouble::getDisplayString( RpcCalcFlags flags ) const
{
    char* temp = new char[256];
    sprintf( temp, "%.15g", data );  // Translate double to formatted string
    return( temp );
}    


// Return the fullscreen display string for this object.
char* RpcDouble::getFullscreenString( RpcCalcFlags flags ) const
{
    char* temp = new char[256];
    sprintf( temp, "%.15g\n", data ); // Translate double to formatted string
    return( temp );
}
 

// Addition operator.
RpcObject* RpcDouble::operator +( const RpcObject* rhs )
{
    RpcObject* temp;

    // Branch through all the possible data types.
    switch( rhs -> getType() )
    {
        // Add two doubles
        case DOUBLE_TYPE:
            temp = new RpcDouble( data + ((RpcDouble*)rhs)->getData() );
            break;
        
        // Add a complex to a double
        case COMPLEX_TYPE:
            temp = new RpcComplex( ((RpcComplex*)rhs)->getData() + data );
            break;
            
        // Add an integer const to a double.
        case INTEGERCONST_TYPE:
            temp = new RpcIntegerConst( ((RpcIntegerConst*)rhs)->getData() + data );
            // trap for integer const overflow
            if( (((RpcIntegerConst*)temp)->getData()).overflow )
            {
                delete temp;
                temp = new RpcError(ERRORCODE_CONST_OVERFLOW);
            }
            break;
                
        // Default is an error.    
        default:
            temp = new RpcError(ERRORCODE_BADARG);
            break;
    }
    
    return( temp );
}


// Subtraction operator.
RpcObject* RpcDouble::operator -( const RpcObject* rhs )
{
    RpcObject* temp;
    
    // Branch through all possible data types.
    switch( rhs -> getType() )
    {
        // Subtract two doubles
        case DOUBLE_TYPE:
            temp = new RpcDouble( data - ((RpcDouble*)rhs)->getData() );
            break;
            
        // Subtract a complex from a double
        case COMPLEX_TYPE:
            temp = new RpcComplex( Complex(data) - ((RpcComplex*)rhs)->getData() );
            break;
            
        // Subtract an integer const from a double
        case INTEGERCONST_TYPE:
            temp = new RpcIntegerConst( IntegerConst(data) - ((RpcIntegerConst*)rhs)->getData() );
            break;
        
        // Default is an error.
        default:
            temp = new RpcError(ERRORCODE_BADARG);
            break;
    }
    
    return( temp );
}


// Multiplication operator.
RpcObject* RpcDouble::operator *( const RpcObject* rhs )
{
    RpcObject* temp;
    
    // Branch through all possible data types.
    switch( rhs -> getType() )
    {
        // Multiply two doubles.
        case DOUBLE_TYPE:
            temp = new RpcDouble( data * ((RpcDouble*)rhs)->getData() );
            break;
            
        // Multiply a double by a complex.
        case COMPLEX_TYPE:
            temp = new RpcComplex( Complex(data) * 
                    ((RpcComplex*)rhs)->getData() );
            break;
            
        // Multiply a double by a real matrix.
        case REALMATRIX_TYPE:
            temp = new RpcRealMatrix( ((RpcRealMatrix*)rhs)->getData() * data );
            break;
            
        // Multiply a double by a complex matrix.
        case COMPLEXMATRIX_TYPE:
            temp = new RpcComplexMatrix( ((RpcComplexMatrix*)rhs)->getData() * data );
            break;
            
        case INTEGERCONST_TYPE:            
            temp = new RpcIntegerConst( ((RpcIntegerConst*)rhs)->getData() * data );
            // trap for integer overflow
            if( (((RpcIntegerConst*)rhs)->getData()).overflow )
            {
                delete temp;
                temp = new RpcError(ERRORCODE_CONST_OVERFLOW);
            }
            break;
            
        // Default is an error.
        default:
            temp = new RpcError(ERRORCODE_BADARG);
            break;
    }
    
    return( temp );
}


// Division operator.
RpcObject* RpcDouble::operator /( const RpcObject* rhs )
{
    RpcObject* temp;
    
    // Branch through all possible data types.
    switch( rhs -> getType() )
    {
        // Divide a double by a double.
        case DOUBLE_TYPE:
            temp = new RpcDouble( data / ((RpcDouble*)rhs)->getData() );
            break;
            
        // Divide a double by a complex.
        case COMPLEX_TYPE:
            temp = new RpcComplex( Complex(data) /
                    ((RpcComplex*)rhs)->getData() );
            break;
            
        // Divide a double by an integer constant.
        case INTEGERCONST_TYPE:
            temp = new RpcIntegerConst( IntegerConst(data) /
                    ((RpcIntegerConst*)rhs)->getData() );
            // check for integer const overflow
            if( (((RpcIntegerConst*)temp)->getData()).overflow )
            {
                delete temp;
                temp = new RpcError(ERRORCODE_CONST_OVERFLOW);
            }
            break;
            
        // Default is an error.
        default:
            temp = new RpcError(ERRORCODE_BADARG);
            break;
    }
    
    return( temp );
}


// Change sign operator.
inline RpcObject* RpcDouble::operator -( void )
{
    RpcObject* temp = new RpcDouble( -data );
    return( temp );
}


// Power function.
RpcObject* RpcDouble::Pow( const RpcObject* power )
{
    RpcObject* temp;
    
    // Branch through all possible data types.
    switch( power -> getType() )
    {
        // Raise a double to a double power.
        case DOUBLE_TYPE:
            temp = new RpcDouble( pow(data,((RpcDouble*)power)->getData()) );
            break;
        
        // Raise a double to a complex power.
        case COMPLEX_TYPE:
            temp = new RpcComplex( Complex(data).Pow(
                    ((RpcComplex*)power)->getData() ) );
            break;
                            
        // Default is an error.
        default:
            temp = new RpcError(ERRORCODE_BADARG);
            break;
    }
    
    return( temp );
}


// Square function.
inline RpcObject* RpcDouble::Sq( void )
{
    RpcObject* temp = new RpcDouble( data*data );
    return( temp );
}


// Square root function.
inline RpcObject* RpcDouble::Sqrt( void )
{
    RpcObject* temp = new RpcDouble( sqrt(data) );
    return( temp );
}


// Exponential function.
inline RpcObject* RpcDouble::Exp( void )
{
    RpcObject* temp = new RpcDouble( exp(data) );
    return( temp );
}


// Natural logarithm.
inline RpcObject* RpcDouble::Log( void )
{
    RpcObject* temp = new RpcDouble( log(data) );
    return( temp );
}


// Ten raised to the power x.
inline RpcObject* RpcDouble::Ten_x( void )
{
    RpcObject* temp = new RpcDouble( pow(10.0, data) );
    return( temp );
}


// Common logarithm.
inline RpcObject* RpcDouble::Log10( void )
{
    RpcObject* temp = new RpcDouble( log10(data) );
    return( temp );
}


// Sine function.
inline RpcObject* RpcDouble::Sin( int mode )
{
    double dtemp = data;
    if( mode == MODE_DEG )
    {
        dtemp *= deg2rad;
    }
    
    RpcObject* temp = new RpcDouble( sin(dtemp) );
    return( temp );
}


// Cosine function.
inline RpcObject* RpcDouble::Cos( int mode )
{
    double dtemp = data;
    if( mode == MODE_DEG )
    {
        dtemp *= deg2rad;
    }
    
    RpcObject* temp = new RpcDouble( cos(dtemp) );
    return( temp );
}


// Tangent function.
inline RpcObject* RpcDouble::Tan( int mode )
{
    double dtemp = data;
    if( mode == MODE_DEG )
    {
        dtemp *= deg2rad;
    }
    
    RpcObject* temp = new RpcDouble( tan(dtemp) );
    return( temp );
}


// Inverse sine function.
inline RpcObject* RpcDouble::aSin( int mode )
{
    double dtemp = 1.0;
    if( mode == MODE_DEG )
    {
        dtemp *= deg2rad;
    }
    
    RpcObject* temp = new RpcDouble( asin(data)/dtemp );
    return( temp );
}


// Inverse cosine function.
inline RpcObject* RpcDouble::aCos( int mode )
{
    double dtemp = 1.0;
    if( mode == MODE_DEG )
    {
        dtemp *= deg2rad;
    }
    
    RpcObject* temp = new RpcDouble( acos(data)/dtemp );
    return( temp );
}


// Inverse tangent function.
inline RpcObject* RpcDouble::aTan( int mode )
{
    double dtemp = 1.0;
    if( mode == MODE_DEG )
    {
        dtemp *= deg2rad;
    }
    
    RpcObject* temp = new RpcDouble( atan(data)/dtemp );
    return( temp );
}


// Hyperbolic sine function.
inline RpcObject* RpcDouble::Sinh( void )
{
    RpcObject* temp = new RpcDouble( sinh(data) );
    return( temp );
}


// Hyperbolic cosine function.
inline RpcObject* RpcDouble::Cosh( void )
{
    RpcObject* temp = new RpcDouble( cosh(data) );
    return( temp );
}


// Hyperbolic tangent function.
inline RpcObject* RpcDouble::Tanh( void )
{
    RpcObject* temp = new RpcDouble( tanh(data) );
    return( temp );
}


// Inverse hyperbolic sine function.
inline RpcObject* RpcDouble::aSinh( void )
{
    RpcObject* temp = new RpcDouble( asinh(data) );
    return( temp );
}


// Inverse hyperbolic cosine function.
inline RpcObject* RpcDouble::aCosh( void )
{
    RpcObject* temp = new RpcDouble( acosh(data) );
    return( temp );
}


// Inverse hyperbolic tangent function.
inline RpcObject* RpcDouble::aTanh( void )
{
    RpcObject* temp = new RpcDouble( atanh(data) );
    return( temp );
}


// Euler Gamma function.
// Note: gamma() is really LogGamma[].  Who the hell chose that
// convention?
inline RpcObject* RpcDouble::Gamma( void )
{
    RpcObject* temp = new RpcDouble( exp(gamma(data)) );
    return( temp );
}


// Log gamma function.
inline RpcObject* RpcDouble::LogGamma( void )
{
    RpcObject* temp = new RpcDouble( gamma(data) );
    return( temp );
}


// Factorial function.
inline RpcObject* RpcDouble::Factorial( void )
{
    RpcObject* temp = new RpcDouble( exp(gamma(data+1.0)) );
    return( temp );
}


// Error function.
inline RpcObject* RpcDouble::Erf( void )
{
    RpcObject* temp = new RpcDouble( erf(data) );
    return( temp );
}


// Complementary error function.
inline RpcObject* RpcDouble::Erfc( void )
{
    RpcObject* temp = new RpcDouble( erfc(data) );
    return( temp );
}


// Real part of number.
inline RpcObject* RpcDouble::Real( void )
{
    // Returns same number.
    RpcObject* temp = new RpcDouble( data );
    return( temp );
}


// Imaginary part of number.
inline RpcObject* RpcDouble::Imag( void )
{   // Returns zero-valued double.
    RpcObject* temp = new RpcDouble( 0.0 );
    return( temp );
}


// Floor function.
inline RpcObject* RpcDouble::Floor( void )
{
    RpcObject* temp = new RpcDouble( floor(data) );
    return( temp );
}


// Ceiling function.
inline RpcObject* RpcDouble::Ceil( void )
{
    RpcObject* temp = new RpcDouble( ceil(data) );
    return( temp );
}


// Nearest integer function.
inline RpcObject* RpcDouble::NearInt( void )
{
    RpcObject* temp = new RpcDouble( floor(data+0.5) );
    return( temp );
}


// Inverse function.
inline RpcObject* RpcDouble::Inv( void )
{
    RpcObject* temp = new RpcDouble( 1.0/data );
    return( temp );
}


// Absolute value
inline RpcObject* RpcDouble::Abs( void )
{
    RpcObject* temp;
    if( data >= 0 )
    {
        temp = new RpcDouble( data );
    }
    else
    {
        temp = new RpcDouble( -data );
    }
    return( temp );
}



// Convert to integer constant.
inline RpcObject* RpcDouble::ToConst( void )
{
    RpcObject* temp = new RpcIntegerConst( data );
    return( temp );
}


// Return the data.
inline double RpcDouble::getData( void )
{
    return( data );
}


// arch-tag: rpc double precision real data type
