/*  rpc -- A console-based reverse polish notation calculator
    Copyright (C) 2000, 2001, 2002 Paul Pelzl

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



/*  RpcIntegerConst.cpp
 *  Definition of a the rpc integer constant data type.
 */
 
#include "RpcIntegerConst.h"



// Default constructor: creates a zero element.
RpcIntegerConst::RpcIntegerConst()
{
    type = INTEGERCONST_TYPE;
    data = IntegerConst( 0UL );
}


// Default destructor.
RpcIntegerConst::~RpcIntegerConst()
{
}


// Constructor to initialize specific value.
RpcIntegerConst::RpcIntegerConst( IntegerConst input )
{
    type = INTEGERCONST_TYPE;
    data = input;
}


// Constructor to initialize specific value.
RpcIntegerConst::RpcIntegerConst( double input )
{
    type = INTEGERCONST_TYPE;
    data = IntegerConst( input );
}


// Copy constructor.
RpcIntegerConst::RpcIntegerConst( const RpcIntegerConst& oldObject )
{
    type = INTEGERCONST_TYPE;
    data = oldObject.data;
}


// Assignment operator.
RpcIntegerConst& RpcIntegerConst::operator =( const RpcIntegerConst& oldObject )
{ 
    if( this != &oldObject)
    {
        type = INTEGERCONST_TYPE;
        data = oldObject.data;
    }
    
    return( *this );
}


// Return the display string for this object.
char* RpcIntegerConst::getDisplayString( RpcCalcFlags flags ) const
{
    char* temp = new char[256];
    char scratch[256];
    
    // translate the constant to a formatted string, in dec/hex/oct/bin format.
    if( flags.constMode == MODE_DEC )
    {
        sprintf( temp, "# %lud", data.longdata );
    }
    else if( flags.constMode == MODE_HEX )
    {
        sprintf( temp, "# %lXh", data.longdata );
    }
    else if( flags.constMode == MODE_OCT )
    {
        sprintf( temp, "# %loo", data.longdata );
    }
    else
    {
        toBinary( scratch );
        sprintf( temp, "# %sb", scratch );
    }
        
    return( temp );
}    


// Return the fullscreen display string for this object.
char* RpcIntegerConst::getFullscreenString( RpcCalcFlags flags ) const
{
    char* temp = new char[256];
    char scratch[256];
    
    // translate the constant to a formatted string, in dec/hex/oct/bin format.
    if( flags.constMode == MODE_DEC )
    {
        sprintf( temp, "# %lud", data.longdata );
    }
    else if( flags.constMode == MODE_HEX )
    {
        sprintf( temp, "# %lXh", data.longdata );
    }
    else if( flags.constMode == MODE_OCT )
    {
        sprintf( temp, "# %loo", data.longdata );
    }
    else
    {
        toBinary( scratch );
        sprintf( temp, "# %sb", scratch );
    }
        
    return( temp );   
}
 

// Addition operator.
RpcObject* RpcIntegerConst::operator +( const RpcObject* rhs )
{
    RpcObject* temp;

    // Branch through all the possible data types.
    switch( rhs -> getType() )
    {
        // Add a double to an integer constant
        case DOUBLE_TYPE:
            temp = new RpcIntegerConst( data + ((RpcDouble*)rhs)->getData() );
            // trap overflow error
            if( (((RpcIntegerConst*)temp)->getData()).overflow )
            {
                delete temp;
                temp = new RpcError(ERRORCODE_CONST_OVERFLOW);
            }                
            break;

        // Add an integer constant to an integer constant.
        case INTEGERCONST_TYPE:
            temp = new RpcIntegerConst( data + ((RpcIntegerConst*)rhs)->getData() );
            // trap overflow error
            if( (((RpcIntegerConst*)temp)->getData()).overflow )
            {
                delete temp;
                temp = new RpcError(ERRORCODE_CONST_OVERFLOW);
            }
            break;
                        
        // Default is an error.    
        default:
            temp = new RpcError(ERRORCODE_BADARG);
            break;
    }
    
    return( temp );
}


// Subtraction operator.
RpcObject* RpcIntegerConst::operator -( const RpcObject* rhs )
{
    RpcObject* temp;
    
    // Branch through all possible data types.
    switch( rhs -> getType() )
    {
        // Subtract a double from an integer const
        case DOUBLE_TYPE:
            temp = new RpcIntegerConst( data - ((RpcDouble*)rhs)->getData() );
            break;
            
        // Subtract an integer const from an integer const
        case INTEGERCONST_TYPE:
            temp = new RpcIntegerConst( data - ((RpcIntegerConst*)rhs)->getData() );
            break;
            
        // Default is an error.
        default:
            temp = new RpcError(ERRORCODE_BADARG);
            break;
    }
    
    return( temp );
}


// Multiplication operator.
RpcObject* RpcIntegerConst::operator *( const RpcObject* rhs )
{
    RpcObject* temp;
    
    // Branch through all possible data types.
    switch( rhs -> getType() )
    {
        // Multiply an integer constant by a double.
        case DOUBLE_TYPE:
            temp = new RpcIntegerConst( data * ((RpcDouble*)rhs)->getData() );
            // trap overflow error
            if( (((RpcIntegerConst*)temp)->getData()).overflow )
            {
                delete temp;
                temp = new RpcError(ERRORCODE_CONST_OVERFLOW);
            }
            break;

        // Multiply an integer constant by an integer constant.            
        case INTEGERCONST_TYPE:
            temp = new RpcIntegerConst( data * ((RpcIntegerConst*)rhs)->getData() );
            // temp overflow error
            if( (((RpcIntegerConst*)temp)->getData()).overflow )
            {
                delete temp;
                temp = new RpcError(ERRORCODE_CONST_OVERFLOW);
            }
            break;
            
        // Default is an error.
        default:
            temp = new RpcError(ERRORCODE_BADARG);
            break;
    }
    
    return( temp );
}


// Division operator.
RpcObject* RpcIntegerConst::operator /( const RpcObject* rhs )
{
    RpcObject* temp;
    
    // Branch through all possible data types.
    switch( rhs -> getType() )
    {
        // Divide an integer const by a double.
        case DOUBLE_TYPE:
            temp = new RpcIntegerConst( data / ((RpcDouble*)rhs)->getData() );
            // trap overflow
            if( (((RpcIntegerConst*)temp)->getData()).overflow )
            {
                delete temp;
                temp = new RpcError(ERRORCODE_CONST_OVERFLOW);
            }
            break;
            
        // Divide an integer const by an integer const.
        case INTEGERCONST_TYPE:
            temp = new RpcIntegerConst( data / ((RpcIntegerConst*)rhs)->getData() );
            // trap overflow
            if( (((RpcIntegerConst*)temp)->getData()).overflow )
            {
                delete temp;
                temp = new RpcError(ERRORCODE_CONST_OVERFLOW);
            }
            break;
            
        // Default is an error.
        default:
            temp = new RpcError(ERRORCODE_BADARG);
            break;
    }
    
    return( temp );
}



// Mod operator.
RpcObject* RpcIntegerConst::Mod( const RpcObject* rhs )
{
    RpcObject* temp;
    
    // Branch through all possible data types.
    switch( rhs -> getType() )
    {
        // Evaluate an integer constant mod another integer const.
        case INTEGERCONST_TYPE:
            temp = new RpcIntegerConst( data % ((RpcIntegerConst*)rhs)->getData() );
            break;
            
        // Default is an error.
        default:
            temp = new RpcError(ERRORCODE_BADARG);
            break;
    }
    
    return( temp );
}



// Convert to RpcDouble.
RpcObject* RpcIntegerConst::Real( void )
{
    RpcObject* temp = new RpcDouble( (double) data.longdata );
    return( temp );
}
    

void RpcIntegerConst::toBinary( char* destString ) const
{
    unsigned long temp = data.longdata;
    char digit[10];
    char scratch[50];
    int i=0, len;
    
    strcpy(scratch, "");
    
    do
    {
        sprintf( digit, "%lu", temp%2UL );        
        strcat( scratch, digit );
        temp /= 2UL;
    } while (temp != 0UL);
    
    len = strlen(scratch);
    
    for(i=0; i<len; i++)
    {
        destString[i] = scratch[len-i-1];
    }
    
    destString[len] = '\0';
}

    
    



// Return the data.
inline IntegerConst RpcIntegerConst::getData( void )
{
    return( data );
}



// arch-tag: rpc integer constant data type
