/*  rpc -- A console-based reverse polish notation calculator
    Copyright (C) 2000, 2001, 2002 Paul Pelzl
    
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


/* RpcInterface.cpp
 * Definition of the class that takes basic input and produces the
 * standard curses stack display.
 */

#include "RpcInterface.h"


// Defining all the possible special function strings.
// Letters entered into the buffer must match with these
// strings in order to select a special function.
// Order is important here--the first string which
// matches the letters entered will be chosen.
// For example, "erf" is chosen before "erfc"
// even though both match the entry string "erf".
// The order should be the same as defined in the
// macros at the beginning of RpcInterface.h.

// By default, these are the strings that will be
// displayed upon special function entry.  If something
// else should be displayed, that must be substituted
// in parseFunBuffer().  See how "<radians mode>" is 
// implemented, for example.
const static char* functStrings[MAXN] = { "sin", "cos", "tan", "sq", "sqrt",
                                          "inv", "exp", "ten^", "ln", "log", 
                                          "asin", "acos", "atan", "rad", "deg", 
                                          "pi", "undo", "floor", "ceil", "nearint",
                                          "gamma", "lngamma", "erf", "erfc",
                                          "sinh", "cosh", "tanh", "asinh",
                                          "acosh", "atanh", "real", "imag",
                                          "view", "transpose", "const", "dec",
                                          "hex", "bin", "oct", "abs", "arg",
                                          "rect", "polar" };

const static char* errorStrings[5] = { "", "<bad argument>",
                                       "<insufficient arguments>",
                                       "<singular matrix>",
                                       "<integer constant overflow>"};
                                       
                                

//  Default constructor, with initializations.
RpcInterface::RpcInterface()
{    
    
    initscr(); 			    // Initialize ncurses
    keypad(stdscr, TRUE);	// Allow standard mapping for arrow keys, etc.
    cbreak(); noecho();

        
    // Check if the fullscreen editor has already been chosen.  If not,
    // pick one.
    strcpy( rpcEditorFile, getenv( "HOME" ) );
    strcat( rpcEditorFile, "/.rpc/editor" );
    FILE* editorFile;
    editorFile = fopen( rpcEditorFile, "r" );
    if( editorFile != NULL )
    {
        fscanf( editorFile, "%s", rpcEditor );
    }
    else
    {
        move(5, 0);
        printw( "What editor do you want to use for fullscreen viewing?  </bin/view>\n" );
        refresh();
        echo();
        getnstr( rpcEditor, 255 );
        noecho();
        if( strlen(rpcEditor) == 0 )
        {
            strcpy(rpcEditor, "/bin/view");
        }
        // Create the .rpc directory
        system("mkdir $HOME/.rpc");
        editorFile = fopen( rpcEditorFile, "w" );
        fprintf( editorFile, "%s", rpcEditor );
    }
    fclose( editorFile );
    
    strcpy( viewCommand, rpcEditor );
    strcat( viewCommand, " ~/.rpc/viewbuffer" );
    
                
    clearEntry();               // Clear the entry buffers.
            
    clearFunctionEntry();	// Clear function entry buffer.
    strcpy(functionDisplay, "");	// Clear function display buffer.
    funExists = false;		// No function input initially
    browsing  = false;		// Not in interactive stack mode

    strcpy( rpcKeyMapFile, getenv( "HOME" ) );
    strcat( rpcKeyMapFile, "/.rpc/keymap" );
    loadUserKeyMap();           // Load the user keymapping file.

    strcpy( rpcViewBufferFile, getenv( "HOME" ) );
    strcat( rpcViewBufferFile, "/.rpc/viewbuffer" );
        
    // Other non-essential initializations
    key       = 0;
    funChoice = NEMPTY;
    funArg    = false;
    polar     = false;
    dataType  = DOUBLE_TYPE;
    
    stackLevel = 1;
    screenBottom = 1;
    horizShift = 0;
    
    buffersUsed = 0;            // Not using any entry buffers initially
    currentBuffer = 0;          // Editing the 0th buffer.
    errorState = 0;             // No error initially.
}


//  Destructor.
RpcInterface::~RpcInterface()
{
    endwin();            // Clean exit ncurses
    
    cout << endl;        // Clean up the display
}


//  Clear the entry buffers.
void RpcInterface::clearEntry( void )
{
    int i;
    for(i=0; i<1024; i++)
    {
        strcpy( entry[i], "" ); 
        strcpy( expEntry[i], " " );     // One space for sign
        expExists[i] = false;
        strcpy( entrySign[i], " " );             // No sign
    }
    entryExists = false;
    
    currentBuffer = 0;
}


//  REALLY clear functionEntry (set to an array of NULL chars).
//  Fixes some problems with using strcpy().
void RpcInterface::clearFunctionEntry( void )
{
    int i;
    for( i=0; i<10; i++ )
        functionEntry[i] = '\0';
}


//  Display entry buffer.
void RpcInterface::displayEntry( void )
{
    int idx = 0;
    int jmax = 0;
    char temp[20480];
    strcpy( temp, "" );
    parseFunBuffer();                // Figure out if a function is selected	

    
    // Display the function name, e.g. "sin( "
    if( funExists )
    {
        strcat( temp, functionDisplay );    // Display the function string
        if( funArg )
        {
           strcat( temp, "( " );
        }
    }

    
    // Display the function argument, e.g. "1 x 10^-6"
    if( !funExists || funArg )
    {    
        switch( dataType )
        {
            case COMPLEX_TYPE:
                strcat( temp, "(" );
                // REAL PART
                if( entrySign[0][0] == '-' )
                    strcat( temp, entrySign[0] );
                strcat( temp, entry[0] ); // Display mantissa

                if( expExists[0] )
                {
                    strcat( temp, " x10^" );  // Display exponent
                    strcat( temp, expEntry[0] );
                }

                if( currentBuffer == 1 )
                {
                    if(polar)
                    {
                        strcat( temp, " <" );
                    }
                    else
                    {
                        strcat( temp, ", " );
                    }
                    // IMAG PART
                    if( entrySign[1][0] == '-' )
                        strcat( temp, entrySign[1] );
                    strcat( temp, entry[1] ); // Display mantissa

                    if( expExists[1] )
                    {
                        strcat( temp, " x10^" );  // Display exponent
                        strcat( temp, expEntry[1] );
                    }
                }
                strcat( temp, ")" );
                break;

            case REALMATRIX_TYPE:
                strcat( temp, " [" );
                for(int i=0; i<mtrxRows; i++)
                {
                    strcat( temp, "[" );
                    if( (i+1)*mtrxCols <= ((currentBuffer+2)/2) )
                    {
                        jmax = mtrxCols;
                    }
                    else
                    {   
                        jmax = ((currentBuffer+2)/2) % mtrxCols;
                    } 
                    for(int j=0; j<jmax; j++)
                    {
                        idx = i*mtrxCols + j;
                        if( entrySign[2*idx][0] == '-' )
                            strcat( temp, entrySign[2*idx] );
                        strcat( temp, entry[2*idx] ); // Display mantissa

                        if( expExists[2*idx] )
                        {
                            strcat( temp, " x10^" );  // Display exponent
                            strcat( temp, expEntry[2*idx] );
                        }
                        
                        if( j < jmax-1 )
                        {
                            strcat( temp, ", " );
                        }
                    }

                    strcat( temp, "]" );
                }
                strcat( temp, "]" ); 
                break;

            case COMPLEXMATRIX_TYPE:
                strcat( temp, " [" );
                for(int i=0; i<mtrxRows; i++)
                {
                    strcat( temp, "[" );
                    if( (i+1)*mtrxCols <= ((currentBuffer+2)/2) )
                    {
                        jmax = mtrxCols;
                    }
                    else
                    {
                        jmax = ((currentBuffer+2)/2) % mtrxCols;
                    }
                    for(int j=0; j<jmax; j++)
                    {                        
                        idx = i*mtrxCols + j;
                        strcat( temp, "(" );
                        
                        // Display real part
                        if( entrySign[2*idx][0] == '-' )
                            strcat( temp, entrySign[2*idx] );
                        strcat( temp, entry[2*idx] );   // Display mantissa
                        
                        if( expExists[2*idx] )
                        {
                            strcat( temp, " x10^" );  // Display exponent
                            strcat( temp, expEntry[2*idx] );
                        }
                        
                        if( !(currentBuffer == 2*idx) )
                        {
                            if(polar)
                            {
                                strcat( temp, " <" );
                            }
                            else
                            {
                                strcat( temp, ", " );
                            }

                            // Display imaginary part
                            if( entrySign[2*idx+1][0] == '-' )
                                strcat( temp, entrySign[2*idx+1] );
                            strcat( temp, entry[2*idx+1] );   // Display mantissa

                            if( expExists[2*idx+1] )
                            {
                                strcat( temp, " x10^" );    // Display exponent
                                strcat( temp, expEntry[2*idx+1] );
                            }
                        }
                        
                        strcat( temp, ")" );
                        
                        if( j < jmax-1 )
                        {
                            strcat( temp, ", " );
                        }
                    }                    

                    strcat( temp, "]" );
                }
                strcat( temp, "]" );
                break;                                                             
               
                
            default:
                // Assume entry of double type.
                if( entrySign[0][0] == '-' )
                    strcat( temp, entrySign[0] );
                strcat( temp, entry[0] ); // Display mantissa

                if( expExists[0] )
                {
                    strcat( temp, " x10^" ); // Display exponent
                    strcat( temp, expEntry[0] );
                }
                break;
        }    
    }
    
    if( funExists && funArg )
    {
        strcat( temp, " )" );
    }
    

    //  Display the buffer via curses.
    move(LINES-1, 0);
    clrtoeol();

    if( strlen(temp) <= SWIDTH )
    {
        move(LINES-1, SWIDTH-strlen(temp));
        printw( temp );
    }
    else
    {
        printw( temp+strlen(temp)-SWIDTH );
    }

    refresh();
}




//  Create the standard stack display.
void RpcInterface::displayMain( RpcCalc& calc )
{
    char angleString[4];
    char baseString[4];
    char complexString[4];
    
    char tempstring[33];
    char displayString[DISP_SIZE];
    int shiftBit = 0;
    
    erase();   

    if( calc.flags.angleMode == MODE_RAD )
        strcpy( angleString, "RAD" );
    else
        strcpy( angleString, "DEG" );
    
    if( calc.flags.constMode == MODE_DEC )
        strcpy( baseString, "DEC" );
    else if( calc.flags.constMode == MODE_HEX )
        strcpy( baseString, "HEX" );
    else if( calc.flags.constMode == MODE_BIN )
        strcpy( baseString, "BIN" );
    else
        strcpy( baseString, "OCT" );

    if( calc.flags.complexMode == MODE_RECT )
        strcpy( complexString, "REC" );
    else
        strcpy( complexString, "POL" );
    
    if( errorState != 0 )
        shiftBit = 2;


    int Cnt;
    move(0, 0);
    attron(A_BOLD);
    printw("rpc v%s              Press '?' for help\n", VERSION);
    printw("angle: ");
    attroff(A_BOLD);
    printw("%s     ", angleString);
    attron(A_BOLD);
    printw("base: ");
    attroff(A_BOLD);
    printw("%s     ", baseString);
    attron(A_BOLD);
    printw("complex: ");
    attroff(A_BOLD);
    printw("%s\n", complexString);
    attron(A_BOLD);
    printw("------------------------------------------\n");
    attroff(A_BOLD);

    if( errorState != 0 )
    {
        attron(A_BOLD);
        printw("Error: %s\n", errorStrings[errorState]);
        printw("------------------------------------------\n");
        attroff(A_BOLD);
    }

    //  Print contents of stack in reverse order.
    for( Cnt=LINES-5-shiftBit; Cnt>0; Cnt-- )
    {
        if( Cnt>calc.numElements() )
            printw( "%2d:\n", Cnt );
        else
        {
            calc.getDisplayString(displayString, Cnt);
            if( strlen(displayString) > 29 )
            {
                snprintf( tempstring, 27, "%s", displayString );
                strcat( tempstring, "..." );
            }
            else
            {
                strcpy( tempstring, displayString );
            }
            printw( "%2d:     %30s\n", Cnt, tempstring );
        }
    }
    attron(A_BOLD);
    printw("------------------------------------------");    
    attroff(A_BOLD);
    move(LINES-1, SWIDTH);
    refresh();                   // Update the modified screen
}



//  Translate the entry buffers into a valid RpcObject.
RpcObject* RpcInterface::getEntry( RpcCalc& calc )
{
    RpcObject* temp = NULL;
    double* matrixElements;
    double a, b;
    Complex* cMatrixElements;
    int i;

    switch( dataType )
    {
        case DOUBLE_TYPE:
            // Use the first entry buffer
            temp = new RpcDouble( parseBuffer(0) );
            break;
            
        case COMPLEX_TYPE:
            // Use the first two entry buffers
            temp = new RpcComplex( Complex(polar, calc.flags.angleMode, 
                   parseBuffer(0), parseBuffer(1)) );
            break;
            
        case REALMATRIX_TYPE:
            // Use the first 2(mxn) entry buffers
            matrixElements = new double[mtrxRows*mtrxCols];
            for(i=0; i<mtrxRows*mtrxCols; i++)
            {
                // Note: for a real matrix, every other buffer is actually
                //       used.  So we use parseBuffer(2*i) instead of
                //       parseBuffer(i).
                matrixElements[i] = parseBuffer(2*i);
            }
            
            temp = new RpcRealMatrix( matrixElements, mtrxRows, mtrxCols );                                  
            delete[] matrixElements;                     
            break;
           
        case COMPLEXMATRIX_TYPE:
            refresh();
            // Use the first 2(mxn) entry buffers
            cMatrixElements = new Complex[mtrxRows*mtrxCols];
            for(i=0; i<mtrxRows*mtrxCols; i++)
            {
                a = parseBuffer(2*i);      // get the real part (or magnitude)
                b = parseBuffer(2*i+1);    // get the imag part (or phase)
                cMatrixElements[i] = Complex(polar, calc.flags.angleMode, a, b);
            }
            
            temp = new RpcComplexMatrix( cMatrixElements, mtrxRows, mtrxCols );
            delete[] cMatrixElements;
            refresh();
            break;
     
        default:
            // Return an error if the data type is not defined
            temp = new RpcError;
            break;
    }
    
    clearEntry();        // Kill the data in the entry buffers
    
    dataType = DOUBLE_TYPE;    
    polar = false;
    return( temp );
}


//  Display the browsing screen.
void RpcInterface::displayBrowse( RpcCalc& calc )
{
    
    int screenShift;
    char tempstring[33];   
    char displayString[DISP_SIZE];
    char levelString[10];
    char levelString2[10];
    int maxLen, numSpaces, i;
    
    // Shift the screen down when user keymapping is used,
    // because the little help screen at the bottom is reduced.
    if( userKeyMap )
        screenShift = 2;
    else
        screenShift = 0;


    // Make sure screen is properly aligned (during scrolling)
    if( stackLevel > LINES + screenBottom - 7 + screenShift )
        screenBottom = stackLevel - LINES + 7 - screenShift;
    else if( stackLevel < screenBottom )
        screenBottom = stackLevel;


    erase();

    int Cnt;
    move(0, 0);
    attron(A_BOLD);
    printw("rpc v%s       --       (STACK BROWSING MODE)\n", VERSION );
    printw("-----------------------------------------------\n");
    attroff(A_BOLD);

    // Figure out how many spaces to use for the level number.
    sprintf( levelString, "%d", calc.numElements() );
    sprintf( levelString2, "%d", LINES+screenBottom-7 + screenShift );
    if( strlen(levelString) > strlen(levelString2) )
        maxLen = strlen(levelString);
    else
        maxLen = strlen(levelString2);

    // Print contents of stack in reverse order, down to screen bottom.
    for( Cnt=LINES+screenBottom-7 + screenShift; Cnt>screenBottom-1; Cnt-- )
    {
        // Generate the level number.
        sprintf( levelString2, "%d", Cnt );
        numSpaces = maxLen - strlen( levelString2 );
        
        // Put the appropriate number of spaces before the number...
        for(i=0; i<numSpaces; i++)
        {
            levelString[i] = ' ';
        }
        levelString[numSpaces] = '\0';
        
        // Print the level number itself...
        strcat( levelString, levelString2 );
        strcat( levelString, ":" );
        
        // Put the appropriate number of spaces after the number.
        for(i=strlen(levelString); i<8; i++)
        {
            levelString[i] = ' ';
        }
        levelString[8] = '\0';
        
        
        // Generate the display.
        if( Cnt>calc.numElements() )
            printw( "%s\n", levelString );
        else
        {
            calc.getDisplayString( displayString, Cnt );        
            if( strlen( displayString ) > 29 )
            {
                snprintf( tempstring, 27, "%s", displayString );
                strcat( tempstring, "..." );
            }
            else
            {
                strcpy( tempstring, displayString );
            }
        
            if( Cnt != stackLevel )
            {
                printw( "%s%30s\n", levelString, tempstring );
            }
            else
            {
                // Adjust for horizontal scrolling
                if( horizShift > strlen(displayString)-29 )
                {
                    horizShift = strlen(displayString)-29;
                }
                
                if( strlen( displayString ) > 29 )
                {
                    for(int k=0; k<29; k++)
                    {
                        tempstring[k] = displayString[k+horizShift];
                    }
                    tempstring[29] = '\0';
                }                    
                
                attron(A_REVERSE);        // "Highlight" with reverse video
                printw( "%s%30s\n", levelString, tempstring );
                attroff(A_REVERSE);
            }
        }
    }
    attron(A_BOLD);
    printw("-----------------------------------------------\n");
    attroff(A_BOLD);

    // If user keymapping is on, don't use the mini help screen.
    if( userKeyMap)
    {
        printw( "            Press 'q' to quit browsing." );
    }
    else  // ...otherwise, print this help screen.
    {
        printw("'v' to view; '\\' to delete; 'd' to dropn;\n");
        printw("'k' to keepn; 'r' to roll up; 'R' to roll down;\n");
        printw("'u' to undo all changes; 'q' to quit browsing.");
    }

    move( LINES-1, SWIDTH2 );
    refresh();
}


//  Browse the stack.
void RpcInterface::browse( RpcCalc& calc )
{

    calc.backup();              // Back up before entering interactive mode.
   
    stackLevel   = 1;           // Leave cursor at level 1 initially.
   
    screenBottom = 1;           // This variable stores the lowest level displayed
                                // in stack browsing mode.  Used to figure out how
                                // to shift the screen during scrolling.

    displayBrowse( calc );            // Show the "stack browsing" display

    // Main browsing loop.
    int endloop = 1;
   
    while( endloop )
    {
        key = getch();

        if( !userKeyMap )
        {
            //  IF USING DEFAULT KEYMAP, GO HERE
            switch( key )
            {
                case KEY_UP:        // Move up the stack (up arrow)
                    horizShift = 0;
                    if( calc.numElements() > stackLevel )
                        stackLevel++;
                    displayBrowse( calc );
                    break;

                case KEY_DOWN:      // Move down the stack (down arrow)
                    horizShift = 0;
                    if( stackLevel > 1 )
                        stackLevel--;
                    displayBrowse( calc );
                    break;

                case KEY_PPAGE:     // Page up through stack
                    horizShift = 0;
                    if( calc.numElements()-screenBottom > LINES+LINES-12 )
                    {
                        stackLevel += LINES-6;
                        screenBottom += LINES-6;
                    }
                    else
                        stackLevel = calc.numElements();
                    displayBrowse( calc );
                    break;

                case KEY_NPAGE:        // Page down through stack
                    horizShift = 0;
                    if( screenBottom > LINES-6 )
                    {
                        stackLevel -= LINES-6;
                        screenBottom -= LINES-6;
                    }
                    else
                        stackLevel = 1;
                    displayBrowse( calc );
                    break;

                case KEY_RIGHT:     // Scroll right through a stack element
                    horizShift++;
                    displayBrowse( calc );
                    break;
                    
                case KEY_LEFT:      // Scroll left through a stack element
                    if( horizShift > 0 )
                    {
                        horizShift--;
                        displayBrowse( calc );
                    }
                    break;
                    
                case '\012':        // CR echoes the selected item to stack level 1.
                case KEY_ENTER:     // ncurses enter mapping does same thing.
                    horizShift = 0;
                    calc.enter( calc.viewElement(stackLevel) );
                    stackLevel++;            // Move one item up the stack, since an item was
                                             // pushed to level 1.
                    displayBrowse( calc );
                    break;

                case '\\':          // '\' drops the selected item.
                    horizShift = 0;
                    if( calc.numElements() >= stackLevel )
                    {
                        calc.drop( stackLevel );
                        if( (stackLevel > calc.numElements()) &&
                                (calc.numElements() > 0) )
                        {
                            stackLevel = calc.numElements();
                        }
                    }
                    if( calc.numElements() == 0 )
                        endloop = 0;
                    displayBrowse( calc );
                    break;

                case 'd':           // Drop first n stack items
                    horizShift = 0;
                    if( calc.numElements() >= stackLevel )
                    {
                        calc.dropn( stackLevel );
                        stackLevel = 1;
                    }
                    if( calc.numElements() == 0 )
                        endloop = 0;
                    displayBrowse( calc );
                    break;

                case 'k':           // Keep first n stack items
                    horizShift = 0;
                    calc.keepn( stackLevel );
                    displayBrowse( calc );
                    break;

                case 'r':           // Cyclically roll stack items up to level n
                    horizShift = 0;
                    calc.rollUp( stackLevel );
                    displayBrowse( calc );
                    break;

                case 'R':           // Cyclically roll stack items down to level n
                    horizShift = 0;
                    calc.rollDown( stackLevel );
                    displayBrowse( calc );
                    break;

                case 12:            // Refresh display (^L, like vi mapping)
                case 18:            // (^R also works)
                    windowRefresh( calc );
                    break;
                    
                case 'q':           // Exit stack browsing mode
                    horizShift = 0;
                    endloop = 0;
                    break;

                case 'u':           // Undo all changes
                    horizShift = 0;
                    calc.undo();
                    calc.backup();  // Prevents strange "double undo" effects
                    stackLevel = 1;
                    displayBrowse( calc );
                    break;
                 
                case 'v':           // View a stack element
                    horizShift = 0;
                    // Write the fullscreen data to a buffer file
                    FILE *bufferFile;
                    bufferFile = fopen( rpcViewBufferFile, "w" );
                    char fullscreenString[FULLSCREEN_SIZE];
                    calc.getFullscreenString( fullscreenString, stackLevel );
                    fprintf( bufferFile, fullscreenString );
                    fclose( bufferFile );               

                    // Call 'vi' to look at the file
                    endwin();        // exit ncurses so vi can take over
                    system( viewCommand );   // View fullscreen buffer
                    initscr(); 			// Initialize ncurses
                    keypad(stdscr, TRUE);	// Allow standard mapping for arrow keys, etc.
                    cbreak(); noecho();
                    displayBrowse( calc );
                    break;                                  
                    
            }


        }
        else
        {

            //  IF USING USER-DEFINED KEYMAP, GO HERE
            if( key == moveUpChar )        // Move up the stack
            {
                horizShift = 0;
                if( calc.numElements() > stackLevel )
                    stackLevel++;
                displayBrowse( calc );
            }
            else if( key == moveDownChar ) // Move down the stack
            {
                horizShift = 0;
                if( stackLevel > 1 )
                    stackLevel--;
                displayBrowse( calc );
            }
            else if( key == pageUpChar )   // Page up the stack
            {
                horizShift = 0;
                if( calc.numElements()-screenBottom > LINES+LINES-12 )
                {
                    stackLevel += LINES-6;
                    screenBottom += LINES-6;
                }
                else
                    stackLevel = calc.numElements();
                displayBrowse( calc );
            }
            else if( key == pageDownChar ) // Page down the stack
            {
                horizShift = 0;
                if( screenBottom > LINES-6 )
                {
                    stackLevel -= LINES-6;
                    screenBottom -= LINES-6;
                }
                else
                {
                    stackLevel = 1;
                }
                displayBrowse( calc );
            }
            else if( key == scrollRightChar )
            {
                horizShift++;
                displayBrowse( calc );
            }
            else if( key == scrollLeftChar )
            {                                    
                if( horizShift > 0 )
                {
                    horizShift--;
                    displayBrowse( calc );
                }
            }                                              
            else if( key == enterChar )    // Echo selected item to level 1.            
            {
                horizShift = 0;
                calc.enter( calc.viewElement(stackLevel) );
                stackLevel++;            // Move one item up the stack, since an item was
                                         // pushed to level 1.
                displayBrowse( calc );
            }
            else if( key == dropChar )     // Drop the selected item.
            {
                horizShift = 0;
                if( calc.numElements() >= stackLevel )
                {
                    calc.drop( stackLevel );
                    if( (stackLevel > calc.numElements()) &&
                            (calc.numElements() > 0) )
                    {
                        stackLevel = calc.numElements();
                    }
                }
                if( calc.numElements() == 0 )
                    endloop = 0;
                displayBrowse( calc );
            }
            else if( key == dropnChar )    // Drop first n stack items
            {
                horizShift = 0;
                if( calc.numElements() >= stackLevel )
                {
                    calc.dropn( stackLevel );
                    stackLevel = 1;
                }
                if( calc.numElements() == 0 )
                    endloop = 0;
                displayBrowse( calc );
            }
            else if( key == keepnChar )    // Keep first n stack items
            {
                horizShift = 0;
                calc.keepn( stackLevel );
                displayBrowse( calc );
            }
            else if( key == rollUpChar )   // Cyclically roll stack items up
            {
                horizShift = 0;
                calc.rollUp( stackLevel );
                displayBrowse( calc );
            }
            else if( key == rollDownChar ) // Cyclically roll stack items down
            {
                horizShift = 0;
                calc.rollDown( stackLevel );
                displayBrowse( calc );
            }
            else if( key == refreshChar )  // Refresh display
            {
                windowRefresh( calc );
            }
            else if( key == 'q' )          // Exit stack browsing mode
            {
                horizShift = 0;
                endloop = 0;
            }
            else if( key == undoChar )     // Undo all changes
            {
                horizShift = 0;
                calc.undo();
                calc.backup();  // Prevents strange "double undo" effects
                stackLevel = 1;
                displayBrowse( calc );
            }
            else if(key == viewChar )
            {
                horizShift = 0;
                // Write the fullscreen data to a buffer file
                FILE *bufferFile;
                bufferFile = fopen( rpcViewBufferFile, "w" );
                char fullscreenString[FULLSCREEN_SIZE];
                calc.getFullscreenString( fullscreenString, stackLevel );
                fprintf( bufferFile, fullscreenString );
                fclose( bufferFile );               

                // Call 'vi' to look at the file
                endwin();        // exit ncurses so vi can take over
                system( viewCommand );   // Look at fullscreen buffer
                initscr(); 			// Initialize ncurses
                keypad(stdscr, TRUE);	// Allow standard mapping for arrow keys, etc.
                cbreak(); noecho();
                displayBrowse( calc );
            }
        }
    }
}


//  Get a number from the entry buffer.
double RpcInterface::parseBuffer( int bufnum )
{
    int Cnt;
    char temp[25];
    strcpy( temp, "" );
    
    if( bufnum >=0 && bufnum < 1024 )
    {
        if( expEntry[bufnum][0] == ' ' )    // If no minus sign in exponent, then
            for(Cnt=0; Cnt<4; Cnt++)        // "delete" the space in expEntry[0]
                expEntry[bufnum][Cnt] = expEntry[bufnum][Cnt+1];  // ( atof() doesn't like it )

        strcpy( temp, entrySign[bufnum] );
        strcat( temp, entry[bufnum] );      // Concatenate entry and expEntry
        strcat( temp, "e" );                // into one string that atof() can
        strcat( temp, expEntry[bufnum] );   // convert.
    }
    
    return( atof( temp ) );
}


//  Translate function entry buffer into a valid display function.
//  This is done by comparison with a lookup table of possible functions.

//  Note:  This is a total rewrite from the 0.3x parseFunBuffer().
//  The new version offers a more flexible framework that is easily
//  expandable.
void RpcInterface::parseFunBuffer( void )
{

    int i, temp;
    unsigned int functionEntryLen;
    functionEntryLen = strlen( functionEntry );


    if( functionEntryLen > 0 )
    {
        i    = 0;
        temp = 0;
        //  Loop through each of the function strings, and find the first
        //  one that matches the input.
        while( (i<MAXN) && (temp==0) )
        {
            if( strlen( functStrings[i] ) >= functionEntryLen )
            {
                if( strncmp( functStrings[i], functionEntry, functionEntryLen ) == 0 )
                {
                    temp = 1;
                    funChoice = i;
                    strcpy( functionDisplay, functStrings[i] );
                    funExists = true;
                }
            }
            i++;
        }
    }
    
    // Special consideration for functions that do not take an
    // argument, and which look different on display as a result
    funArg = false;                // Assume no argument is required at this point...
    switch( funChoice )
    {
        case NRAD:
            strcpy( functionDisplay, "<radians mode>" );
            break;
            
        case NDEG:
            strcpy( functionDisplay, "<degrees mode>" );
            break;
        
        case NDEC:
            strcpy( functionDisplay, "<decimal constant mode>" );
            break;
            
        case NHEX:
            strcpy( functionDisplay, "<hex constant mode>" );
            break;
            
        case NBIN:
            strcpy( functionDisplay, "<binary constant mode>" );
            break;
            
        case NOCT:
            strcpy( functionDisplay, "<octal constant mode>" );
            break;
        
        case NRECT:
            strcpy( functionDisplay, "<rect complex representation>" );
            break;
            
        case NPOLAR:
            strcpy( functionDisplay, "<polar complex representation>" );
            break;
            
        case NUNDO:
            strcpy( functionDisplay, "<undo>" );
            break;

        case NPI:
            strcpy( functionDisplay, "PI" );
            break;

        case NVIEW:
            strcpy( functionDisplay, "<view>" );
            break;
            
        default:
            funArg = true;        // ...but the default is that an argument IS required.
            break;
    }        
}	    



//  Boolean:  Is the character valid for input to the numeric buffer?
bool RpcInterface::validNumInput( int input )
{
    return( intStrchr( "1234567890.", input ) );
}
    


// Take input for the function buffer.
// Does a lookup to decide what keypresses are allowed,
// depending on what letters are already in the buffer.
// For example, if 's' is in the buffer, 'i' can be entered (sin)
// or 'q' can be entered (sq or sqrt).

// Note: this is a total rewrite from the 0.3x funInput().
// A more flexible, more easily expandable framework is offered
// here.
void RpcInterface::funInput( char key )
{
    char stemp[2], stemp2[20];
    stemp[0] = key;
    stemp[1] = '\0';          // now stemp is a string of one character
    
    strcpy( stemp2, functionEntry );
    strcat( stemp2, stemp );  // append stemp to functionEntry, store in stemp2
    
    int i, j, temp, match, fun1=0, fun2=0;
    unsigned int stemp2len;
    stemp2len = strlen( stemp2 );
    
    // Perform three separate checks.  First find out what function was
    // matched at the last keypress, then see if the new buffer matches
    // anything.  stemp2 must either match a *different* function or 
    // match two separate functions (e.g. erf and erfc) before we allow
    // adding the new character to the entry buffer.  Otherwise the
    // result is a little confusing for the user, because they may type
    // e.g. "exp", but the screen is only updated when they press 'e'--
    // so they think the 'xp' were not entered.
    
    // I know there may be some un-elegant redundant computation here,
    // but it would be too easy to screw something up in the future
    // otherwise.
    if( stemp2len > 1 )
    {	
        i    = 0;
        temp = 0;
        // First check whether the new buffer matches a different function
        // from the old buffer.
        while( (i<MAXN) && (temp==0) )
        {
            if( strlen( functStrings[i] ) >= strlen( functionEntry ) )
            {
                if( strncmp( functStrings[i], functionEntry, strlen( functionEntry ) ) == 0 )
                {
                    fun1 = i;
                    temp = 1;
                }
            }
            i++;
        }


        fun2 = fun1;

        j    = 0;
        temp = 0;
        while( (j<MAXN) && (temp==0) )
        {
            if( strlen( functStrings[j] ) >= stemp2len )
            {
                if( strncmp( functStrings[j], stemp2, stemp2len ) == 0 )
                {
                    fun2 = j;
                    temp = 1;
                }
            }
            j++;
        }


	// Next check if the new buffer matches at least two separate functions.
        match = 0;
        for(i=0; i<MAXN; i++)
            if( strlen( functStrings[i] ) >= stemp2len )
                if( strncmp( functStrings[i], stemp2, stemp2len ) == 0 )
                    match++;


        // If either case holds, allow entry of the next character.
        if( (fun2 != fun1) || (match>1) )
        {
            funExists = true;
            funChoice = fun2;
            strcpy( functionEntry, stemp2 );
            displayEntry();
        }
    }
    else        // special case, only one character in buffer
    {
        j    = 0;
        temp = 0;
        while( (j<MAXN) && (temp==0) )
        {
            if( strlen( functStrings[j] ) >= stemp2len )
            {
                if( strncmp( functStrings[j], stemp2, stemp2len ) == 0 )
                {
                    funExists = true;
                    funChoice = j;
                    strcpy( functionEntry, stemp2 );
                    displayEntry();
                    temp = 1;
                }
            }
            j++;
        }
    }
}




// Calculates the special functions, based on the function buffer
// contents.  Two cases must be considered:  the function takes
// an argument off of the stack, or it does not.
void RpcInterface::specialFunction( RpcCalc& calc )
{

    bool requiresArg = true;

    funExists = false;
    clearFunctionEntry();
    strcpy( functionDisplay, "" );
    
    RpcDouble* temp;
    RpcObject* ent;
      
    
    // This first switch is for the case when the special function
    // requires no argument.
    switch( funChoice )	
    {
        case NRAD:
            calc.flags.angleMode = MODE_RAD;        // Set radian mode
            errorState = ERRORCODE_NOERR;
            displayMain( calc );         // Update the main display	
            displayEntry();              // Update the entry buffer display
            requiresArg = false;
            break;
            
        case NDEG:		
            calc.flags.angleMode = MODE_DEG;        // Set degrees mode
            errorState = ERRORCODE_NOERR;
            displayMain( calc );         // Update the main display
            displayEntry();              // Update the entry buffer display
            requiresArg = false;
            break;  

    	  case NDEC:		
            calc.flags.constMode = MODE_DEC;   // Set decimal mode
            errorState = ERRORCODE_NOERR;
            displayMain( calc );         // Update the main display
            displayEntry();              // Update the entry buffer display
            requiresArg = false;
            break;  

        case NHEX:		
            calc.flags.constMode = MODE_HEX;   // Set hex mode
            errorState = ERRORCODE_NOERR;
            displayMain( calc );         // Update the main display
            displayEntry();              // Update the entry buffer display
            requiresArg = false;
            break;  

    	  case NBIN:		
            calc.flags.constMode = MODE_BIN;   // Set binary mode
            errorState = ERRORCODE_NOERR;
            displayMain( calc );         // Update the main display
            displayEntry();              // Update the entry buffer display
            requiresArg = false;
            break;  

        case NOCT:		
            calc.flags.constMode = MODE_OCT;   // Set octal mode
            errorState = ERRORCODE_NOERR;
            displayMain( calc );         // Update the main display
            displayEntry();              // Update the entry buffer display
            requiresArg = false;
            break;  
            
        case NRECT:
            calc.flags.complexMode = MODE_RECT; // set rect display mode
            errorState = ERRORCODE_NOERR;
            displayMain( calc );      
            displayEntry();
            requiresArg = false;
            break;
            
        case NPOLAR:
            calc.flags.complexMode = MODE_POLAR; // set polar display mode
            errorState = ERRORCODE_NOERR;
            displayMain( calc );
            displayEntry();
            requiresArg = false;
            break;
                                        
        case NPI:
            calc.backup();

            temp = new RpcDouble( M_PI ); 
            calc.enter( temp );             // Put pi on the stack
            delete( temp );

            errorState = ERRORCODE_NOERR;
            displayMain( calc );
            displayEntry();
            requiresArg = false;
            break;
    
        case NUNDO:
            clearEntry();                   // Clear entry buffer
            calc.undo();
            
            errorState = ERRORCODE_NOERR;
            displayMain( calc );
            displayEntry();
            requiresArg = false;
            break;
        
        case NVIEW:         // Fullscreen view of the top stack item
            // Write the fullscreen data to a buffer file
            FILE *bufferFile;
            bufferFile = fopen( rpcViewBufferFile, "w" );
            char fullscreenString[FULLSCREEN_SIZE];
            calc.getFullscreenString( fullscreenString, 1 );
            fprintf( bufferFile, fullscreenString );
            fclose( bufferFile );               
            
            // Call 'vi' to look at the file
            endwin();        // exit ncurses so vi can take over
            system( viewCommand );   // Look at fullscreen buffer
            initscr(); 			// Initialize ncurses
            keypad(stdscr, TRUE);	// Allow standard mapping for arrow keys, etc.
            cbreak(); noecho();
            displayMain( calc );
            displayEntry();
            requiresArg = false;
            break;
          
        default:
            break;
            
    }
    

    if( requiresArg )
    {    
        // If the special function does require an argument, control
        // flow passes here.  Push the buffer entry onto the stack
        // if it exists, then execute the appropriate function.
        calc.backup();
        
        if( entryExists )
        {
            ent = getEntry(calc);
            calc.enter( ent );
            delete ent;
        }
        
        if( calc.numElements() == 0)
            funChoice = NEMPTY;     // If there is no argument available
                                    // for the function, nothing happens.
                                    // See following switch statement.
    
        // Implementation of the basic scientific calc functions
        switch( funChoice )
        {
            case NSIN:
                errorState = calc.sin();
                break;
    
            case NCOS:
                errorState = calc.cos();
                break;
    
            case NTAN:
                errorState = calc.tan();
                break;
    
            case NEXP:
                errorState = calc.exp();
                break;
                
            case NINV:
                errorState = calc.inv();
                break;
    
            case NLN:
                errorState = calc.log();
                break;
        
            case NLOG:
                errorState = calc.log10();
                break;
        
            case NASIN:
                errorState = calc.asin();
                break;
        
            case NACOS:
                errorState = calc.acos();
                break;
    
            case NATAN:
                errorState = calc.atan();
                break;
        
            case NSQ:
                errorState = calc.sq();
                break;
    
            case NSQRT:
                errorState = calc.sqrt();
                break;
        
            case NTEN:
                errorState = calc.ten_x();
                break;
    
            case NFLOOR:
                errorState = calc.floor();
                break;
        
            case NCEIL:
                errorState = calc.ceil();
                break;
        
            case NNEARINT:
                errorState = calc.nearint();
                break;
    
            case NERF:
                errorState = calc.erf();
                break;
    
            case NERFC:
                errorState = calc.erfc();
                break;
    
            case NGAMMA:
                errorState = calc.gamma();
                break;
    
            case NLNGAMMA:
                errorState = calc.lngamma();
                break;
                
            case NSINH:
                errorState = calc.sinh();
                break;
                
            case NCOSH:
                errorState = calc.cosh();
                break;
                
            case NTANH:
                errorState = calc.tanh();
                break;
            
            case NASINH:
                errorState = calc.asinh();
                break;
                
            case NACOSH:
                errorState = calc.acosh();        
                break;
                
            case NATANH:
                errorState = calc.atanh();
                break;
                
            case NREAL:
                errorState = calc.real();
                break;
                
            case NIMAG:
                errorState = calc.imag();
                break;
                
            case NTRANS:
                errorState = calc.trans();
                break;
                
            case NCONST:
                errorState = calc.toConst();
                break;
    
            case NABS:
                errorState = calc.abs();
                break;
                
            case NARG:
                errorState = calc.arg();
                break;
                
            case NEMPTY:    // Don't do anything if there is no argument
                errorState = ERRORCODE_TOOFEW;   // available for the function.
                break;
                
            default:
                errorState = ERRORCODE_NOERR;
                break;
        }
        
        displayMain( calc );
        displayEntry();
    }
}
    


// Display the help pages.
void RpcInterface::help( void )
{
    erase();
    move(0, 0);
    attron(A_BOLD);
    printw( "rpc v%s -- ONLINE HELP PAGES\n", VERSION );
    printw( "-----------------------------------------\n" );
    attroff(A_BOLD);
    printw( "Default hotkeys:\n" );
    printw( "<Enter>  --  enter number on stack\n" );
    printw( "             duplicate last stack entry\n" );
    printw( "      \\  --  drop last stack item\n" );
    printw( " <PgDn>  --  swap last two stack items\n" );
    printw( "      |  --  clear stack\n\n" );
    printw( " +  --  add             -  -- subtract\n"  );
    printw( " *  --  multiply        /  -- divide\n" );
    printw( " ^  --  power (x^y)     !  -- factorial\n" );
    printw( " %%  --  mod\n\n" );
    printw( "<Right>  --  change sign\n" );
    printw( "<Space>  --  scientific notation\n" );
    printw( "   <Up>  --  enter stack browsing mode\n" );
    printw( "     ^L  --  refresh display\n" );
    printw( "      S  --  setup the keymapping\n" );
    printw( "      ?  --  this help screen\n" );
    printw( "      Q  --  nice quit\n" );
    attron(A_BOLD);
    printw( "-----------------------------------------\n" );
    attroff(A_BOLD);
    printw( "Press any key to continue : " );
    refresh();
    getch();
    
    erase();
    move(0, 0);
    attron(A_BOLD);
    printw( "rpc v%s -- ONLINE HELP PAGES\n", VERSION );
    printw( "-----------------------------------------\n" );
    attroff(A_BOLD);
    printw( "To enter a simple (real) number, simply\n" );
    printw( "start typing digits.\n\n" );
    printw( "To enter a complex number, press '(',\n" );
    printw( "then enter the real part, then enter ','\n" );
    printw( "followed by the imaginary part.\n" );
    printw( "Alternatively, press '(', then enter the\n" );
    printw( "magnitude, then enter '<' followed by the\n" );
    printw( "angle.  The complex number will appear as\n" );
    printw( "an ordered pair.\n\n" );
    printw( "To enter a matrix, press '[', then\n" );
    printw( "enter the matrix elements separated by\n" );
    printw( "commas.  Press '[' again to start\n" );
    printw( "a new row of the matrix.\n\n" );
    printw( "To enter a complex matrix, simply press\n" );
    printw( "'(' after beginning matrix entry." );
    move(20, 0);
    attron(A_BOLD);
    printw( "-----------------------------------------\n" );
    attroff(A_BOLD);
    printw( "Press any key to continue : " );
    refresh();
    getch();
    
    erase();
    move(0, 0);
    attron(A_BOLD);
    printw( "rpc v%s -- ONLINE HELP PAGES\n", VERSION );
    printw( "-------------------------------------------------------\n" );
    attroff(A_BOLD);
    printw( "To select a special function, type the first few\n" );
    printw( "letters of the function (until it is identified\n" );
    printw( "uniquely).  Then press <Enter> to execute.  The\n" );
    printw( "following functions are executed in this fashion:\n\n" );
    printw( "   sin     asin     cos     acos    tan     atan\n" );
    printw( "   sq      sqrt     exp     ln      ten^x   log\n" );
    printw( "   erf     erfc     gamma   lngamma inv     floor\n" );
    printw( "   ceil    nearint  sinh    asinh   cosh    acosh\n" );	
    printw( "   tanh    atanh    real    imag    const\n\n" );
    printw( "          rad  -- enters radian mode\n" );
    printw( "          deg  -- enters degree mode\n" );
    printw( "          pi   -- enters 3.1415...\n" );
    printw( "          undo -- restores last stack\n" );
    printw( "          view -- views last stack element\n\n" );
    printw( "If you make an error in entering a function,\n" );
    printw( "<Backspace> will clear it.\n" );
    attron(A_BOLD);
    printw( "-------------------------------------------------------\n" );
    attroff(A_BOLD);
    printw( "Press any key to continue : " );
    refresh();
    getch();
    
    erase();
    move(0, 0);
    attron(A_BOLD);
    printw( "rpc v%s -- ONLINE HELP PAGES\n", VERSION );
    printw( "----------------------------------------------\n" );
    attroff(A_BOLD);
    printw( "In stack browsing mode, Up/Down and PgUp/\n" );
    printw( "PgDn scroll through the stack.  The\n" );
    printw( "following (default) hotkeys are\n" );
    printw( "available:\n\n" );
    printw( "  <Right> -- scroll right through selection\n" );
    printw( "   <Left> -- scroll left through selection\n" );
    printw( "  <Enter> -- echo selection to level 1\n" );
    printw( "        \\ -- delete selection\n" );
    printw( "        r -- cyclically roll entries up\n" );
    printw( "        R -- cyclically roll entries down\n" );
    printw( "        k -- keep entries below selected\n" );
    printw( "             level (inclusive)\n" );
    printw( "        d -- drop entries below selected\n" );
    printw( "             level (inclusive)\n" );
    printw( "        v -- view selection (fullscreen)\n" );
    printw( "        q -- exit interactive stack\n\n" );
    attron(A_BOLD);
    printw( "----------------------------------------------\n" );
    attroff(A_BOLD);
    printw( "Press 'a' for About, 'w' for Warranty\n" );
    printw( "information, or any other key to\n" );
    printw( "continue : " );
    refresh();
    key = getch();
    if( key == 'a' )
        about();
    else if( key == 'w' )
        warranty();
}


// Display an "about" page.
void RpcInterface::about( void )
{
    erase();
    move(0, 0);
    attron(A_BOLD);
    printw( "rpc v%s\n", VERSION );
    printw( "-----------------------------------------\n\n" );
    attroff(A_BOLD);
    printw( "A console-based RPN calculator\napplication.\n\n" );
    printw( "Copyright (c) 2000, 2001, 2002 Paul Pelzl\n\n\n" );
    printw( "\"Because, frankly, graphical calculator\n" );
    printw( " programs are pure evil.  " );
    attron(A_BOLD);
    printw( "rpc" );
    attroff(A_BOLD);
    printw( ", on the\n" );
    printw( " other hand, is only a little bit evil.\"\n\n" );
    attron(A_BOLD);
    printw( "-----------------------------------------\n" );
    attroff(A_BOLD);
    printw( "Press any key to continue : " );
    refresh();
    getch();
}


// Display a warranty page.
void RpcInterface::warranty( void )
{
    erase();
    move(0, 0);
    attron(A_BOLD);
    printw( "rpc v%s\n", VERSION );
    printw( "-----------------------------------------------------------------------------\n\n" );
    attroff(A_BOLD);
    printw( "rpc -- A console-based reverse polish notation calculator\n" );
    printw( "Copyright (C) 2000, 2001, 2002 Paul Pelzl\n\n" );
    printw( "This program is free software; you can redistribute it and/or\n" );
    printw( "modify it under the terms of the GNU General Public License\n" );
    printw( "as published by the Free Software Foundation; either version 2\n" );
    printw( "of the License, or (at your option) any later version.\n\n" );
    printw( "This program is distributed in the hope that it will be useful,\n" );
    printw( "but WITHOUT ANY WARRANTY; without even the implied warranty of\n" );
    printw( "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n" );
    printw( "GNU General Public License for more details.\n\n" );
    printw( "You should have received a copy of the GNU General Public License\n" );
    printw( "along with this program; if not, write to the Free Software\n" );
    printw( "Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.\n\n" );
    attron(A_BOLD);
    printw( "-----------------------------------------------------------------------------\n" );
    attroff(A_BOLD);
    printw( "Press any key to continue : " );
    refresh();
    getch();
}


//  Refresh the display.  Called either upon user request or
//  on xterm resize.
void RpcInterface::windowRefresh( RpcCalc& calc )
{
    endwin();               // Reset the ncurses window
    refresh();
    if( browsing )	
    {
        displayBrowse( calc );    // Show the proper screen in interactive mode
    }
    else
    {
        displayMain( calc );      // Otherwise show standard display
        displayEntry();
    }
}

    
//  Guts of the program.
//  Read raw input, decide what to do with it, redisplay the screen, etc.
//  Basic arithmetic operations (anything activated by a hotkey) are also 
//  implemented here, for simplicity.
void RpcInterface::run( RpcCalc& calc )
{
    int itemp;
    char stemp[2];
    RpcObject* ent;
    
    int endloop = 1;
    
    displayMain( calc );         // Start out by showing the main display
    
    while(endloop)
    {
        key = getch();           // Input a keypress


        //  USE THIS SWITCH FOR DEFAULT KEYMAPPING
        if( !userKeyMap )
        {    
            switch( key )
            {
                case '\012':         // CR:  duplicate or enter number on stack
                case KEY_ENTER:      // ncurses enter mapping does same thing
                    if( funExists )
                        specialFunction( calc );
                    else if( !entryExists )
                    {
                        if( calc.numElements() > 0 )    // requires nonempty stack
                        {
                            calc.backup();
                            errorState = calc.dup();
                            displayMain( calc );
                        }
                    }
                    else
                    {
                        calc.backup();
                        ent = getEntry(calc);
                        errorState = calc.enter( ent );
                        delete ent;
                        displayMain( calc );
                    }    
                    break;

                case '+':                // add
                    if( funExists )
                    {
                        specialFunction( calc );
                    }
                    if( entryExists )
                    {
                        ent = getEntry(calc);
                        calc.enter( ent );      // Enter buffer on stack
                        delete ent;
                        displayMain( calc );
                    }
                    if( calc.numElements() > 1 )        // requires two elements
                    {
                        calc.backup();
                        errorState = calc.add();
                        displayMain( calc );
                    }
                    break;

                case '-':                // subtract
                    if( funExists )
                    {
                        specialFunction( calc );
                    }
                    if( entryExists )
                    {
                        ent = getEntry(calc);
                        calc.enter( ent );      // Enter buffer on stack
                        delete ent;
                        displayMain( calc );
                    }
                    if( calc.numElements() > 1 )        // requires two elements
                    {
                        calc.backup();
                        errorState = calc.subtract();
                        displayMain( calc );
                    }
                    break;

                case '*':                        // multiply
                    if( funExists )
                    {
                        specialFunction( calc );
                    }
                    if( entryExists )
                    {
                        ent = getEntry(calc);
                        calc.enter( ent );      // Enter buffer on stack
                        delete ent;
                        displayMain( calc );
                    }
                    if( calc.numElements() > 1 )        // requires two elements
                    {
                        calc.backup();
                        errorState = calc.multiply();
                        displayMain( calc );
                    }
                    break;
                
                case '/':                        // divide
                    if( funExists )
                    {
                        specialFunction( calc );
                    }
                    if( entryExists )
                    {
                        ent = getEntry(calc);
                        calc.enter( ent );      // Enter buffer on stack
                        delete ent;
                        displayMain( calc );
                    }
                    if( calc.numElements() > 1 )        // requires two elements
                    {
                        calc.backup();
                        errorState = calc.divide();
                        displayMain( calc );
                    }
                    break;

                case '^':                // raise to a power
                    if( funExists )
                    {
                        specialFunction( calc );
                    }
                    if( entryExists )
                    {
                        ent = getEntry(calc);
                        calc.enter( ent );      // Enter buffer on stack
                        delete ent;
                        displayMain( calc );
                    }
                    if( calc.numElements() > 1 )        // requires two elements
                    {
                        calc.backup();
                        errorState = calc.pow();
                        displayMain( calc );
                    }
                    break;

                case '%':                // Mod function
                    if( funExists )
                    {
                        specialFunction( calc );
                    }
                    if( entryExists )
                    {
                        ent = getEntry(calc);
                        calc.enter( ent );      // Enter buffer on stack
                        delete ent;
                        displayMain( calc );
                    }
                    if( calc.numElements() > 1 )        // requires two elements
                    {
                        calc.backup();
                        errorState = calc.mod();
                        displayMain( calc );
                    }
                    break;
                    
                // factorial, calculated from gamma() to allow type double argument.
                // Note: n! = EulerGamma[n+1]
                case '!':                      
                    if( funExists )
                    {
                        specialFunction( calc );
                    }
                    if( entryExists )          
                    { 
                        ent = getEntry(calc);
                        calc.enter( ent );        // Enter buffer on stack
                        delete ent;
                        displayMain( calc );
                    }
                    if( calc.numElements() > 0 )        // requires nonempty stack
                    {
                        calc.backup();
                        errorState = calc.factorial();
                        displayMain( calc );
                    }
                    break;

                case ' ':                // begin scientific notation entry
                    if( !expExists[currentBuffer] && entryExists )
                    {
                        expExists[currentBuffer] = true;
                        displayEntry();
                    }
                    break;        

                case '(':                // enter a complex number
                    if( dataType == DOUBLE_TYPE )
                    {
                        entryExists = true;
                        polar = false;
                        dataType = COMPLEX_TYPE;
                        displayEntry();
                    }
                    else if( dataType == REALMATRIX_TYPE )
                    {
                        entryExists = true;
                        polar = false;
                        dataType = COMPLEXMATRIX_TYPE;
                        displayEntry();
                    }                                       
                    break;
                    
                case '[':                // enter a matrix, or begin adding a
                                         // new matrix row
                    if( dataType == DOUBLE_TYPE )
                    {
                        entryExists = true;
                        dataType = REALMATRIX_TYPE;
                        mtrxRows = 1;
                        mtrxCols = 1;
                        displayEntry();
                    }
                    else if( dataType == REALMATRIX_TYPE )
                    {
                        entryExists = true;
                        // We must be editing the last column of the matrix
                        // if we want to go to the next row.
                        if( ((currentBuffer+2)/2)%mtrxCols == 0 )
                        {
                            currentBuffer += 2;
                            mtrxRows++;
                            displayEntry();
                        }
                    }
                    else if( dataType == COMPLEXMATRIX_TYPE )
                    {
                        entryExists = true;
                        // We must be editing the last column of the matrix
                        // if we want to go to the next row.
                        if( ((currentBuffer+2)/2)%mtrxCols == 0 )
                        {
                            currentBuffer++;
                            mtrxRows++;
                            displayEntry();
                        }
                    }
                    break;

                                        
                case ',':                // separate entries for complex, matrix
                    if( dataType == COMPLEX_TYPE && currentBuffer == 0 )
                    {
                        currentBuffer = 1;
                        displayEntry();
                    }
                    else if( dataType == REALMATRIX_TYPE )
                    {
                        if( mtrxRows == 1 )
                        {
                            currentBuffer += 2;
                            mtrxCols++;
                        }
                        // Don't allow a comma after the last entry in
                        // a row...
                        else if( ((currentBuffer+2)/2) % mtrxCols != 0 )
                        {
                            currentBuffer += 2;
                        }
                        displayEntry();
                    }
                    else if( dataType == COMPLEXMATRIX_TYPE )
                    {
                        if( mtrxRows == 1 )
                        {
                            currentBuffer++;
                            if( currentBuffer%2 == 0 )
                            {
                                mtrxCols++;
                            }
                        }
                        // Don't allow a comma after the last entry in
                        // a row...
                        else if( (currentBuffer+1) % (2*mtrxCols) != 0 )
                        {
                            currentBuffer++;
                        }
                        displayEntry();
                    }
                    break;

                case '<':                // complex entry, polar format
                    if( dataType == COMPLEX_TYPE && currentBuffer == 0 )
                    {
                        currentBuffer = 1;
                        polar = true;
                        displayEntry();
                    }
                    else if( dataType == COMPLEXMATRIX_TYPE && mtrxRows == 1
                              && currentBuffer%2 == 0)
                    {
                        currentBuffer++;
                        polar = true;
                        displayEntry();
                    }
                    break;
                                              
                                    
                case '\\':               // backslash kills last item 
                    if( entryExists || funExists )
                    {
                        clearEntry();    // clear entry buffers
                        funExists = false;  // clear the function buffers
                        clearFunctionEntry();
                        strcpy( functionDisplay, "" );
                        displayEntry();
                    }
                    else if( calc.numElements() > 0 )
                    {
                        calc.backup();
                        errorState = calc.drop();     // pop it into oblivion
                        displayMain( calc );
                        displayEntry();
                    }
                    break;

                case '\010':                // BS kills char from entry buffers		
                case '\177':                // DEL kills char from entry buffers
                case KEY_BACKSPACE:         // ncurses backspace mapping
                case KEY_DC:                // ncurses delete character mapping
                    if( funExists )
                    {
                        funExists = false;
                        clearFunctionEntry();
                        strcpy( functionDisplay, "" );
                        displayEntry();
                    }
                    else if( entryExists )
                    {
                        if( strlen(expEntry[currentBuffer]) == 1 && expExists[currentBuffer])  // delete exponent
                        {
                            expExists[currentBuffer] = false;
                            strcpy( expEntry[currentBuffer], " " );
                            displayEntry();
                        }
                        else if( strlen(expEntry[currentBuffer]) > 1 )   // delete char from exponent buffer
                        {
                            itemp = strlen(expEntry[currentBuffer]);
                            expEntry[currentBuffer][itemp-1] = '\0';
                            displayEntry();
                        }
                        else if( strlen(entry[currentBuffer]) > 0 )      // delete char from entry buffer
                        {
                            itemp = strlen(entry[currentBuffer]);
                            entry[currentBuffer][itemp-1] = '\0';
                            displayEntry();
                        }
                        else if( currentBuffer > 0 )       // delete current buffer entirely
                        {
                            if( dataType == REALMATRIX_TYPE )
                            {
                                if( mtrxRows > 1 )
                                {    
                                    if( currentBuffer == 2*(mtrxRows-1)*mtrxCols )
                                        mtrxRows--;
                                }
                                else
                                {
                                    mtrxCols--;
                                }
                                
                                currentBuffer -= 2;
                            }
                            else if( dataType == COMPLEXMATRIX_TYPE )
                            {
                                if( mtrxRows > 1)
                                {
                                    if( currentBuffer == 2*(mtrxRows-1)*mtrxCols )
                                        mtrxRows--;
                                }
                                else if( (currentBuffer+2) % (2*mtrxCols) == 0 )
                                {
                                    mtrxCols--;
                                }
                                
                                currentBuffer--;
                            }
                            else
                            {
                                currentBuffer--;
                            }
                            displayEntry();
                        }
                        else if( dataType == COMPLEX_TYPE ||
                            dataType == REALMATRIX_TYPE || dataType == COMPLEXMATRIX_TYPE )
                        {
                            dataType = DOUBLE_TYPE;
                            entryExists = false;
                            displayEntry();
                        }                            
                        else
                        {
                            clearEntry();
                            displayEntry();
                        }
                    }
                    else if( strlen(entry[0]) == 0 && entrySign[0][0] == '-' )
                    {
                        strcpy( entrySign[0], " " );
                        displayEntry();
                    }
                    else if( dataType == COMPLEX_TYPE )
                    {
                        dataType = DOUBLE_TYPE;
                        displayEntry();
                    }
                    displayEntry();
                    break;
                    
                case '|':                      // clear all ( '|' is shift-drop )
                    calc.backup();
                    calc.clearStack();
                    displayMain( calc );
                    break;        

                case KEY_RIGHT:             // change sign (right arrow)
                case KEY_LEFT:              // allow left arrow also
                    if( expExists[currentBuffer] )
                    {
                        if( expEntry[currentBuffer][0] == ' ' )
                            expEntry[currentBuffer][0] = '-';
                        else
                            expEntry[currentBuffer][0] = ' ';
                        displayEntry();
                    }
                    else if( !entryExists )
                    {
                        if( calc.numElements() > 0 )
                        {
                            calc.backup();
                            errorState = calc.changeSign();
                            displayMain( calc );
                            displayEntry();
                        }
                    }
                    else
                    {
                        if( entrySign[currentBuffer][0] == ' ' )
                            entrySign[currentBuffer][0] = '-';
                        else
                            entrySign[currentBuffer][0] = ' ';
                        
                        if( (currentBuffer==0) && strlen(entry[currentBuffer])==0 )
                        {
                            clearEntry();
                        }
                        
                        displayEntry();
                    }
                    break;           
			
                case KEY_PPAGE:             // ncurses PgUp
                case KEY_NPAGE:             // ncurses PgDn
                    calc.backup();
                    errorState = calc.swap();  // swap last two elements
                    displayMain( calc );
                    displayEntry();
                    break;        

                case KEY_UP:         // Enter stack browsing mode (up arrow)
                    if( calc.numElements() > 0 )
                    {
                        browsing = true;
                        browse( calc );
                        browsing = false;
                        displayMain( calc );
                        displayEntry();
                    }
                    break;        

                case '#':            // Input integer constant
                    constInput( calc );
                    displayMain( calc );
                    displayEntry();
                    break;
                    
                case 12:             // Refresh the display (^L, the vi mapping)
                case 18:             // (^R works also)
                    windowRefresh( calc );
                    break;
                    
                case '?':                   // ? for help
                    help();
                    displayMain( calc );
                    displayEntry();
                    break;        

                case 'Q':                   // Q to quit
                    endloop = 0;
                    break;        

                case 'S':                   // S to setup keys
                    setupUserKeyMap();
                    displayMain( calc );
                    displayEntry();
                    break;
                    
                default:                    // put numbers in buffer
                    if( validNumInput(key) )
                    {
                        if( !entryExists )
                        {
                            entryExists = true;
                        }
                        
                        if( expExists[currentBuffer] ) // put numbers in exponent...
                        {
                            if( strlen(expEntry[currentBuffer]) < 4 )
                            {
                                stemp[0] = key;
                                stemp[1] = '\0';
                                strcat(expEntry[currentBuffer], stemp);
                                displayEntry();
                            }
                        }
                        else if( strlen(entry[currentBuffer]) < ENTRYSIZE )   // ...or mantissa.
                        {
                            stemp[0] = key;
                            stemp[1] = '\0';
                            strcat(entry[currentBuffer], stemp);
                            displayEntry();
                        }
                    }
                    else
                        funInput(key);      // If the input is not numeric,
                                            //        try function input.
                    break;    
            }        



        }
        else
        {
            if( key == enterChar )
            {
                if( funExists )
                    specialFunction( calc );
                else if( !entryExists )
                {
                    if( calc.numElements() > 0 )    // requires nonempty stack
                    {
                        calc.backup();
                        errorState = calc.dup();
                        displayMain( calc );
                    }
                }
                else
                {
                    calc.backup();
                    ent = getEntry(calc);
                    errorState = calc.enter( ent );
                    delete ent;
                    displayMain( calc );
                }    
            }
            else if( key == addChar )
            {
                if( funExists )
                {
                    specialFunction( calc );
                }
                if( entryExists )
                {
                    ent = getEntry(calc);
                    calc.enter( ent );      // Enter buffer on stack
                    delete ent;
                    displayMain( calc );
                }
                if( calc.numElements() > 1 )        // requires two elements
                {
                    calc.backup();
                    errorState = calc.add();
                    displayMain( calc );
                }
            }
            else if( key == subtractChar )
            {
                if( funExists )
                {
                    specialFunction( calc );
                }
                if( entryExists )
                {
                    ent = getEntry(calc);                   
                    calc.enter( ent );      // Enter buffer on stack
                    delete ent;
                    displayMain( calc );
                }
                if( calc.numElements() > 1 )        // requires two elements
                {
                    calc.backup();
                    errorState = calc.subtract();
                    displayMain( calc );
                }
            }
            else if( key == multiplyChar )
            {
                if( funExists )
                {
                    specialFunction( calc );
                }
                if( entryExists )
                {
                    ent = getEntry(calc);                   
                    calc.enter( ent );      // Enter buffer on stack
                    delete ent;
                    displayMain( calc );
                }
                if( calc.numElements() > 1 )        // requires two elements
                {
                    calc.backup();
                    errorState = calc.multiply();
                    displayMain( calc );
                }
            }
            else if( key == divideChar )
            {
                if( funExists )
                {
                    specialFunction( calc );
                }
                if( entryExists )
                {
                    ent = getEntry(calc);                   
                    calc.enter( ent );      // Enter buffer on stack
                    delete ent;
                    displayMain( calc );
                }
                if( calc.numElements() > 1 )        // requires two elements
                {
                    calc.backup();
                    errorState = calc.divide();
                    displayMain( calc );
                }
            }
            else if( key == powerChar )
            {
                if( funExists )
                {
                    specialFunction( calc );
                }
                if( entryExists )
                {
                    ent = getEntry(calc);                   
                    calc.enter( ent );      // Enter buffer on stack
                    delete ent;
                    displayMain( calc );
                }
                if( calc.numElements() > 1 )        // requires two elements
                {
                    calc.backup();
                    errorState = calc.pow();
                    displayMain( calc );
                }
            }
            else if( key == modChar )
            {
                if( funExists )
                {
                    specialFunction( calc );
                }
                if( entryExists )
                {
                    ent = getEntry(calc);                   
                    calc.enter( ent );      // Enter buffer on stack
                    delete ent;
                    displayMain( calc );
                }
                if( calc.numElements() > 1 )        // requires two elements
                {
                    calc.backup();
                    errorState = calc.mod();
                    displayMain( calc );
                }
            }
            // factorial, calculated from gamma() to allow type double argument.
            // Note: n! = EulerGamma[n+1]
            else if( key == factorialChar )
            {
                if( funExists )
                {
                    specialFunction( calc );
                }
                if( entryExists )
                {
                    ent = getEntry(calc);                   
                    calc.enter( ent );        // Enter buffer on stack
                    delete ent;
                    displayMain( calc );
                }
                if( calc.numElements() > 1 )        // requires two elements
                {
                    calc.backup();
                    errorState = calc.factorial();
                    displayMain( calc );
                }
            }
            else if( key == exponentChar )
            {
                if( !expExists[currentBuffer] && entryExists )
                {
                    expExists[currentBuffer] = true;
                    displayEntry();
                }
            }
            else if( key == complexChar )
            {
                if( dataType == DOUBLE_TYPE )
                {
                    entryExists = true;
                    dataType = COMPLEX_TYPE;
                    displayEntry();
                }
                else if( dataType == REALMATRIX_TYPE )
                {
                    entryExists = true;
                    dataType = COMPLEXMATRIX_TYPE;
                    displayEntry();
                }                                       
            }
            else if( key == matrixChar )
            {
                if( dataType == DOUBLE_TYPE )
                {
                    entryExists = true;
                    dataType = REALMATRIX_TYPE;
                    mtrxRows = 1;
                    mtrxCols = 1;
                    displayEntry();
                }
                else if( dataType == REALMATRIX_TYPE )
                {
                    entryExists = true;
                    // We must be editing the last column of the matrix
                    // if we want to go to the next row.
                    if( ((currentBuffer+2)/2)%mtrxCols == 0 )
                    {
                        currentBuffer += 2;
                        mtrxRows++;
                        displayEntry();
                    }
                }
                else if( dataType == COMPLEXMATRIX_TYPE )
                {
                    entryExists = true;
                    // We must be editing the last column of the matrix
                    // if we want to go to the next row.
                    if( ((currentBuffer+2)/2)%mtrxCols == 0 )
                    {
                        currentBuffer++;
                        mtrxRows++;
                        displayEntry();
                    }
                }
            }
            else if( key == separatorChar )
            {
                if( dataType == COMPLEX_TYPE && currentBuffer == 0 )
                {
                    currentBuffer = 1;
                    displayEntry();
                }
                else if( dataType == REALMATRIX_TYPE )
                {
                    if( mtrxRows == 1 )
                    {
                        currentBuffer += 2;
                        mtrxCols++;
                    }
                    // Don't allow a comma after the last entry in
                    // a row...
                    else if( ((currentBuffer+2)/2) % mtrxCols != 0 )
                    {
                        currentBuffer += 2;
                    }
                    displayEntry();
                }
                else if( dataType == COMPLEXMATRIX_TYPE )
                {
                    if( mtrxRows == 1 )
                    {
                        currentBuffer++;
                        if( currentBuffer%2 == 0 )
                        {
                            mtrxCols++;
                        }
                    }
                    // Don't allow a comma after the last entry in
                    // a row...
                    else if( (currentBuffer+1) % (2*mtrxCols) != 0 )
                    {
                        currentBuffer++;
                    }
                    displayEntry();
                }
            }
            else if( key == polarChar )   // Complex number, polar format
            {
                if( dataType == COMPLEX_TYPE && currentBuffer == 0 )
                {
                    currentBuffer = 1;
                    polar = true;
                    displayEntry();
                }
                else if( dataType == COMPLEXMATRIX_TYPE && mtrxRows == 1
                          && currentBuffer%2 == 0)
                {
                    currentBuffer++;
                    polar = true;
                    displayEntry();
                }
            }                
            else if( key == dropChar )
            {
                if( entryExists || funExists )
                {
                    clearEntry();    // clear entry buffers
                    funExists = false;  // clear the function buffers
                    clearFunctionEntry();
                    strcpy( functionDisplay, "" );
                    displayEntry();
                }
                else if( calc.numElements() > 0 )
                {
                    calc.backup();
                    errorState = calc.drop();     // pop it into oblivion
                    displayMain( calc );
                    displayEntry();
                }
            }
            else if( key == backspaceChar )
            {
                if( funExists )
                {
                    funExists = false;
                    clearFunctionEntry();
                    strcpy( functionDisplay, "" );
                    displayEntry();
                }
                else if( entryExists )
                {
                    if( strlen(expEntry[currentBuffer]) == 1 && expExists[currentBuffer])  // delete exponent
                    {
                        expExists[currentBuffer] = false;
                        strcpy( expEntry[currentBuffer], " " );
                        displayEntry();
                    }
                    else if( strlen(expEntry[currentBuffer]) > 1 )   // delete char from exponent buffer
                    {
                        itemp = strlen(expEntry[currentBuffer]);
                        expEntry[currentBuffer][itemp-1] = '\0';
                        displayEntry();
                    }
                    else if( strlen(entry[currentBuffer]) > 0 )      // delete char from entry buffer
                    {
                        itemp = strlen(entry[currentBuffer]);
                        entry[currentBuffer][itemp-1] = '\0';
                        displayEntry();
                    }
                    else if( currentBuffer > 0 )       // delete current buffer entirely
                    {
                        if( dataType == REALMATRIX_TYPE )
                        {
                            if( mtrxRows > 1 )
                            {    
                                if( currentBuffer == 2*(mtrxRows-1)*mtrxCols )
                                    mtrxRows--;
                            }
                            else
                            {
                                mtrxCols--;
                            }

                            currentBuffer -= 2;
                        }
                        else if( dataType == COMPLEXMATRIX_TYPE )
                        {
                            if( mtrxRows > 1)
                            {
                                if( currentBuffer == 2*(mtrxRows-1)*mtrxCols )
                                    mtrxRows--;
                            }
                            else if( (currentBuffer+2) % (2*mtrxCols) == 0 )
                            {
                                mtrxCols--;
                            }

                            currentBuffer--;
                        }
                        else
                        {
                            currentBuffer--;
                        }
                        displayEntry();
                    }
                    else if( dataType == COMPLEX_TYPE ||
                        dataType == REALMATRIX_TYPE || dataType == COMPLEXMATRIX_TYPE )
                    {
                        dataType = DOUBLE_TYPE;
                        entryExists = false;
                        displayEntry();
                    }                            
                    else
                    {
                        clearEntry();
                        displayEntry();
                    }
                }
                else if( strlen(entry[0]) == 0 && entrySign[0][0] == '-' )
                {
                    strcpy( entrySign[0], " " );
                    displayEntry();
                }
                else if( dataType == COMPLEX_TYPE )
                {
                    dataType = DOUBLE_TYPE;
                    displayEntry();
                }
                displayEntry();
            }
            else if( key == clearChar )
            {
                calc.backup();      // clear all ( '|' is shift-drop )
                calc.clearStack();
                displayMain( calc );
            }
            else if( key == signChar )
            {
                if( expExists[currentBuffer] )
                {
                    if( expEntry[currentBuffer][0] == ' ' )
                        expEntry[currentBuffer][0] = '-';
                    else
                        expEntry[currentBuffer][0] = ' ';
                    displayEntry();
                }
                else if( !entryExists )
                {
                    if( calc.numElements() > 0 )
                    {
                        calc.backup();
                        errorState = calc.changeSign();
                        displayMain( calc );
                        displayEntry();
                    }
                }
                else
                {
                    if( entrySign[currentBuffer][0] == ' ' )
                        entrySign[currentBuffer][0] = '-';
                    else
                        entrySign[currentBuffer][0] = ' ';

                    if( (currentBuffer==0) && strlen(entry[currentBuffer])==0 )
                    {
                        clearEntry();
                    }

                    displayEntry();
                }
            }
            else if( key == swapChar )
            {
                calc.backup();
                errorState = calc.swap();  // swap last two elements
                displayMain( calc );
                displayEntry();
            }
            else if( key == browseChar )
            {
                if( calc.numElements() > 0 )
                {
                    browsing = true;
                    browse( calc );
                    browsing = false;
                    displayMain( calc );
                    displayEntry();
                }
            }
            else if( key == integerChar )
            {
                constInput( calc );
                displayMain( calc );
                displayEntry();
            }               
            else if( key == refreshChar )
            {
                windowRefresh( calc );
            }
            else if( key == '?' )
            {
                help();
                displayMain( calc );
                displayEntry();
            }
            else if( key == 'Q' )
            {
                endloop = 0;
            }
            else if( key == 'S' )
            {
                setupUserKeyMap();
                displayMain( calc );
                displayEntry();
            }
            else
            {

                if( validNumInput(key) )
                {
                    if( !entryExists )
                    {
                        entryExists = true;
                    }

                    if( expExists[currentBuffer] ) // put numbers in exponent...
                    {
                        if( strlen(expEntry[currentBuffer]) < 4 )
                        {
                            stemp[0] = key;
                            stemp[1] = '\0';
                            strcat(expEntry[currentBuffer], stemp);
                            displayEntry();
                        }
                    }
                    else if( strlen(entry[currentBuffer]) < ENTRYSIZE )   // ...or mantissa.
                    {
                        stemp[0] = key;
                        stemp[1] = '\0';
                        strcat(entry[currentBuffer], stemp);
                        displayEntry();
                    }
                }
                else
                    funInput(key);      // If the input is not numeric,
                                        //        try function input.

            }                    
        }
    }
}



// Load the user-defined keymapping.
void RpcInterface::loadUserKeyMap( void )
{
    ifstream keymapfile;
    keymapfile.open( rpcKeyMapFile );
    
    if( !keymapfile.bad() )
    {
        char version[10];
        keymapfile.read(version, sizeof(version) );  // Load the version info.
    
        // Only load the keymap file if the version information is currect.
        // Otherwise assume default keymapping.
        // This method prevents trying to load from an incompatible data file.
        if( strcmp(version, VERSION) == 0 )
        {
            keymapfile.read((char*)(&userKeyMap), sizeof(userKeyMap) );  // Read keymap flag

            keymapfile.read((char*)(&enterChar), sizeof(enterChar) );    // Then read the keymap.
            keymapfile.read((char*)(&addChar), sizeof(addChar) );
            keymapfile.read((char*)(&subtractChar), sizeof(subtractChar) );
            keymapfile.read((char*)(&multiplyChar), sizeof(multiplyChar) );
            keymapfile.read((char*)(&divideChar), sizeof(divideChar) );
            keymapfile.read((char*)(&powerChar), sizeof(powerChar) );
            keymapfile.read((char*)(&modChar), sizeof(modChar) );
            keymapfile.read((char*)(&factorialChar), sizeof(factorialChar) );
            keymapfile.read((char*)(&exponentChar), sizeof(exponentChar) );
            keymapfile.read((char*)(&complexChar), sizeof(complexChar) );
            keymapfile.read((char*)(&polarChar), sizeof(polarChar) );
            keymapfile.read((char*)(&matrixChar), sizeof(matrixChar) );
            keymapfile.read((char*)(&separatorChar), sizeof(separatorChar) );
            keymapfile.read((char*)(&integerChar), sizeof(integerChar) );
            keymapfile.read((char*)(&dropChar), sizeof(dropChar) );
            keymapfile.read((char*)(&backspaceChar), sizeof(backspaceChar) );
            keymapfile.read((char*)(&clearChar), sizeof(clearChar) );
            keymapfile.read((char*)(&signChar), sizeof(signChar) );
            keymapfile.read((char*)(&swapChar), sizeof(swapChar) );
            keymapfile.read((char*)(&browseChar), sizeof(browseChar) );
            keymapfile.read((char*)(&refreshChar), sizeof(refreshChar) );
            
            keymapfile.read((char*)(&moveUpChar), sizeof(moveUpChar) );
            keymapfile.read((char*)(&moveDownChar), sizeof(moveDownChar) );
            keymapfile.read((char*)(&pageUpChar), sizeof(pageUpChar) );
            keymapfile.read((char*)(&pageDownChar), sizeof(pageDownChar) );
            keymapfile.read((char*)(&scrollRightChar), sizeof(scrollRightChar) );
            keymapfile.read((char*)(&scrollLeftChar), sizeof(scrollLeftChar) );
            keymapfile.read((char*)(&dropnChar), sizeof(dropnChar) );
            keymapfile.read((char*)(&keepnChar), sizeof(keepnChar) );
            keymapfile.read((char*)(&rollUpChar), sizeof(rollUpChar) );
            keymapfile.read((char*)(&rollDownChar), sizeof(rollDownChar) );
            keymapfile.read((char*)(&viewChar), sizeof(viewChar) );
            keymapfile.read((char*)(&undoChar), sizeof(undoChar) );
        }
    }
    else
    {
        userKeyMap = false;        //  If there is no keymap file, go with default keymap.
    }
    
    keymapfile.close();
}
    

// Save the user-defined keymapping.
void RpcInterface::saveUserKeyMap( void )
{
    ofstream keymapfile;
    keymapfile.open( rpcKeyMapFile );
    
    if( keymapfile.bad() )
    {
        // Try creating the directory first
        system( "mkdir $HOME/.rpc" );
        keymapfile.close();
        keymapfile.open( rpcKeyMapFile );
    }        
    
    if( keymapfile.bad() )
    {
        cerr << "\nError: Unable to save keymap to ~/.rpc/keymap.\n";
    }
    else
    {
        char version[10];
        strcpy( version, VERSION );
        keymapfile.write(version, sizeof(version) );  // Load the version info.        

        keymapfile.write((char*)(&userKeyMap), sizeof(userKeyMap) );  // Write keymap flag

        keymapfile.write((char*)(&enterChar), sizeof(enterChar) );    // Then write the keymap.
        keymapfile.write((char*)(&addChar), sizeof(addChar) );
        keymapfile.write((char*)(&subtractChar), sizeof(subtractChar) );
        keymapfile.write((char*)(&multiplyChar), sizeof(multiplyChar) );
        keymapfile.write((char*)(&divideChar), sizeof(divideChar) );
        keymapfile.write((char*)(&powerChar), sizeof(powerChar) );
        keymapfile.write((char*)(&modChar), sizeof(modChar) );
        keymapfile.write((char*)(&factorialChar), sizeof(factorialChar) );
        keymapfile.write((char*)(&exponentChar), sizeof(exponentChar) );
        keymapfile.write((char*)(&complexChar), sizeof(complexChar) );
        keymapfile.write((char*)(&polarChar), sizeof(polarChar) );
        keymapfile.write((char*)(&matrixChar), sizeof(matrixChar) );
        keymapfile.write((char*)(&separatorChar), sizeof(separatorChar) );
        keymapfile.write((char*)(&integerChar), sizeof(integerChar) );
        keymapfile.write((char*)(&dropChar), sizeof(dropChar) );
        keymapfile.write((char*)(&backspaceChar), sizeof(backspaceChar) );
        keymapfile.write((char*)(&clearChar), sizeof(clearChar) );
        keymapfile.write((char*)(&signChar), sizeof(signChar) );
        keymapfile.write((char*)(&swapChar), sizeof(swapChar) );
        keymapfile.write((char*)(&browseChar), sizeof(browseChar) );
        keymapfile.write((char*)(&refreshChar), sizeof(refreshChar) );
    
        keymapfile.write((char*)(&moveUpChar), sizeof(moveUpChar) );
        keymapfile.write((char*)(&moveDownChar), sizeof(moveDownChar) );
        keymapfile.write((char*)(&pageUpChar), sizeof(pageUpChar) );
        keymapfile.write((char*)(&pageDownChar), sizeof(pageDownChar) );
        keymapfile.write((char*)(&scrollRightChar), sizeof(scrollRightChar) );
        keymapfile.write((char*)(&scrollLeftChar), sizeof(scrollLeftChar) );
        keymapfile.write((char*)(&dropnChar), sizeof(dropnChar) );
        keymapfile.write((char*)(&keepnChar), sizeof(keepnChar) );
        keymapfile.write((char*)(&rollUpChar), sizeof(rollUpChar) );
        keymapfile.write((char*)(&rollDownChar), sizeof(rollDownChar) );
        keymapfile.write((char*)(&viewChar), sizeof(viewChar) );
        keymapfile.write((char*)(&undoChar), sizeof(undoChar) );
    }        

    keymapfile.close();
}


// Set up the user-defined keymapping.
void RpcInterface::setupUserKeyMap( void )
{
    numKeysUsed = 0;
    numBrowseKeysUsed = 0;
    
    erase();
    move(0, 0);
    attron(A_BOLD);
    printw( "rpc v%s -- KEYMAPPING SETUP\n", VERSION );
    printw( "------------------------------------------------------------" );
    move(11, 0);
    printw( "------------------------------------------------------------" );
    attroff(A_BOLD);
    move(3, 0);
    printw( "Do you wish to use the default keymapping (y/n)? " );
    refresh();
    
    key = getch();
    if( key == 'n' )
    {
        userKeyMap = true;
        
        move(3, 0);
        clrtoeol();
        printw( "You may select keys that are not numbers, lowercase\n" );
        printw( "letters, or one of the reserved keys 'Q', 'S', '?'\n" );
        printw( "and '.'.  Hit 'd' to choose the default.\n\n" );
        printw( "Press the key you wish to use for" );

        setUserKey( 8, "        enter/dup <enter>: ", 
            '\012', enterChar );
        
        // Browse mode uses the same enter key...
        browseKeysUsed[numBrowseKeysUsed] = enterChar;
        numBrowseKeysUsed++;
        
        setUserKey( 8, "        addition <+>: ", 
            '+', addChar );            

        setUserKey( 8, "        subtraction <->: ",
            '-', subtractChar );

        setUserKey( 8, "        multiplication <*>: ", 
            '*', multiplyChar );

        setUserKey( 8, "        division </>: " , 
            '/', divideChar );
            
        setUserKey( 8, "        raising to a power <^>: ", 
            '^', powerChar );
        
        setUserKey( 8, "        the mod operator <%>: ",
            '%', modChar );
            
        setUserKey( 8, "        factorial <!>: ", 
            '!', factorialChar );
            
        setUserKey( 8, "        scientific notation < >: ", 
            ' ', exponentChar );
        
        setUserKey( 8, "        entering a complex number <(>: ",
            '(', complexChar );        
        
        setUserKey( 8, "        entering a matrix <[>: ",
            '[', matrixChar );
        
        setUserKey( 8, "        entering the complex imaginary part <,>: ",
            ',', separatorChar );
        
        setUserKey( 8, "        entering a complex angle (<): ",
            '<', polarChar );
        
        setUserKey( 8, "        entering an integer constant <#>: ",
            '#', integerChar );
            
        setUserKey( 8, "        drop <\\>: ", 
            '\\', dropChar );
        
        // Browse mode uses the same drop key...
        browseKeysUsed[numBrowseKeysUsed] = dropChar;
        numBrowseKeysUsed++;
            
        setUserKey( 8, "        backspace <BS>: ", 
            KEY_BACKSPACE, backspaceChar );

        setUserKey( 8, "        clearing the stack <|>: ", 
            '|', clearChar );
            
        setUserKey( 8, "        sign change <Right>: ", 
            KEY_RIGHT, signChar );

        setUserKey( 8, "        swap <PgDn>: ", 
            KEY_NPAGE, swapChar );
            
        setUserKey( 8, "        browsing the stack <Up>: ", 
            KEY_UP, browseChar );
            
        setUserKey( 8, "        refreshing the display <^L>: ",
            12, refreshChar );
    
        // Browse mode uses the same refresh key...
        browseKeysUsed[numBrowseKeysUsed] = refreshChar;
        numBrowseKeysUsed++;
        

        move(3, 0);
        clrtoeol();
        printw("Now you can set the keys to be used in stack\n" );
        clrtoeol();
        printw("browsing mode.  You may use any key except for 'q'\n" );
        clrtoeol();
        printw("and those you chose for enter/dup, drop, and\n" );
        clrtoeol();
        printw("sign change.  Hit <enter> to use the default.\n\n" );
        clrtoeol();
        printw("Press the key you wish to use for" );

        setBrowseUserKey( 9, "        scrolling up the stack <Up>: ",
            KEY_UP, moveUpChar );

        setBrowseUserKey( 9, "        scrolling down the stack <Down>: ",
            KEY_DOWN, moveDownChar );
   
        setBrowseUserKey( 9, "        paging up the stack <PgUp>: ",
            KEY_PPAGE, pageUpChar );
   
        setBrowseUserKey( 9, "        paging down the stack <PgDn>: ",
            KEY_NPAGE, pageDownChar );
        
        setBrowseUserKey( 9, "        scrolling right through an entry <Right>: ",
            KEY_RIGHT, scrollRightChar );
        
        setBrowseUserKey( 9, "        scrolling left through an entry <Left>: ",
            KEY_LEFT, scrollLeftChar );

        setBrowseUserKey( 9, "        dropping n entries <d>: ",
            'd', dropnChar );

        setBrowseUserKey( 9, "        keeping n entries <k>: ",
            'k', keepnChar );

        setBrowseUserKey( 9, "        rolling stack entries up <r>: ",
            'r', rollUpChar );

        setBrowseUserKey( 9, "        rolling stack entries down <R>: ",
            'R', rollDownChar );
        
        setBrowseUserKey( 9, "        fullscreen viewing a stack entry <v>: ",
            'v', viewChar );

        setBrowseUserKey( 9, "        undoing all changes <u>: ",
            'u', undoChar );

        saveUserKeyMap();             // Save the keymap to a data file.


        for( int i=3; i<11; i++)
        {
            move(i, 0);
            clrtoeol();
        }
        
        move(3, 0);
        printw( "Keymap saved.  Press any key to continue : " );
        refresh();
        key = getch();

    }
    else
    {
        userKeyMap = false;
        saveUserKeyMap();

        move(3, 0);
        clrtoeol();
        printw( "Using default keymapping.  Press any key to continue : " );
        refresh();
        key = getch();
    }    
}
    


// Accept input of an integer constant.  This routine is flagged by
// hitting '#'.  It allows entry of characters 0-f (hex set), and the user
// may optionally suffix his entry with <space>(d/h/o/b) to indicate the
// desired base.
void RpcInterface::constInput( RpcCalc& calc )
{
    
    strcpy( constEntry, "" );
    strcpy( constBase, "" );
    chooseBase = false;    
    
    displayConstEntry();
    
    int endloop = 1;
    char stemp[2] = { '\0', '\0' };
    RpcObject* ent;
    
    while( endloop )
    {
        key = getch();
        
        switch( key )
        {
            case '\012':
            case KEY_ENTER:     // Enter the integer constant on the stack
                ent = getConstEntry(calc);
                errorState = calc.enter( ent );
                delete ent;
                
                clearEntry();
                dataType = DOUBLE_TYPE;
                endloop = 0;
                displayConstEntry();
                break;
                
            case '\010':
            case '\177':
            case KEY_BACKSPACE: // Delete last character entered
            case KEY_DC:
                if( chooseBase )
                {
                    if( strlen(constBase) > 0 )
                    {
                        strcpy( constBase, "" );
                    }
                    else
                    {
                        chooseBase = false;
                    }
                }
                else if( strlen(constEntry) > 0 )
                {
                    constEntry[strlen(constEntry)-1] = '\0';                   
                }
                else
                {
                    endloop = 0;
                }
                displayConstEntry();
                break;
                
            case ' ':           // Space indicates entry in a specific base.
                chooseBase = true;
                displayConstEntry();
                break;
                
            default:
                if( chooseBase )
                {
                    if( intStrchr( "bodh", key ) )
                    {
                        stemp[0] = key;
                        strcpy( constBase, stemp );
                        displayConstEntry();
                    }
                }
                else
                {
                    if( intStrchr( "1234567890abcdefABCDEF", key ) &&
                             strlen( constEntry ) < CONST_ENTRYSIZE )
                    {
                        stemp[0] = key;
                        strcat( constEntry, stemp );
                        displayConstEntry();
                    }
                }
                break;
        }
    }       
}            
                
                
                
                

RpcObject* RpcInterface::getConstEntry( RpcCalc& calc )
{
    RpcObject* temp = NULL;
    unsigned long longtemp;
    
    // If no base specifier is given, assume the desired base
    //     is the mode the calculator is currently in.
    if( strlen(constBase) == 0 )
    {
        if( calc.flags.constMode == MODE_HEX )
        {
            strcpy( constBase, "h" );
        }
        else if( calc.flags.constMode == MODE_BIN )
        {
            strcpy( constBase, "b" );
        }
        else if( calc.flags.constMode == MODE_OCT )
        {
            strcpy( constBase, "o" );
        }
        else
        {
            strcpy( constBase, "d" );
        }
    }
    
    switch( constBase[0] )
    {
        case 'h':
            longtemp = strtoul( constEntry, NULL, 16 );     // treat as hex
            break;
            
        case 'b':
            longtemp = strtoul( constEntry, NULL, 2);       // treat as binary
            break;
            
        case 'o':
            longtemp = strtoul( constEntry, NULL, 8);       // treat as octal
            break;
            
        case 'd':
        default:
            longtemp = strtoul( constEntry, NULL, 10 );     // treat as decimal
            break;            
    }
                                    
    temp = new RpcIntegerConst( IntegerConst(longtemp) );
    
    // clean up the details
    strcpy( constEntry, "" );
    strcpy( constBase, "" );
    chooseBase = false;
        
    return( temp );
}
                


//  Display the entry buffer for a constant entry
void RpcInterface::displayConstEntry( void )
{

    char temp[2048];
    
    strcpy( temp, "# " );    
    strcat( temp, constEntry );
    
    if( chooseBase )
    {
        strcat( temp, " " );
        strcat( temp, constBase );
    }
    
    // Display the buffer via curses.
    move(LINES-1, 0);
    clrtoeol();

    if( strlen(temp) <= SWIDTH )
    {
        move(LINES-1, SWIDTH-strlen(temp));
        printw( temp );
    }
    else
    {
        printw( temp+strlen(temp)-SWIDTH );
    }

    refresh();
}
    
               
                
//  Boolean:  Is the character valid to use as a hotkey?
//  Make sure the input does not match numbers or lowercase letters.
bool RpcInterface::validHotKey( int input )
{
    bool valid = true;
    
    // Make sure the key doesn't match the "illegal keys":
    // It can't be a number or decimal, lowercase, or ?, Q, or S.
    if( intStrchr( "1234567890.abcdefghijklmnopqrstuvwxyz?QS", input ) )
    {
        valid = false;
    }

    // Make sure the key doesn't match a key which is already used
    for( int i=0; i<numKeysUsed; i++ )
    {
        if( input == keysUsed[i] )
            valid = false;
    }

    return( valid );
}



//  Boolean:  Is the character valid to use as a hotkey in browsing mode?
//  Make sure the input does not match numbers or lowercase letters.
bool RpcInterface::validBrowseHotKey( int input )
{
    bool valid = true;

    // Make sure the key doesn't match the "illegal key"
    if( input == int('q') )
    {
        valid = false;
    }

    // Make sure the key doesn't match a key which is already used
    for( int i=0; i<numBrowseKeysUsed; i++ )
    {
        if( input == browseKeysUsed[i] )
            valid = false;
    }

    return( valid );
}



//  Modify a single user keymap for the stack hotkeys
void RpcInterface::setUserKey( int lineNum, char* promptString,
    int defaultKey, int &userKey )
{
    bool valid;
    
    move(lineNum+1, 0);
    clrtoeol();
        
    valid = false;
    while( !valid )
    {
        move(lineNum, 0);
        clrtoeol();
        printw( "%s", promptString );
        refresh();
        
        key = getch();
        if( key == 'd' && validHotKey( defaultKey ) )
        {
            userKey = defaultKey;
            keysUsed[numKeysUsed] = defaultKey;
            numKeysUsed++;
            valid = true;
        }
        else if( validHotKey( key ) )
        {
            userKey = key;
            keysUsed[numKeysUsed] = key;
            numKeysUsed++;
            valid = true;
        }
        else
            printw( "\nThat key is invalid, or is already used." );
            
    }
}



//  Modify a single user keymap for browsing mode.
void RpcInterface::setBrowseUserKey( int lineNum, char* promptString,
    int defaultKey, int &userKey )
{
    int key;
    bool valid;

    move(lineNum+1, 0);
    clrtoeol();

    valid = false;
    while( !valid )
    {
        move(lineNum, 0);
        clrtoeol();
        printw( "%s", promptString );
        refresh();

        key = getch();
        if( key == '\012' && validBrowseHotKey( defaultKey ) )
        {
            userKey = defaultKey;
            browseKeysUsed[numBrowseKeysUsed] = defaultKey;
            numBrowseKeysUsed++;
            valid = true;
        }
        else if( validBrowseHotKey( key ) )
        {
            userKey = key;
            browseKeysUsed[numBrowseKeysUsed] = key;
            numBrowseKeysUsed++;
            valid = true;
        }
        else
            printw( "\nThat key is invalid, or is already used." );

    }
}



//  Tests whether the integer input matches one of the characters
//  in the string.  Replacement for strchr(), which expects a char
//  in the second position (and will demote the int to a char).
//  Returns true if there is a match.
bool RpcInterface::intStrchr( char* charString, int input )
{
    int i=0;
    bool output = false;

    while( *(charString+i) != '\0' )
    {
        if( int( *(charString+i) ) == input )
            output = true;
            
        i++;
    }
    
    return( output );
}



// arch-tag: rpc curses interface class
