/*  rpc -- A console-based reverse polish notation calculator
    Copyright (C) 2000, 2001, 2002 Paul Pelzl

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



/* RpcInterface.h
 * Declaration of the class that handles basic input and creates the
 * standard stack display using curses.
 */
 

#include <iostream>

#if defined(USE_NCURSES) && !defined(RENAMED_NCURSES)
#include <ncurses.h>
#else
#include <curses.h>
#endif

#include <csignal>
#include <cstdlib>
#include "RpcCalc.h"

using namespace std;


// Assign numbers to basic operations.
#define NSIN  0
#define NCOS  1
#define NTAN  2
#define NSQ   3
#define NSQRT 4
#define NINV  5
#define NEXP  6
#define NTEN  7
#define NLN   8
#define NLOG  9
#define NASIN 10
#define NACOS 11
#define NATAN 12
#define NRAD  13
#define NDEG  14
#define NPI   15
#define NUNDO 16
#define NFLOOR 17
#define NCEIL 18
#define NNEARINT 19
#define NGAMMA 20
#define NLNGAMMA 21
#define NERF 22
#define NERFC 23
#define NSINH 24
#define NCOSH 25
#define NTANH 26
#define NASINH 27
#define NACOSH 28
#define NATANH 29
#define NREAL 30
#define NIMAG 31
#define NVIEW 32
#define NTRANS 33
#define NCONST 34
#define NDEC 35
#define NHEX 36
#define NBIN 37
#define NOCT 38
#define NABS 39
#define NARG 40
#define NRECT 41
#define NPOLAR 42

// Always leave NEMPTY as last number.
#define NEMPTY 43
#define MAXN 43

#define SWIDTH 38
#define SWIDTH2 47

#define ENTRYSIZE 19
#define CONST_ENTRYSIZE 32

class RpcInterface
{
    private:
        char entry[1024][20];        // Buffer for mantissa data entry.
                                     //        Actually 1024 buffers are allowed
                                     //        to provide for matrices, etc.
        bool entryExists;            // Something is entered.
        char entrySign[1024][2];     // Stores the sign char for each buffer.
        
        char expEntry[1024][5];      // Buffer(s) for exponent.
        bool expExists[1024];        // Is there an exponent?
        int buffersUsed;             // Number of the above buffers actually
                                     //        being used
        int currentBuffer;           // Which buffer is being edited?
        
        char functionEntry[10];      // Buffer for function entry.
        char functionDisplay[50];    // Buffer for the function to be
                                     //        displayed when a function is
                                     //        chosen.
        bool funExists;              // Has a function been chosen?
        int funChoice;               // Number corresponding to the current
                                     //        choice of special function.
        bool funArg;                 // Does the function require an argument?
        bool polar;                  // Is polar format being used for
                                     //       complex entry?
        unsigned int dataType;       // The type of data being entered.
        
        bool browsing;               // Is the user browsing the stack?
                                     //        (Used during window resize.)
        bool userKeyMap;             // Has the user defined his/her own
                                     //        key mapping?
        
        char rpcKeyMapFile[256];     // Full path to user's keymap file.
        char rpcViewBufferFile[256]; // Full path to the file that stores fullscreend data
        
        char rpcEditor[256];         // Full path to a fullscreen editor
        char rpcEditorFile[256];     // Full path to the file that stores the choice of editor
        
        char viewCommand[512];    // System command for fullscreen viewing
        
        int keysUsed[30];            // The keys used in user-defined keymapping.
        int browseKeysUsed[30];      // Same, but for browsing mode keymapping.
        
        int numKeysUsed;             // Number of keys chosen in user-defined
                                     //        keymapping.
        int numBrowseKeysUsed;       // Same thing, but for stack browsing keys.
        
        int errorState;              // Stores the error state of the calc.
        
        int mtrxRows;                // Store the number of matrix rows
        int mtrxCols;                //    and columns currently being entered.
        
        int key;                     // Stores a keypress.
        
        char constEntry[1024];
        char constBase[2];
        bool chooseBase;
                                             
        int enterChar;               // User-defined keymapping variables.
        int addChar;
        int subtractChar;
        int multiplyChar;
        int divideChar;
        int powerChar;
        int modChar;
        int factorialChar;
        int exponentChar;
        int complexChar;
        int polarChar;
        int matrixChar;
        int separatorChar;
        int integerChar;
        int dropChar;
        int backspaceChar;
        int clearChar;
        int signChar;
        int swapChar;
        int browseChar;
        int refreshChar;
    
        int moveUpChar;
        int moveDownChar;
        int pageUpChar;
        int pageDownChar;
        int scrollLeftChar;
        int scrollRightChar;
        int dropnChar;
        int keepnChar;
        int rollUpChar;
        int rollDownChar;
        int undoChar;
        int viewChar;
        
        int stackLevel;                  // Used in browsing mode
        int screenBottom;
        unsigned int horizShift;         // For scrolling left and right

        void displayEntry( void );       // Display the entry buffer
        void displayMain( RpcCalc& );    // Display the stack data
        void displayBrowse( RpcCalc& );  // Display the browse screen
        void displayConstEntry( void );  // Display the entry buffer when an
                                         //      integer constant is being
                                         //      entered.
        
        void browse( RpcCalc& );         // Browse the stack
        
        void help( void );               // Display a help page
        void about( void );              // About screen
        void warranty( void );           // Warranty screen
        
        RpcObject* getEntry( RpcCalc & );   // Translate the entry buffer(s)
                                         //     into an RpcObject.
        double parseBuffer( int );       // Get a number from an entry buffer
        void parseFunBuffer( void );     // Translate function entry buffer
                                         //     into a valid function string to
                                         //     be displayed.
        bool validNumInput( int );       // Boolean: can this character be
                                         //     entered into the numeric input
                                         //     buffer?
                                         
        void funInput( char );           // Accept characters used to choose
                                         //        a function, and make sure
                                         //        they are valid.
                                         
        void specialFunction( RpcCalc& ); // Calculates a special function.
        void clearEntry( void );         // Clear the entry buffers.
        void clearFunctionEntry( void ); // REALLY clear 'functionEntry'

        void loadUserKeyMap( void );     // Load the user-defined keymapping
        void saveUserKeyMap( void );     // Save the user-defined keymapping
        void setupUserKeyMap( void );    // Set up the user-defined keymapping.

        void constInput( RpcCalc& );      // Accept input of an integer constant
        RpcObject* getConstEntry( RpcCalc& ); // Translate the constant entry buffer
                                              //      into a valid RpcObject.
        
        
        bool validHotKey( int );         // Boolean:  Is the char okay to use
                                         //     as a hotkey?
        bool validBrowseHotKey( int );   // Same, but for browsing mode keys

        void setUserKey( int, char *, int, int &);  // Modify one user keymap.
        void setBrowseUserKey( int, char *, int, int &);  // Same, but for browsing keys

        bool intStrchr( char *, int );   // Check if the character in the int
                                         //       is present in the string

    public:
        RpcInterface();           // Initialization
        ~RpcInterface();
        
        // No nasty pointer tricks in this class, so default assignment
        // and copy constructor are okay.
        

        void windowRefresh( RpcCalc& ); // Refresh the screen.  Called either
                                        //    by user input or by xterm resize.

        void run( RpcCalc& );   // Get input and do stuff with it.
};
        


// arch-tag: rpc curses interface class (header)
