/*  rpc -- A console-based reverse polish notation calculator
    Copyright (C) 2000, 2001, 2002 Paul Pelzl

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



/*  RpcObject.cpp
 *  Base class from which all types of rpc data will be derived.
 */

 
#include "RpcObject.h"

// Default constructor.  Shouldn't be called directly.
RpcObject::RpcObject()
{
    type = DOUBLE_TYPE;
}


// Default destructor.
RpcObject::~RpcObject()
{
}


// Create an RpcObject with specified type.
RpcObject::RpcObject( unsigned int input )
{
    type = input;
}


// Copy constructor.
RpcObject::RpcObject( const RpcObject& oldObject )
{
    type = oldObject.type;
}


// Assignment operator.
RpcObject& RpcObject::operator =( const RpcObject& oldObject )
{
    if( this != &oldObject )        // Check self-assignment.
    {
        type = oldObject.type;
    }
    
    return( *this );
}


// Return the display string for the specified object.
// An error is returned unless the derived class overrides this.
char* RpcObject::getDisplayString( RpcCalcFlags flags ) const
{
    char* temp = new char[256];
    strcpy( temp, "<no display available>" );
    return( temp );
}


// Return the fullscreen display string for the specified object.
// An error is returned unless the derived class overrides this.
char* RpcObject::getFullscreenString( RpcCalcFlags flags ) const
{
    char* temp = new char[256];
    strcpy( temp, "<no display available>" );
    return( temp );
} 

// Addition operator.  Returns an error object.
// Should be overridden by all classes for which
// addition makes sense.
inline RpcObject* RpcObject::operator +( const RpcObject* rhs )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Subtraction operator.  Returns an error object.
// Should be overridden by all classes for which
// subtraction makes sense.
inline RpcObject* RpcObject::operator -( const RpcObject* rhs )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Multiplication operator.  Returns an error object.
// Should be overridden by all classes for which
// multiplication makes sense.
inline RpcObject* RpcObject::operator *( const RpcObject* rhs )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Division operator.  Returns an error object.
// Should be overridden by all classes for which
// division makes sense.
inline RpcObject* RpcObject::operator /( const RpcObject* rhs )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Change sign operator.  Returns an error object.
// Should be overridden by all classes for which
// changeSign makes sense.
inline RpcObject* RpcObject::operator -( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Power function.  Returns an error object.
// Should be overridden by all classes for which
// power makes sense.
inline RpcObject* RpcObject::Pow( const RpcObject* rhs )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Mod function. Returns an error object.
// Should be overriden by all classes for which mod
// makes sense.
inline RpcObject* RpcObject::Mod( const RpcObject* rhs )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Square root.  Returns an error object.
// Should be overridden by all classes for which sqrt makes sense.
inline RpcObject* RpcObject::Sqrt( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Square.  Returns an error object.
// Should be overridden by all classes for which square makes sense.
inline RpcObject* RpcObject::Sq( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Exponential function.  Returns an error object.
// Should be overridden by all classes for which exp makes sense.
inline RpcObject* RpcObject::Exp( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Natural logarithm.  Returns an error object.
// Should be overridden by all classes for which log makes sense.
inline RpcObject* RpcObject::Log( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Ten raised to the power x.  Returns an error object.
// Should be overridden by all classes for which ten_x makes sense.
inline RpcObject* RpcObject::Ten_x( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Common logarithm.  Returns an error object.
// Should be overridden by all classes for which log10 makes sense.
inline RpcObject* RpcObject::Log10( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Sine function.  Returns an error object.
// Should be overridden by all classes for which sine makes sense.
inline RpcObject* RpcObject::Sin( int mode )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Cosine function.  Returns an error object.
// Should be overridden by all classes for which cosine makes sense.
inline RpcObject* RpcObject::Cos( int mode )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Tangent function.  Returns an error object.
// Should be overridden by all classes for which tangent makes sense.
inline RpcObject* RpcObject::Tan( int mode )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Inverse sine function.  Returns an error object.
// Should be overridden by all classes for which asin makes sense.
inline RpcObject* RpcObject::aSin( int mode )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Inverse cosine function.  Returns an error object.
// Should be overridden by all classes for which acos makes sense.
inline RpcObject* RpcObject::aCos( int mode )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Inverse tangent function.  Returns an error object.
// Should be overridden by all classes for which inverse tangent makes sense.
inline RpcObject* RpcObject::aTan( int mode )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Hyperbolic sine function.  Returns an error object.
// Should be overridden by all classes for which sinh makes sense.
inline RpcObject* RpcObject::Sinh( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Hyperbolic cosine function.  Returns an error object.
// Should be overridden by all classes for which cosh makes sense.
inline RpcObject* RpcObject::Cosh( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Hyperbolic tangent function.  Returns an error object.
// Should be overridden by all classes for which tanh makes sense.
inline RpcObject* RpcObject::Tanh( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Inverse hyperbolic sine function.  Returns an error object.
// Should be overridden by all classes for which asinh makes sense.
inline RpcObject* RpcObject::aSinh( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Inverse hyperbolic cosine function.  Returns an error object.
// Should be overridden by all classes for which acosh makes sense.
inline RpcObject* RpcObject::aCosh( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Inverse hyperbolic tangent function.  Returns an error object.
// Should be overridden by all classes for which atanh makes sense.
inline RpcObject* RpcObject::aTanh( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Euler gamma function.  Returns an error object.
// Should be overridden by all classes for which gamma makes sense.
inline RpcObject* RpcObject::Gamma( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Log gamma function.  Returns an error object.
// Should be overridden by all classes for which LogGamma makes sense.
inline RpcObject* RpcObject::LogGamma( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Factorial function.  Returns an error object.
// Should be overridden by all classes for which factorial makes sense.
inline RpcObject* RpcObject::Factorial( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Error function.  Returns an error object (not because it is the error
// function...).  Should be overridden by all classes for which erf
// makes sense.
inline RpcObject* RpcObject::Erf( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Complementary error function.  Returns an error object.
// Should be overridden by all classes for which erfc makes sense.
inline RpcObject* RpcObject::Erfc( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Real part of number.  Returns an error object.
// Should be overridden by all classes for which real() makes sense.
inline RpcObject* RpcObject::Real( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Imaginary part of number.  Returns an error object.
// Should be overridden by all classes for which imag() makes sense.
inline RpcObject* RpcObject::Imag( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Floor function.  Returns an error object.
// Should be overridden by all classes for which floor makes sense.
inline RpcObject* RpcObject::Floor( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Ceiling function.  Returns an error object.
// Should be overridden by all classes for which ceiling makes sense.
inline RpcObject* RpcObject::Ceil( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Nearest integer function.  Returns an error object.
// Should be overridden by all classes for which nearest integer makes sense.
inline RpcObject* RpcObject::NearInt( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Inverse function.  Returns an error object.
// Should be overridden by all classes for which inverse makes sense.
inline RpcObject* RpcObject::Inv( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Matrix transpose.
inline RpcObject* RpcObject::Trans( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Convert to integer constant.
inline RpcObject* RpcObject::ToConst( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Magnitude
inline RpcObject* RpcObject::Abs( void )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}


// Argument of complex number
inline RpcObject* RpcObject::Arg( int angleMode )
{
    RpcObject* temp = new RpcObject( SIMPLE_ERROR_TYPE );
    return( temp );
}   


// Return the data type.
unsigned int RpcObject::getType( void ) const
{
    return( type );
}



// arch-tag: rpc object base class
