/*  rpc -- A console-based reverse polish notation calculator
    Copyright (C) 2000, 2001, 2002 Paul Pelzl

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



/*  RpcObject.h
 *  Declaration of a class from which all the different types of
 *  rpc data will be derived.
 */

#ifndef RPCOBJECT_H
#define RPCOBJECT_H 1

 
#define DOUBLE_TYPE 0
#define COMPLEX_TYPE 1
#define REALMATRIX_TYPE 2
#define COMPLEXMATRIX_TYPE 3
#define INTEGERCONST_TYPE 4

#define ERROR_TYPE 255
#define SIMPLE_ERROR_TYPE 256


#include <cstdio>
#include <cstring>
#include "RpcCalcFlags.h"

using namespace std;

/* Note: an RpcObject with type SIMPLE_ERROR_TYPE is constructed from
 * RpcObject::RpcObject(), not a derived class.  There is no extra
 * error information embedded in the object.  This case is reserved 
 * for functions which make no sense when applied to the data type
 * for which they are called.
 */


class RpcObject
{
    protected:
        unsigned int type;   // Flag for type of data.

    public:    
        RpcObject();         // Create a default object.
        virtual ~RpcObject();        // Destroy object.
        
        RpcObject( unsigned int );    // Create an RpcObject specified type.
        
        
        RpcObject( const RpcObject& );        // copy constructor
        RpcObject& operator =( const RpcObject& );        // assignment
        
        unsigned int getType( void ) const;        // Return the type flag.

        virtual char* getDisplayString( RpcCalcFlags ) const;       // Return the display string for the object.
        virtual char* getFullscreenString( RpcCalcFlags ) const;    // Return the fullscreen display string for the object
        
        // All the following functions return SIMPLE_ERROR_TYPE objects by 
        // default.  Each operator should be overridden in all derived classes
        // for which it makes sense.
        virtual RpcObject* operator +( const RpcObject* );  // Add an object to 'this'
        virtual RpcObject* operator -( const RpcObject* );  // Subtract an object from 'this'
        virtual RpcObject* operator *( const RpcObject* );  // Multiply 'this' by an object
        virtual RpcObject* operator /( const RpcObject* );  // Divide 'this' by an object
        
        virtual RpcObject* operator -( void );        // Change sign of object       
        
        virtual RpcObject* Pow( const RpcObject* );   // Power function
        
        virtual RpcObject* Mod( const RpcObject* );   // Mod function
        
        virtual RpcObject* Sqrt( void );       // Square root
        virtual RpcObject* Sq( void );         // Square
        
        virtual RpcObject* Exp( void );        // Exponentiate
        virtual RpcObject* Log( void );        // Logarithm
        
        virtual RpcObject* Ten_x( void );      // Ten raised to the power x
        virtual RpcObject* Log10( void );      // Common logarithm
        
        virtual RpcObject* Sin( int );         // Sine
        virtual RpcObject* Cos( int );         // Cosine
        virtual RpcObject* Tan( int );         // Tangent
        
        virtual RpcObject* aSin( int );        // Inverse sine
        virtual RpcObject* aCos( int );        // Inverse cosine
        virtual RpcObject* aTan( int );        // Inverse tangent
        
        virtual RpcObject* Sinh( void );       // Hyperbolic sine
        virtual RpcObject* Cosh( void );       // Hyperbolic cosine
        virtual RpcObject* Tanh( void );       // Hyperboliic tangent
        
        virtual RpcObject* aSinh( void );      // Inverse hyperbolic sine
        virtual RpcObject* aCosh( void );      // Inverse hyperbolic cosine
        virtual RpcObject* aTanh( void );      // Inverse hyperbolic tangent
        
        virtual RpcObject* Gamma( void );      // Euler gamma function
        virtual RpcObject* LogGamma( void );   // Log gamma        
        virtual RpcObject* Factorial( void );  // Factorial function
        
        virtual RpcObject* Erf( void );        // Error function
        virtual RpcObject* Erfc( void );       // Complementary error function
        
        virtual RpcObject* Real( void );       // Real part
        virtual RpcObject* Imag( void );       // Imaginary part
        
        virtual RpcObject* Floor( void );      // Floor function
        virtual RpcObject* Ceil( void );       // Ceiling function
        virtual RpcObject* NearInt( void );    // Nearest integer
        
        virtual RpcObject* Inv( void );        // Inverse function
        
        virtual RpcObject* Trans( void );      // Matrix transpose
        
        virtual RpcObject* ToConst( void );      // Convert to integer constant
        
        virtual RpcObject* Abs( void );        // Magnitude
        virtual RpcObject* Arg( int );         // Phase of complex number
};


#endif  /* RPCOBJECT_H */



// arch-tag: rpc object base class (header)
