/*  rpc -- A console-based reverse polish notation calculator
    Copyright (C) 2000, 2001, 2002 Paul Pelzl

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include "RpcRealMatrix.h"


// Default constructor -- a 1x1 zero element.
RpcRealMatrix::RpcRealMatrix()
{
   type = REALMATRIX_TYPE;
   data = RealMatrix();
}


// Default destructor
RpcRealMatrix::~RpcRealMatrix()
{
}


// Copy constructor
RpcRealMatrix::RpcRealMatrix( const RpcRealMatrix& oldMatrix )
{
   type = REALMATRIX_TYPE;
   data = oldMatrix.data;
}


// Assignment operator.
RpcRealMatrix& RpcRealMatrix::operator =( const RpcRealMatrix& rhs )
{
   if( this != &rhs )   // check assignment
   {
      type = REALMATRIX_TYPE;
      data = rhs.data;
   }
   
   return( *this );
}


// Initialize a specific real matrix.
RpcRealMatrix::RpcRealMatrix( RealMatrix value )
{
   type = REALMATRIX_TYPE;
   data = value;
}


// Initialize a specific real matrix.
RpcRealMatrix::RpcRealMatrix( double* elements, int m, int n )
{
   type = REALMATRIX_TYPE;
   data = RealMatrix( elements, m, n );
}


// Return the display string for this object.
char* RpcRealMatrix::getDisplayString( RpcCalcFlags flags ) const
{
   int i=0, j=0;
   int m = data.m, n = data.n;
   char* temp = new char[ m * n * 30 ];
   char numBuffer[30];
   strcpy( temp, "[" );
   for( i=0; i<m; i++ )
   {
      strcat( temp, "[ " );
      for( j=0; j<n; j++ )
      {
         sprintf( numBuffer, "%.15g", data.mtrxdata[i*n+j] );
         strcat( temp, numBuffer );
         if( j < (n-1) )
         {
            strcat( temp, ", " );
         }
      }
      strcat( temp, " ]" );
   }
   strcat( temp, "]" );
   return( temp );
}


// Return the fullscreen display string for this object.
char* RpcRealMatrix::getFullscreenString( RpcCalcFlags flags ) const
{
   int i=0, j=0;
   int m = data.m, n = data.n;
   char* temp = new char[ m * n * 30 ];
   char numBuffer[30];
   int width, maxwidth[n];
   
   //First get the maximum field width for each column
   for( j=0; j<n; j++ )
   {
      maxwidth[j] = 0;
      for( i=0; i<m; i++ )
      {   
         sprintf( numBuffer, "%-.15g", data.mtrxdata[i*n+j] );
         width = strlen( numBuffer );
         if( width > maxwidth[j] )
            maxwidth[j] = width;
      }
   }
      
   
   strcpy( temp, "[" );
   for( i=0; i<m; i++ )
   {
      if( i != 0 )
         strcat( temp, " " );
      strcat( temp, "[ " );
      for( j=0; j<n; j++ )
      {
         sprintf( numBuffer, "%*.15g", maxwidth[j], data.mtrxdata[i*n+j] );
         strcat( temp, numBuffer );
         if( j < (n-1) )
         {
            strcat( temp, ", " );
         }
      }
      
      if( i == m-1 )
      {
         strcat( temp, " ]" );
      }
      else
      {
         strcat( temp, " ]\n" );
      }
   }
   strcat( temp, "]\n" );
   
   return( temp );
}



// Addition operator.
RpcObject* RpcRealMatrix::operator +(const RpcObject* rhs)
{
   RpcObject* temp;
   ComplexMatrix* dummy;
   
   // Branch according to data type.
   switch( rhs -> getType() )
   {
      // Add a real matrix to a real matrix.
      case REALMATRIX_TYPE:
         // check dimension.
         if( data.m == ((RpcRealMatrix*)rhs)->getData().m && 
               data.n == ((RpcRealMatrix*)rhs)->getData().n )
         {
            temp = new RpcRealMatrix( data + ((RpcRealMatrix*)rhs)->getData() );
         }
         else
         {
            temp = new RpcError;
         }         
         break;
         
      // Add a complex matrix to a real matrix.
      case COMPLEXMATRIX_TYPE:
         // check dimension.
         if( data.m == ((RpcComplexMatrix*)rhs)->getData().m &&
               data.n == ((RpcComplexMatrix*)rhs)->getData().n )
         {
            dummy = data + ((RpcComplexMatrix*)rhs)->getData();
            temp = new RpcComplexMatrix( *dummy );
            delete dummy;
         }
         else
         {
            temp = new RpcError;
         }
         break;
         
      // Default is an error.
      default:
         temp = new RpcError;
         break;
   }
   
   return( temp );
}

         

// Subtraction operator.
RpcObject* RpcRealMatrix::operator -(const RpcObject* rhs)
{
   RpcObject* temp;
   ComplexMatrix* dummy;
   
   // Branch according to data type.
   switch( rhs -> getType() )
   {
      // Subtract a real matrix from a real matrix.
      case REALMATRIX_TYPE:
         // check dimension.
         if( data.m == ((RpcRealMatrix*)rhs)->getData().m && 
               data.n == ((RpcRealMatrix*)rhs)->getData().n )
         {
            temp = new RpcRealMatrix( data - ((RpcRealMatrix*)rhs)->getData() );
         }
         else
         {
            temp = new RpcError;
         }
         break;
         
         
      // Subtract a complex matrix from a real matrix.
      case COMPLEXMATRIX_TYPE:
         // check dimension
         if( data.m == ((RpcComplexMatrix*)rhs)->getData().m &&
               data.n == ((RpcComplexMatrix*)rhs)->getData().n )
         {
            dummy = data - ((RpcComplexMatrix*)rhs)->getData();
            temp = new RpcComplexMatrix( *dummy );
            delete dummy;
         }
         else
         {
            temp = new RpcError;
         }
         break;
         
      // Default is an error.
      default:
         temp = new RpcError;
         break;
   }
   
   return( temp );
}



// Multiplication operator.
RpcObject* RpcRealMatrix::operator *( const RpcObject* rhs )
{
   RpcObject* temp;
   ComplexMatrix* dummy;
   
   // Branch according to dta type.
   switch( rhs -> getType() )
   {
      // Multiply a real matrix by a double.
      case DOUBLE_TYPE:
         temp = new RpcRealMatrix( data * ((RpcDouble*)rhs)->getData() );
         break;
      
      // Multiply a real matrix by a complex.
      case COMPLEX_TYPE:
         dummy = data * ((RpcComplex*)rhs)->getData();
         temp = new RpcComplexMatrix( *dummy );
         delete dummy;
         break;
      
      // Multiply two real matrices.
      case REALMATRIX_TYPE:
         // check dimension.
         if( data.n == ((RpcRealMatrix*)rhs)->getData().m )
         {
            temp = new RpcRealMatrix( data * ((RpcRealMatrix*)rhs)->getData() );
         }
         else
         {
            temp = new RpcError;
         }
         break;
         
      case COMPLEXMATRIX_TYPE:
         // check dimension
         if( data.n == ((RpcComplexMatrix*)rhs)->getData().m )
         {
            dummy = data * ((RpcComplexMatrix*)rhs)->getData();
            temp = new RpcComplexMatrix( *dummy );
            delete dummy;
         }
         else
         {
            temp = new RpcError;
         }
         break;
         
      // Default is an error.
      default:
         temp = new RpcError;
         break;
   }
   
   return( temp );
}



// Division operator.
RpcObject* RpcRealMatrix::operator /( const RpcObject* rhs )
{
   RpcObject* temp;
   ComplexMatrix* dummy;
   
   // Branch according to dta type.
   switch( rhs -> getType() )
   {
      // Divide a real matrix by a double.
      case DOUBLE_TYPE:
         temp = new RpcRealMatrix( data / ((RpcDouble*)rhs)->getData() );
         break;

      // Divide a real matrix by a complex.
      case COMPLEX_TYPE:
         dummy = data / ((RpcComplex*)rhs)->getData() ;
         temp = new RpcComplexMatrix( *dummy );
         delete dummy;
         break;
           
      // Default is an error.
      default:
         temp = new RpcError;
         break;
   }
   
   return( temp );
}




// Change sign operator.
inline RpcObject* RpcRealMatrix::operator -( void )
{
   RpcObject* temp = new RpcRealMatrix( -data );
   return( temp );
}


// Matrix inversion.
inline RpcObject* RpcRealMatrix::Inv( void )
{
   RpcObject* temp;
   
   // check for square matrix
   if( data.m == data.n )
   {
      RealMatrix inverse = data.Inv();
      if( inverse.singular )
      {
         temp = new RpcError( ERRORCODE_SINGULAR );
      }
      else
      {
         temp = new RpcRealMatrix( inverse );
      }
   }
   else
   {
      temp = new RpcError;
   }
   
   return( temp );
}



// Matrix inversion.
inline RpcObject* RpcRealMatrix::Trans( void )
{
   RpcObject* temp = new RpcRealMatrix( data.Trans() );
   return( temp );
}



// Return the data.
inline RealMatrix RpcRealMatrix::getData( void )
{
   return( data );
}




// arch-tag: rpc real matrix data type
