/*  rpc -- A console-based reverse polish notation calculator
    Copyright (C) 2000, 2001, 2002 Paul Pelzl

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


/*  RpcStack.cpp
 *  Class definition for a stack that contains RpcObjects.
 *  Memory is dynamically allocated (and deallocated).  Contains
 *  the standard push() and pop(), as well as various other
 *  more advanced features accessible from stack browsing mode.
 */

#include "RpcStack.h"



// Default constructor.  Creates an empty stack.
RpcStack::RpcStack()
{
    actualSize = DEFAULT_STACK_SIZE;
    element = new RpcObject*[DEFAULT_STACK_SIZE];
    count = 0;
}


// Default destructor.  Frees memory.
RpcStack::~RpcStack()
{    
    int i;
    
    for(i=0; i<count; i++)
    {
        delete element[i];
    }
    
    delete[] element;
}


// Copy constructor.  Very carefully preserves proper
// pointer typing.
RpcStack::RpcStack( const RpcStack& newStack )
{
    int i=0;
    
    count = newStack.count;
    actualSize = newStack.actualSize;
    
    // Allocate the array
    element = new RpcObject*[actualSize];

    // Copy over the new memory, carefully preserving proper storage type.
    for(i=0; i<count; i++)
    {
        switch( newStack.element[i] -> getType() )
        {
            case DOUBLE_TYPE:
                element[i] = new RpcDouble( *((RpcDouble*)(newStack.element[i])) );
                break;
            
            case COMPLEX_TYPE:
                element[i] = new RpcComplex( *((RpcComplex*)(newStack.element[i])) );
                break;
             
            case REALMATRIX_TYPE:
                element[i] = new RpcRealMatrix( *((RpcRealMatrix*)(newStack.element[i])) );
                break;
            
            case COMPLEXMATRIX_TYPE:
                element[i] = new RpcComplexMatrix( *((RpcComplexMatrix*)(newStack.element[i])) );
                break;
                
            case INTEGERCONST_TYPE:
                element[i] = new RpcIntegerConst( *((RpcIntegerConst*)(newStack.element[i])) );
                break;
               
            default:
                cerr << "Error in RpcStack::RpcStack( const RpcStack& ) : Type " << 
                    newStack.element[i] -> getType() << " data not supported in RpcStack.\n";
                break;
        }
    }
}


// Assignment operator.  Carefully preserves proper pointer typing.
RpcStack& RpcStack::operator =( const RpcStack& newStack )
{    
    int i=0;
    
    if( this != &newStack )        // Check self-assignment
    {
        // Delete the current memory.
        for(i=0; i<count; i++)
        {
            delete element[i];
        }
        
        if( newStack.count > actualSize )  // Need to reallocate pointer array
        {
            delete[] element;
            actualSize = newStack.actualSize;
            element = new RpcObject*[actualSize];
        }
    
        count = newStack.count;
        
        // Copy over the new memory, carefully using proper storage type.
        for(i=0; i<count; i++)
        {
            switch( newStack.element[i] -> getType() )
            {
                case DOUBLE_TYPE:
                    element[i] = new RpcDouble( *((RpcDouble*)(newStack.element[i])) );
                    break;
                
                case COMPLEX_TYPE:
                    element[i] = new RpcComplex( *((RpcComplex*)(newStack.element[i])) );
                    break;

                case REALMATRIX_TYPE:
                    element[i] = new RpcRealMatrix( *((RpcRealMatrix*)(newStack.element[i])) );
                    break;
                    
                case COMPLEXMATRIX_TYPE:
                    element[i] = new RpcComplexMatrix( *((RpcComplexMatrix*)(newStack.element[i])) );
                    break;
                    
                case INTEGERCONST_TYPE:
                    element[i] = new RpcIntegerConst( *((RpcIntegerConst*)(newStack.element[i])) );
                    break;
                                        
                default:
                    cerr << "Error in RpcStack::operator = : Type " <<
                        newStack.element[i] -> getType() << " data not supported in RpcStack.\n";
                    break;
            }
        }
    }
    
    return( *this );
}



// Place an RpcObject on the stack.
void RpcStack::push( const RpcObject* input )
{
    if( count >= actualSize )
    {
        // Need to reallocate memory.
        actualSize += STACK_SIZE_ADDON;
        RpcObject** temp = new RpcObject*[actualSize];
        
        // Copy memory over to new temp stack.  (Actually
        // only copying over pointers to the data.)
        for( int i=0; i<count; i++ )
        {
            temp[i] = element[i];
        }
        
        // Get rid of the old stack, and replace with the new one.
        delete[] element;
        element = temp;
    }
        
    switch( input -> getType() )
    {
        case DOUBLE_TYPE:
            element[count] = new RpcDouble( *((RpcDouble*)input) );
            break;
    
        case COMPLEX_TYPE:
            element[count] = new RpcComplex( *((RpcComplex*)input) );
            break;
        
        case REALMATRIX_TYPE:
            element[count] = new RpcRealMatrix( *((RpcRealMatrix*)input) );
            break;
            
        case COMPLEXMATRIX_TYPE:
            element[count] = new RpcComplexMatrix( *((RpcComplexMatrix*)input) );
            break;      
            
        case INTEGERCONST_TYPE:
            element[count] = new RpcIntegerConst( *((RpcIntegerConst*)input) );
            break;        
            
        case ERROR_TYPE:
        default:
            element[count] = new RpcError( *((RpcError*)input) );
            break;
    }
        
    count++;
}
        


// Pop an RpcObject off the stack.
RpcObject* RpcStack::pop( void )
{
    RpcObject* temp = NULL;
    
    if( (count <= actualSize - STACK_SIZE_ADDON) && 
            (actualSize - STACK_SIZE_ADDON >= DEFAULT_STACK_SIZE) )
    {
        // We can deallocate some memory in this case.
        actualSize -= STACK_SIZE_ADDON;
        RpcObject** tempStack = new RpcObject*[actualSize];
        
        // Copy memory over to new temp stack.  (Actually, only
        // copy over pointers to the data.)
        for(int i=0; i<count; i++)
        {
            tempStack[i] = element[i];
        }
        
        // Delete the old stack, and replace with the new one.
        delete[] element;
        element = tempStack;
    }
    
    
    if( count > 0 )
    {
        count--;
        temp = element[count];        // Just return a pointer to the data.
                                      // The memory is freed at a higher level.
    }
    else
    {
        temp = new RpcError;
    }
        
    return( temp );
}



// Clear all stack elements.
void RpcStack::clearAll( void )
{
    // Delete the old stack memory.
    for( int i=0; i<count; i++ )
    {
        delete element[i];
    }
    delete[] element;
    
    // Create new stack memory, with original allocation parameters.
    actualSize = DEFAULT_STACK_SIZE;
    element = new RpcObject*[DEFAULT_STACK_SIZE];
    count = 0;    
}



// Drop a specific stack element.
void RpcStack::drop( int level )
{
    RpcObject** temp = NULL;
    int i;
    
    if( level >= 0 && level < count )
    {
        // Create a new temp stack.
        temp = new RpcObject*[actualSize];
        
        // Copy over the elements below the specified level.
        for(i=0; i<level; i++)
        {
            temp[i] = element[i];
        }
        
        // Delete the specified element (no leak here!)
        delete element[level];
        
        // Copy over the elements above the specified level
        for(i=level+1; i<count; i++)
        {
            temp[i-1] = element[i];
        }
        
        // Clean up
        delete[] element;
        element = temp;
        count--;
    }
}
    


// Keep only the top n elements.
void RpcStack::keepn( int num )
{
    int i, coeff;
    
    if( num > 0 && num < count )
    {
        // Create a temporary stack.
        
        // First, choose an appropriate actualSize.
        if( num > DEFAULT_STACK_SIZE )
        {
            coeff = (num - DEFAULT_STACK_SIZE) / STACK_SIZE_ADDON;
            coeff++;
            actualSize = DEFAULT_STACK_SIZE + (coeff*STACK_SIZE_ADDON);
        }
        else
        {
            actualSize = DEFAULT_STACK_SIZE;
        }
        
        // Allocate the temporary stack.
        RpcObject** temp = new RpcObject*[actualSize];
        
        // Copy over the appropriate elements.
        for(i=0; i<num; i++)
        {
            temp[i] = element[count-num+i];
        }
        
        // Delete the remaining elements.
        for(i=0; i<(count-num); i++)
        {
            delete element[i];
        }
        
        // Delete the old memory stack, and copy over the new one.
        delete[] element;
        element = temp;
        count = num;
    }
}



// Cyclically roll top n elements toward top of stack.
void RpcStack::rollUp( int num )
{
    RpcObject* temp = NULL;
    int i;
    
    if( num > 0 && num <= count )
    {
        // Just switch all the pointers contained in element[] to point
        // to the right memory.
        temp = element[count-1];
        
        for(i=count-1; i>(count-num); i--)
        {
            element[i] = element[i-1];
        }
        
        element[count-num] = temp;
    }
}



// Cyclically roll top n elements toward bottom of stack.
void RpcStack::rollDown( int num )
{
    RpcObject* temp = NULL;
    int i;
    
    if( num > 0 && num <= count )
    {
        // Just switch all the pointers contained in element[] to point
        // to the right memory.
        temp = element[count-num];
        
        for(i=count-num; i<(count-1); i++)
        {
            element[i] = element[i+1];
        }
        
        element[count-1] = temp;
    }
}



// Return the number of stack elements.
int RpcStack::numElements( void ) const
{
    return( count );
}


// Return the display string from the object at the specified level.
char* RpcStack::getDisplayString( int level, RpcCalcFlags flags ) const
{
    char* temp;
    
    if( (level>0) && (level<=count) )
    {
        temp = element[level-1] -> getDisplayString( flags );
    }
    else
    {
        temp = new char[256];
        strcpy( temp, "<no object at this level>" );
    }
    
    return( temp );
}



// Return the fullscreen display string from the object at the specified level.
char* RpcStack::getFullscreenString( int level, RpcCalcFlags flags ) const
{
    char* temp;
    
    if( (level>0) && (level<=count) )
    {
        temp = element[level-1] -> getFullscreenString( flags );
    }
    else
    {
        temp = new char[256];
        strcpy( temp, "<no object at this level>\n" );
    }
    
    return( temp );
}



// Return the data type flag of the top stack element.
unsigned int RpcStack::getTopType( void )
{
    unsigned int temp;
    
    if( count > 0 )
    {
        temp = element[count-1] -> getType();
    }
    else
    {
        temp = ERROR_TYPE;
    }
    
    return( temp );
}


// Return constant access to the element at the specified level.
// Use with caution!  Make sure the use *really is* constant.
RpcObject* const RpcStack::viewElement( int level ) const
{
    RpcObject* temp = NULL;
    
    if( level>0 && level<=count )
    {
        temp = element[level-1];
    }
    else
    {
        temp = stackError;        // Return error state
    }
    
    return( temp );
}


// Save stack to a data file.
void RpcStack::save( ofstream& datafile ) const
{
    int dataType;
    char version[10];
    double doubleData = 0.0;
    Complex complexData = 0.0;
    RealMatrix matrixData;
    ComplexMatrix cMatrixData;
    IntegerConst longdata;
    
    strcpy( version, VERSION );
    datafile.write(version, sizeof(version)); // Save the rpc version info.
    
    datafile.write((char*)(&count), sizeof(count));      // Write number of elements.
    datafile.write((char*)(&actualSize), sizeof(actualSize));
    
    for(int i=0; i<count; i++)
    {
        dataType = element[i] -> getType();
        datafile.write((char*)(&dataType), sizeof(dataType) );
        
        switch( dataType )
        {
            case DOUBLE_TYPE:
                doubleData = ((RpcDouble*)(element[i]))->getData();
                datafile.write((char*)(&doubleData), sizeof(doubleData) );
                break;
                
            case COMPLEX_TYPE:
                complexData = ((RpcComplex*)(element[i]))->getData();
                datafile.write((char*)(&complexData), sizeof(complexData) );
                break;
            
            case REALMATRIX_TYPE:
                matrixData = ((RpcRealMatrix*)(element[i]))->getData();
                datafile.write((char*)(&(matrixData.m)), sizeof(matrixData.m) );
                datafile.write((char*)(&(matrixData.n)), sizeof(matrixData.n) );
                for(int j=0; j<matrixData.m*matrixData.n; j++)
                {
                    datafile.write((char*)(&(matrixData.mtrxdata[j])),
                                       sizeof(matrixData.mtrxdata[j]) );
                }
                break;                
            
            case COMPLEXMATRIX_TYPE:
                cMatrixData = ((RpcComplexMatrix*)(element[i]))->getData();
                datafile.write((char*)(&(cMatrixData.m)), sizeof(cMatrixData.m) );
                datafile.write((char*)(&(cMatrixData.n)), sizeof(cMatrixData.n) );
                for(int j=0; j < cMatrixData.m * cMatrixData.n; j++)
                {
                    datafile.write((char*)(&((cMatrixData.mtrxdata[j]).re)),
                            sizeof(double) );
                    datafile.write((char*)(&((cMatrixData.mtrxdata[j]).im)), 
                            sizeof(double) );
                }
                break;
                
            case INTEGERCONST_TYPE:
                longdata = ((RpcIntegerConst*)(element[i]))->getData();
                datafile.write((char*)(&longdata), sizeof(longdata));
                break;
                
        }
    }
}    


// Load stack from a data file.
void RpcStack::load( ifstream& datafile )
{
    int dataType, actualSizeTemp;
    char version[10];
    double doubleData = 0.0;
    Complex complexData = 0.0;
    RealMatrix matrixData;
    ComplexMatrix cMatrixData;
    int mtrxRows, mtrxCols;
    double* mtrxdata;
    Complex* cmtrxdata;
    double real, imag;
    IntegerConst longdata;
    
    datafile.read(version, sizeof(version));
    
    // Only load the stack if the version matches.
    if( strcmp(version, VERSION) == 0 )
    {
        datafile.read((char*)(&count), sizeof(count));
        datafile.read((char*)(&actualSizeTemp), sizeof(actualSizeTemp) );

        // Reallocate memory if necessary.
        if( count >= actualSize)
        {
            delete[] element;
            actualSize = actualSizeTemp;
            element = new RpcObject*[actualSize];
        }
        
        for(int i=0; i<count; i++)
        {
            datafile.read((char*)(&dataType), sizeof(dataType) );
            
            switch( dataType )
            {
                case DOUBLE_TYPE:
                    datafile.read((char*)(&doubleData), sizeof(doubleData) );
                    element[i] = new RpcDouble( doubleData );
                    break;
                    
                case COMPLEX_TYPE:
                    datafile.read((char*)(&complexData), sizeof(complexData) );
                    element[i] = new RpcComplex( complexData );
                    break;
                    
                case REALMATRIX_TYPE:
                    // First determine the proper storage size for an mxn matrix
                    datafile.read((char*)(&mtrxRows), sizeof(mtrxRows) );
                    datafile.read((char*)(&mtrxCols), sizeof(mtrxCols) );
                    mtrxdata = new double[mtrxRows*mtrxCols];
                    for(int j=0; j<mtrxRows*mtrxCols; j++)
                    {
                        datafile.read((char*)(&(mtrxdata[j])), sizeof(mtrxdata[j]) );
                    }
                    matrixData = RealMatrix( mtrxdata, mtrxRows, mtrxCols );
                    delete[] mtrxdata;
                    element[i] = new RpcRealMatrix( matrixData );
                    break;    
                
                case COMPLEXMATRIX_TYPE:
                    // First determine the proper storage size for an mxn matrix
                    datafile.read((char*)(&mtrxRows), sizeof(mtrxRows) );
                    datafile.read((char*)(&mtrxCols), sizeof(mtrxCols) );
                    cmtrxdata = new Complex[mtrxRows*mtrxCols];
                    for(int j=0; j<mtrxRows*mtrxCols; j++)
                    {
                        datafile.read((char*)(&real), sizeof(real) );
                        datafile.read((char*)(&imag), sizeof(imag) );
                        cmtrxdata[j] = Complex(real, imag);
                    }
                    cMatrixData = ComplexMatrix( cmtrxdata, mtrxRows, mtrxCols );
                    delete[] cmtrxdata;
                    element[i] = new RpcComplexMatrix( cMatrixData );
                    break;                    
                
                case INTEGERCONST_TYPE:
                    datafile.read((char*)(&longdata), sizeof(longdata) );
                    element[i] = new RpcIntegerConst( longdata );
                    break;
                    
                default:
                    perror( "Error in stack::load().  Unknown data type.\n" );
                    exit(1);
                    break;
            }
        }
    }
}



// arch-tag: rpc data stack
